/*
 * Copyright (c) 2021, NVIDIA CORPORATION. All rights reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

#include "ClientCommon.h"
#include "CloudXRFileLogger.h"

#include <thread>
#include <chrono>

using std::chrono::milliseconds;
using std::this_thread::sleep_for;

// macro to switch between CloudXRFileLogger for logging, vs writing stdout here directly.
#define LOG_TO_FILE 1

// this is called by helper macros, and by the LogMessage stub implementation for the client app.
extern "C" void dispatchLogMsg(cxrLogLevel level, cxrMessageCategory category, void *extra, const char *tag, const char *fmt, ...)
{
    // Not using category right now.
    va_list aptr;
    va_start(aptr, fmt);
#if !LOG_TO_FILE
    const int bufsize = 8192;
    char buffer[bufsize];
    vsnprintf(buffer, bufsize, fmt, aptr);
#else
    g_logFile.logva(level, tag, fmt, aptr);
#endif
    va_end(aptr);

#if !LOG_TO_FILE
    // now build output message with full log prefix...
    char message[bufsize];
    time_t t;
    struct tm *tmptr;
    time(&t);
    tmptr = localtime(&t);
    sprintf(message, "%02d:%02d:%02d  %c (%s)  %s\n",
        tmptr->tm_hour, tmptr->tm_min, tmptr->tm_sec,
        cxrLLToChar(level), tag, buffer);

#if defined(WIN32)
    if (IsDebuggerPresent()) // if debugger, print to Output panel.
        OutputDebugStringA(message);
#endif
    // we always print to stdout.
    fputs(message, stdout);
#endif
}

cxrError ClientCommon::Init(const CloudXR::ClientOptions& options, const std::string &outputPath)
{
    cxrReceiverDesc desc = { 0 };

    // TODO, in the future, options might want to have an override for output
    // directory.  at that point, we'll check that first, and use param otherwise.
    if (outputPath.empty())
    {
        // TODO: should we log an error here?
        desc.appOutputPath[0] = 0;
    }
    else
    {
        strncpy(desc.appOutputPath, outputPath.c_str(), CXR_MAX_PATH - 1);
        desc.appOutputPath[CXR_MAX_PATH - 1] = 0; // just in case we ran out of space...
    }

    m_clientProxy.GetTrackingState = [](void* context, cxrVRTrackingState* trackingState)
    {
        return reinterpret_cast<ClientCommon*>(context)->GetTrackingState(trackingState);
    };
    m_clientProxy.TriggerHaptic = [](void* context, const cxrHapticFeedback* haptic)
    {
        return reinterpret_cast<ClientCommon*>(context)->TriggerHaptic(haptic);
    };
    m_clientProxy.RenderAudio = [](void* context, const cxrAudioFrame* audioFrame)
    {
        return reinterpret_cast<ClientCommon*>(context)->RenderAudio(audioFrame);
    };
    m_clientProxy.ReceiveUserData = [](void* context, const void* data, uint32_t size)
    {
        return reinterpret_cast<ClientCommon*>(context)->ReceiveUserData(data, size);
    };
    m_clientProxy.UpdateClientState = [](void* context, cxrClientState state, cxrError error)
    {
        switch (state)
        {
        case cxrClientState_ConnectionAttemptInProgress:
            CXR_LOGI("Connection attempt in progress.");
            break;
        case cxrClientState_StreamingSessionInProgress:
            CXR_LOGI("Connection attempt succeeded.");
            break;
        case cxrClientState_ConnectionAttemptFailed:
            CXR_LOGE("Connection attempt failed with error: %s", cxrErrorString(error));
            break;
        case cxrClientState_Disconnected:
            CXR_LOGE("Server disconnected with error: %s", cxrErrorString(error));
            break;
        default:
            CXR_LOGI("Client state updated: %d, error: %s", state, cxrErrorString(error));
            break;
        }

        ClientCommon *client = reinterpret_cast<ClientCommon*>(context);
        client->m_clientState = state;
        client->m_clientError = error;
    };

    m_clientProxy.LogMessage = [](void* context, cxrLogLevel level, cxrMessageCategory category, void* extra, const char* tag, const char* const messageText)
    {
        // Here we call our helper fn to output same way as the log macros will.
        // note that at the moment, we don't need/use the client context.
        dispatchLogMsg(level, category, extra, tag, messageText);
    };

    m_clientProxy.clientContext = this;

    desc.requestedVersion = CLOUDXR_VERSION_DWORD;
    desc.deviceDesc = m_deviceDesc;
    desc.clientCallbacks = m_clientProxy;
    desc.debugFlags = m_clientOptions.mDebugFlags;
    desc.logMaxSizeKB = m_clientOptions.mLogMaxSizeKB;
    m_serverIP = m_clientOptions.mServerIP;
    m_userData = m_clientOptions.mUserData;

    cxrGraphicsContext context;
    desc.shareContext = nullptr;
    if (m_clientOptions.mGfxType != cxrGraphicsContext_D3D11)
    {
        GetGraphicsContext(&context, m_clientOptions.mGfxType);
        desc.shareContext = &context; // temporary
    }

    cxrError err = cxrCreateReceiver(&desc, &m_receiver);
    if (err != cxrError_Success)
    {
        CXR_LOGE("Failed to create CloudXR receiver. Error %d, %s.", err, cxrErrorString(err));
        return err;
    }

    return cxrError_Success;
}

cxrError ClientCommon::Connect()
{
    m_connectionDesc.async = true;
    m_connectionDesc.useL4S = m_clientOptions.mUseL4S;
    m_connectionDesc.clientNetwork = m_clientOptions.mClientNetwork;
    m_connectionDesc.topology = m_clientOptions.mTopology;
    cxrError err = cxrConnect(m_receiver, m_clientOptions.mServerIP.c_str(), &m_connectionDesc);

    if (!m_connectionDesc.async)
    {
        if (err != cxrError_Success)
        {
            CXR_LOGE("Failed to connect to CloudXR server at %s. Error %d, %s.", m_clientOptions.mServerIP.c_str(), (int)err, cxrErrorString(err));
            cxrDestroyReceiver(m_receiver);
            m_receiver = nullptr;
            return err;
        }
    }

    uint32_t timeOutMs = 1000;
    RunLoop(m_receiver, m_clientOptions.mNumVideoStreams, timeOutMs);

    cxrDestroyReceiver(m_receiver);

    return cxrError_Success;
}

void ClientCommon::SendAudio(uint8_t* stream, int len)
{
    cxrAudioFrame audioFrame{};
    audioFrame.streamBuffer = (int16_t*)stream;
    audioFrame.streamSizeBytes = len;
    cxrSendAudio(m_receiver, &audioFrame);
}

void ClientCommon::SendUserData(std::string userData)
{
    cxrInputEvent event;
    event.type = cxrInputEventType_Generic;
    event.event.genericInputEvent.data = (uint8_t*)&userData[0];
    event.event.genericInputEvent.sizeInBytes = (uint32_t)userData.size() + 1;
    cxrSendInputEvent(m_receiver, &event);
}

void ClientCommon::ReceiveUserData(const void* data, uint32_t size)
{
    // Logging first dword for testing data channel from server to client.
    uint32_t* frameNumber = (uint32_t*)data;
    CXR_LOGV("Received frame number: %u", *frameNumber);
}

cxrError ClientCommon::RunLoop(cxrReceiverHandle receiver, uint32_t numVideoSteams, uint32_t timeOutMs)
{
    cxrError error = cxrError_Success;
    cxrFramesLatched framesLatched;
    cxrConnectionStats stats;
    int framesUntilStats = 60;

    while (m_clientState != cxrClientState_Exiting)
    {
        if (m_clientState == cxrClientState_ReadyToConnect || m_clientState == cxrClientState_ConnectionAttemptInProgress)
        {
            // status indication via log, using LOGW to make it more visible.
            static int32_t attemptCount = 0;
            if (attemptCount++ % 60 == 0) CXR_LOGW("Waiting for server...");
            sleep_for(milliseconds(50)); // sleep an approx 20hz tick if not connected, to not overload system.
        }
        if (m_clientState == cxrClientState_Disconnected || m_clientState == cxrClientState_ConnectionAttemptFailed)
        {
            CXR_LOGE("Exiting due to connection failure.");
            m_clientState = cxrClientState_Exiting;
        }
        else if (m_clientState == cxrClientState_StreamingSessionInProgress)
        {
            // TODO we currently latch all at once.
            //  I wonder if we could gain some parallelism by latching a frame and submit it,
            //  then latch the next frame (which might sleep) while the render-out is happening.
            error = cxrLatchFrame(receiver, &framesLatched, cxrFrameMask_All, timeOutMs);
            if (error == cxrError_Success)
            {
                SubmitFrame(&framesLatched);
                cxrReleaseFrame(receiver, &framesLatched);
                framesUntilStats--;
            }
            else if (error == cxrError_Frame_Not_Ready)
            {
                CXR_LOGW("LatchFrame failed, frame not ready for %d ms", timeOutMs);
            }
            else
            {
                CXR_LOGE("LatchFrame failed with error: %s", cxrErrorString(error));
            }

            // Log connection stats every 3 seconds
            const int STATS_INTERVAL_SEC = 3;
            if (framesUntilStats <= 0 &&
                cxrGetConnectionStats(receiver, &stats) == cxrError_Success)
            {
                // Capture the key connection statistics
                char statsString[64] = { 0 };
                snprintf(statsString, 64, "FPS: %6.1f    Bitrate (kbps): %5d    Latency (ms): %3d", stats.framesPerSecond, stats.bandwidthUtilizationKbps, stats.roundTripDelayMs);

                // Turn the connection quality into a visual representation along the lines of a signal strength bar
                char qualityString[64] = { 0 };
                snprintf(qualityString, 64, "Connection quality: [%c%c%c%c%c]",
                    stats.quality >= cxrConnectionQuality_Bad ? '#' : '_',
                    stats.quality >= cxrConnectionQuality_Poor ? '#' : '_',
                    stats.quality >= cxrConnectionQuality_Fair ? '#' : '_',
                    stats.quality >= cxrConnectionQuality_Good ? '#' : '_',
                    stats.quality == cxrConnectionQuality_Excellent ? '#' : '_');

                // There could be multiple reasons for low quality however we show only the most impactful to the end user here
                char reasonString[64] = { 0 };
                if (stats.quality <= cxrConnectionQuality_Fair)
                {
                    if (stats.qualityReasons == cxrConnectionQualityReason_EstimatingQuality)
                    {
                        snprintf(reasonString, 64, "Reason: Estimating quality");
                    }
                    else if (stats.qualityReasons & cxrConnectionQualityReason_HighLatency)
                    {
                        snprintf(reasonString, 64, "Reason: High Latency (ms): %3d", stats.roundTripDelayMs);
                    }
                    else if (stats.qualityReasons & cxrConnectionQualityReason_LowBandwidth)
                    {
                        snprintf(reasonString, 64, "Reason: Low Bandwidth (kbps): %5d", stats.bandwidthAvailableKbps);
                    }
                    else if (stats.qualityReasons & cxrConnectionQualityReason_HighPacketLoss)
                    {
                        if (stats.totalPacketsLost == 0)
                        {
                            snprintf(reasonString, 64, "Reason: High Packet Loss (Recoverable)");
                        }
                        else
                        {
                            snprintf(reasonString, 64, "Reason: High Packet Loss (%%): %3.1f", 100.0f * stats.totalPacketsLost / stats.totalPacketsReceived);
                        }
                    }
                }

                CXR_LOGD("%s    %s    %s", statsString, qualityString, reasonString);
                framesUntilStats = (int)stats.framesPerSecond * STATS_INTERVAL_SEC;
            }
        }

        if (!Update())
        {
            CXR_LOGI("Exiting due to user input");
            m_clientState = cxrClientState_Exiting;
        }
    }
    return error;
}