/*
 * Copyright (c) 2019-2020, NVIDIA CORPORATION. All rights reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

#pragma once

#define TEXH_TO_GLINT(texture) (static_cast<GLuint>(reinterpret_cast<intptr_t>(texture)))
#define TEXH_TO_D11TEX(texture) (reinterpret_cast<ID3D11Texture2D*>(texture))

#define GLINT_TO_TEXH(tex) (reinterpret_cast<TextureHandle>(tex))

#ifdef _WIN32
#ifndef WIN32_LEAN_AND_MEAN
#define WIN32_LEAN_AND_MEAN
#endif

#ifndef NOMINMAX
#define NOMINMAX
#endif

// Win32 API
#include <Windows.h>
#define _USE_MATH_DEFINES

// DX
#include <d3d11.h>
#else
#include <string.h>  // For memset
#endif

// OpenVR
#include <openvr.h>

// STL
#include <string>
#include <vector>
#include <map>
#include <queue>
#include <memory>

#define SAFE_RELEASE(x) do { if (x) { (x)->Release(); (x) = NULL; } } while (0)

// CloudXR
#include "CloudXRClient.h"
#include "CloudXRClientOptions.h"

#ifndef LOG_TAG
#define LOG_TAG "CloudXR Client"
#endif
#include "CloudXRFileLogger.h"

// Base class that owns the CloudXR reciever object and manages its lifecycle
class ClientCommon
{
public:
    ClientCommon() {};
    virtual ~ClientCommon() {};

    // Called before Run in order for a derived class to initialize itself
    // and populate the CloudXR device descriptor accordingly
    virtual cxrError Init(const CloudXR::ClientOptions& options, const std::string &outputPath);

    // Sets up the CloudXR receiver, connects it to the server and then starts
    // the main loop of the application to handle received frames
    virtual cxrError Connect();

    // Sends audio data from the client to the server
    virtual void SendAudio(uint8_t* stream, int len);

    // Sends user data from the client to the server
    virtual void SendUserData(std::string userData);

    // cxrClientCallbacks implemented by a derived class
    virtual void GetTrackingState(cxrVRTrackingState* trackingState) = 0;
    virtual void TriggerHaptic(const cxrHapticFeedback *haptic) = 0;
    virtual cxrBool RenderAudio(const cxrAudioFrame *audioFrame) = 0;

    // cxrClientCallbacks implemented by this base class
    virtual void ReceiveUserData(const void* data, uint32_t size);

protected:
    // Called by Run to get information from a derived class to
    // set up and establish the connection
    virtual void GetChaperone(cxrChaperone *chaperone) = 0;
    virtual void GetGraphicsContext(cxrGraphicsContext *context, cxrGraphicsContextType type) = 0;

    // Called by Run and implements the main loop of the client application
    // for XR and Generic mode respectively
    cxrError RunLoop(cxrReceiverHandle receiver, uint32_t numVideoSteams, uint32_t timeOutMs);

    // Called by RunLoop to delegate presentation of frames
    // as well as event handling to a derived class
    virtual cxrError SubmitFrame(cxrFramesLatched* framesLatched) = 0;
    virtual bool Update() = 0;

    cxrClientCallbacks     m_clientProxy = {};

    cxrReceiverHandle      m_receiver = nullptr;
    cxrDeviceDesc          m_deviceDesc = {};
    cxrConnectionDesc      m_connectionDesc = {};
    cxrClientState         m_clientState = cxrClientState_ReadyToConnect;
    cxrError               m_clientError = cxrError_Success;

    CloudXR::ClientOptions m_clientOptions;
    std::string            m_serverIP = {};
    std::string            m_userData = {};
    bool                   m_sendUserData = true;
};

