/*
 * Copyright (c) 2019-2020, NVIDIA CORPORATION. All rights reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

#include "ClientCommon.h"

#include <SDL.h>
#include "ClientOpenVR.h"
#include "CloudXRMatrixHelpers.h"

#ifdef _WIN32
#include <Shlobj.h>
#include <Shlwapi.h>
#include <PathCch.h>

#pragma comment(lib, "Shell32.lib")
#pragma comment(lib, "Shlwapi.lib")
#pragma comment(lib, "Pathcch.lib")
#else
#define TCHAR char
#define strcpy_s strcpy
#define MAX_PATH 256
#include <unistd.h>
#endif

#include <cassert>

namespace
{
    static const char* controllerNames[] =
    {
        "HTC Vive Wand",
        "Valve Index Controller",
        "Oculus Touch Controller", 
        "None",
    };

    static const char* inputsViveNew[] =
    {
        "menu/press",
        "system/press",
        "trackpad/touch",
        "trackpad/press",
        "trackpad/x",
        "trackpad/y",
        "trigger/press",
        "grip/press",
    };

    cxrInputValueType inputValuesViveNew[] =
    {
        cxrInputValueType_boolean, //menu/press
        cxrInputValueType_boolean, //system/press
        cxrInputValueType_boolean, //trackpad/touch
        cxrInputValueType_boolean, //trackpad/press
        cxrInputValueType_int32, //trackpad/x
        cxrInputValueType_int32, //trackpad/y
        cxrInputValueType_float32, //trigger/press (TBD: int32, for determinism?..)
        cxrInputValueType_boolean, //grip/press
    };

    static const char* inputsKnucklesNew[] =
    {
        "system/press",
        "a/touch",
        "a/press",
        "b/touch",
        "b/press",
        "trackpad/touch",
        "trackpad/press",
        "trackpad/scroll",
        "stick/touch",
        "stick/x",
        "stick/y",
        "trigger/touch",
        "trigger/press",
        "grip/touch",
        "grip/squeeze",
    };

    cxrInputValueType inputValuesKnucklesNew[] =
    {
        cxrInputValueType_boolean, //system/press,
        cxrInputValueType_boolean, //a/touch,
        cxrInputValueType_boolean, //a/press,
        cxrInputValueType_boolean, //b/touch,
        cxrInputValueType_boolean, //b/press,
        cxrInputValueType_boolean, //trackpad/touch,
        cxrInputValueType_boolean, //trackpad/press,
        cxrInputValueType_int32,   //trackpad/scroll,
        cxrInputValueType_boolean, //stick/touch,
        cxrInputValueType_float32,   //stick/x,
        cxrInputValueType_float32,   //stick/y,
        cxrInputValueType_boolean, //trigger/touch,
        cxrInputValueType_float32,   //trigger/press, (TBD: int32, for determinism?..)
        cxrInputValueType_boolean, //grip/touch,
        cxrInputValueType_float32,   //grip/squeeze,
    };

    static const char* inputsGeneric[] =
    {
        "/input/system/click",
        "/input/trigger/click",
        "/input/trigger/touch",
        "/input/trigger/value",
        "/input/grip/click",
        "/input/trackpad/click",
        "/input/trackpad/touch",
        "/input/trackpad/x",
        "/input/trackpad/y",
        "/input/application_menu/click",
        "/input/grip/value",
        "/input/joystick/click",
        "/input/joystick/touch",
        "/input/joystick/x",    
        "/input/joystick/y",   
        "/input/x/click",
        "/input/y/click",
        "/input/a/click",
        "/input/b/click",
        "/input/grip/touch",      
        "/input/grip/value",      
        "/input/grip/force",
        "/input/thumbstick/click",
        "/input/thumbstick/touch",
        "/input/thumbstick/x",
        "/input/thumbstick/y",
        "/input/a/click",         
        "/input/b/click",         
    };

    cxrInputValueType inputValuesGeneric[] =
    {
        cxrInputValueType_boolean, //input/system/click
        cxrInputValueType_boolean, //input/trigger/click
        cxrInputValueType_boolean, //input/trigger/touch
        cxrInputValueType_float32, //input/trigger/value
        cxrInputValueType_boolean, //input/grip/click
        cxrInputValueType_boolean, //input/trackpad/click
        cxrInputValueType_boolean, //input/trackpad/touch
        cxrInputValueType_float32, //input/trackpad/x
        cxrInputValueType_float32, //input/trackpad/y
        cxrInputValueType_boolean, //input/application_menu/click
        cxrInputValueType_float32, //input/grip/value
        cxrInputValueType_boolean, //input/joystick/click
        cxrInputValueType_boolean, //input/joystick/touch
        cxrInputValueType_float32, //input/joystick/x
        cxrInputValueType_float32, //input/joystick/y
        cxrInputValueType_boolean, //input/x/click
        cxrInputValueType_boolean, //input/y/click
        cxrInputValueType_boolean, //input/a/click
        cxrInputValueType_boolean, //input/b/click
        cxrInputValueType_boolean, //input/grip/touch
        cxrInputValueType_float32, //input/grip/value
        cxrInputValueType_float32, //input/grip/force
        cxrInputValueType_boolean, //input/thumbstick/click
        cxrInputValueType_boolean, //input/thumbstick/touch
        cxrInputValueType_float32, //input/thumbstick/x
        cxrInputValueType_float32, //input/thumbstick/y
        cxrInputValueType_boolean, //input/a/click
        cxrInputValueType_boolean, //input/b/click
    };

    static constexpr int inputWandLegacyCount = 10;

    static const char* inputsWandLegacy[inputWandLegacyCount] =
    {
        "/input/system/click",
        "/input/trigger/click",
        "/input/trigger/touch",
        "/input/trigger/value",
        "/input/grip/click",
        "/input/trackpad/click",
        "/input/trackpad/touch",
        "/input/trackpad/x",
        "/input/trackpad/y",
        "/input/application_menu/click",        
    };

    cxrInputValueType inputValuesWandLegacy[inputWandLegacyCount] =
    {
        cxrInputValueType_boolean, //input/system/click
        cxrInputValueType_boolean, //input/trigger/click
        cxrInputValueType_boolean, //input/trigger/touch
        cxrInputValueType_float32, //input/trigger/value
        cxrInputValueType_boolean, //input/grip/click
        cxrInputValueType_boolean, //input/trackpad/click
        cxrInputValueType_boolean, //input/trackpad/touch
        cxrInputValueType_float32, //input/trackpad/x
        cxrInputValueType_float32, //input/trackpad/y
        cxrInputValueType_boolean, //input/application_menu/click
    };

    static constexpr int inputKnucklesLegacyCount = 19;

    static const char* inputsKnucklesLegacy[inputKnucklesLegacyCount] =
    {
        "/input/system/click",
        "/input/trigger/click",
        "/input/trigger/touch",
        "/input/trigger/value",
        "/input/grip/click",
        "/input/trackpad/click",
        "/input/trackpad/touch",
        "/input/trackpad/x",
        "/input/trackpad/y",
        "/input/application_menu/click",
        "/input/grip/touch",      
        "/input/grip/value",      
        "/input/grip/force",
        "/input/thumbstick/click",
        "/input/thumbstick/touch",
        "/input/thumbstick/x",
        "/input/thumbstick/y",
        "/input/a/click",         
        "/input/b/click",         
    };

    cxrInputValueType inputValuesKnucklesLegacy[inputKnucklesLegacyCount] =
    {
        cxrInputValueType_boolean, //input/system/click
        cxrInputValueType_boolean, //input/trigger/click
        cxrInputValueType_boolean, //input/trigger/touch
        cxrInputValueType_float32, //input/trigger/value
        cxrInputValueType_boolean, //input/grip/click
        cxrInputValueType_boolean, //input/trackpad/click
        cxrInputValueType_boolean, //input/trackpad/touch
        cxrInputValueType_float32, //input/trackpad/x
        cxrInputValueType_float32, //input/trackpad/y
        cxrInputValueType_boolean, //input/application_menu/click
        cxrInputValueType_boolean, //input/grip/touch
        cxrInputValueType_float32, //input/grip/value
        cxrInputValueType_float32, //input/grip/force
        cxrInputValueType_boolean, //input/thumbstick/click
        cxrInputValueType_boolean, //input/thumbstick/touch
        cxrInputValueType_float32, //input/thumbstick/x
        cxrInputValueType_float32, //input/thumbstick/y
        cxrInputValueType_boolean, //input/a/click
        cxrInputValueType_boolean, //input/b/click
    };

    static constexpr int inputTouchLegacyCount = 21;

    static const char* inputsTouchLegacy[inputTouchLegacyCount] =
    {
        "/input/system/click",
        "/input/application_menu/click",
        "/input/trigger/click",
        "/input/trigger/touch",
        "/input/trigger/value",
        "/input/grip/click",
        "/input/grip/touch",
        "/input/grip/value",
        "/input/joystick/click",
        "/input/joystick/touch",
        "/input/joystick/x",
        "/input/joystick/y",
        "/input/a/click",
        "/input/b/click",
        "/input/x/click", // Touch has X/Y on L controller, so we'll map the raw strings.
        "/input/y/click",
        "/input/a/touch",
        "/input/b/touch",
        "/input/x/touch",
        "/input/y/touch",
        "/input/thumb_rest/touch",
    };

    cxrInputValueType inputValuesTouchLegacy[inputTouchLegacyCount] =
    {
        cxrInputValueType_boolean, //input/system/click
        cxrInputValueType_boolean, //input/application_menu/click
        cxrInputValueType_boolean, //input/trigger/click
        cxrInputValueType_boolean, //input/trigger/touch
        cxrInputValueType_float32, //input/trigger/value
        cxrInputValueType_boolean, //input/grip/click
        cxrInputValueType_boolean, //input/grip/touch
        cxrInputValueType_float32, //input/grip/value
        cxrInputValueType_boolean, //input/joystick/click
        cxrInputValueType_boolean, //input/joystick/touch
        cxrInputValueType_float32, //input/joystick/x
        cxrInputValueType_float32, //input/joystick/y
        cxrInputValueType_boolean, //input/a/click
        cxrInputValueType_boolean, //input/b/click
        cxrInputValueType_boolean, //input/x/click
        cxrInputValueType_boolean, //input/y/click
        cxrInputValueType_boolean, //input/a/touch
        cxrInputValueType_boolean, //input/b/touch
        cxrInputValueType_boolean, //input/x/touch
        cxrInputValueType_boolean, //input/y/touch
        cxrInputValueType_boolean, //input/thumb_rest/touch
    };
}

VisionClientOpenVR::VisionClientOpenVR(bool remapSysBtn) : ClientCommon()
{
    m_hmd = NULL;
    m_init = false;
    m_remapSysBtn = remapSysBtn;
    memset(&m_trackingState, 0, sizeof(m_trackingState));
}

VisionClientOpenVR::~VisionClientOpenVR()
{
    m_init = false;
    m_platform.Destroy();
    VR_Shutdown();
}

cxrError VisionClientOpenVR::Init(const CloudXR::ClientOptions& options, const std::string &outputPath)
{
    m_clientOptions = options; // cache for Connect to use some options later

    EVRInitError eError = VRInitError_None;

    // On Linux, need to create the window and gfx platform here inorder
    // to create the OpenGL rendering context before creating the decoder.
#ifndef _WIN32
    if (!m_sdl.InitWindow("CloudXR Desktop View", 1, 1, false))
    {
        return cxrError_Client_Setup_Failed;
    }

    if (!m_platform.Create(m_sdl.GetDpy(), m_sdl.GetWin(),
               0, 0, 0))
    {
        return cxrError_Client_Setup_Failed;
    }
#endif

    m_platform.MakeCurrent();

    m_hmd = VR_Init(&eError, VRApplication_Scene);

    if (eError != VRInitError_None)
    {
        if (eError == VRInitError_Init_HmdNotFound)
            CXR_LOGE("Error calling VR_Init, no HMD attached.");
        else
            CXR_LOGE("Error calling VR_Init: %d.", eError);
        return cxrError_Client_Setup_Failed;
    }

    int32_t adapterIndex = 0;
    m_hmd->GetDXGIOutputInfo(&adapterIndex);
    CXR_LOGI("    Adapter index is %d", adapterIndex);

    float fps = 0.0f;
    uint32_t width = 0;
    uint32_t height = 0;
    if (vr::IVRExtendedDisplay *d = vr::VRExtendedDisplay())
    {
        int32_t x, y;
        d->GetWindowBounds(&x, &y, &width, &height);
    }

    fps = m_hmd->GetFloatTrackedDeviceProperty(k_unTrackedDeviceIndex_Hmd, Prop_DisplayFrequency_Float);
    CXR_LOGI("  Display: %u x %u @ %.3f", width, height, fps);

    m_deviceDesc.numVideoStreamDescs = CXR_NUM_VIDEO_STREAMS_XR;
    for (uint32_t i = 0; i < m_deviceDesc.numVideoStreamDescs; i++)
    {
        m_deviceDesc.videoStreamDescs[i].format = cxrClientSurfaceFormat_RGB;
        m_deviceDesc.videoStreamDescs[i].width = width / 2;
        m_deviceDesc.videoStreamDescs[i].height = height;
        m_deviceDesc.videoStreamDescs[i].fps = std::min(CXR_MAX_VIDEO_STREAM_FPS, fps);
        m_deviceDesc.videoStreamDescs[i].maxBitrate = options.mMaxVideoBitrate;
    }
    m_deviceDesc.stereoDisplay = true;

    m_deviceDesc.maxResFactor = options.mMaxResFactor;
    m_deviceDesc.predOffset = 0.008f;

    m_deviceDesc.ipd = m_hmd->GetFloatTrackedDeviceProperty(k_unTrackedDeviceIndex_Hmd, Prop_UserIpdMeters_Float);
    CXR_LOGI("  IPD: %.2f", m_deviceDesc.ipd);

    for (int i = 0; i < 2; i++)
    {
        m_hmd->GetProjectionRaw((vr::EVREye)i, &m_deviceDesc.proj[i][0], &m_deviceDesc.proj[i][1], &m_deviceDesc.proj[i][2], &m_deviceDesc.proj[i][3]);
        CXR_LOGI("  Projection %s:  l=%+f, r=%+f, t=%+f, b=%+f", i ? "Right" : "Left", m_deviceDesc.proj[i][0], m_deviceDesc.proj[i][1], m_deviceDesc.proj[i][2], m_deviceDesc.proj[i][3]);
    }

    bool receiveAudio = options.mReceiveAudio;
    bool sendAudio = options.mSendAudio;
    AudioRecordingCallback audioRecordingCallback = [](void* userdata, uint8_t* stream, int len) {
        reinterpret_cast<ClientCommon*>(userdata)->SendAudio(stream, len);
    };
    m_sdl.InitAudio(CXR_AUDIO_SAMPLING_RATE, CXR_AUDIO_CHANNEL_COUNT, CXR_AUDIO_FRAME_LENGTH_MS, 
        receiveAudio, sendAudio, audioRecordingCallback, this);

    m_deviceDesc.receiveAudio = receiveAudio;
    m_deviceDesc.sendAudio = sendAudio;
    m_deviceDesc.embedInfoInVideo = false;
#if defined(_WIN32) || defined(WITH_CUDA)
    m_deviceDesc.foveatedScaleFactor = options.mFoveation;
#else
    m_deviceDesc.foveatedScaleFactor = 0;
#endif
    m_deviceDesc.foveationModeCaps = 0;
    if (m_deviceDesc.foveatedScaleFactor)
    {
        CXR_LOGI("  Foveation enabled (%d%%)", m_deviceDesc.foveatedScaleFactor);
    }
    m_deviceDesc.posePollFreq = 0;
    m_deviceDesc.disablePosePrediction = false;
    m_deviceDesc.angularVelocityInDeviceSpace = false;

    char ctrlName[128] = { 0 };
    vr::VRSystem()->GetStringTrackedDeviceProperty(vr::k_unTrackedDeviceIndex_Hmd, vr::Prop_ModelNumber_String, ctrlName, sizeof(ctrlName), NULL);

    CXR_LOGI("  Controller name: %s\n", ctrlName);
    cxrControllerType physicalControllerType = cxrControllerType_None;
    if (!strcmp(ctrlName, "Index"))
    {
        physicalControllerType = cxrControllerType_ValveIndex;
    }
    else if (!strcmp(ctrlName, "Oculus Touch")) // really to support questlink cabled testing.
    {
        physicalControllerType = cxrControllerType_OculusTouch;
    }
    else
    {
        physicalControllerType = cxrControllerType_HtcVive;
    }

    EVRInputError retErr = VRInputError_None;
    TCHAR path[MAX_PATH * 2] = "\0";
#ifdef _WIN32
    if (GetModuleFileName(NULL, path, MAX_PATH))
        strcpy_s(strrchr(path, '\\'), MAX_PATH, "\\cxr_controller_actions.json");
    else
        CXR_LOGE("Getting the module file path failed.");
#else
    if( readlink( "/proc/self/exe", path, MAX_PATH ) )
        strcpy_s(strrchr(path, '/'), "/vive_controller_actions.json");
    else
        CXR_LOGE("Getting proc/self/exe file path failed.");
#endif
    CXR_LOGI( "Got base exe path as: %s\n", path );

    //set up the manifest first wiki https://github.com/ValveSoftware/openvr/wiki/SteamVR-Input
    retErr = vr::VRInput()->SetActionManifestPath(path);
    if (retErr == vr::VRInputError_None)
        CXR_LOGI("Loaded json manifest: %s, %d", path, retErr);
    else
        CXR_LOGE("Loaded json manifest failed: %s, %d", path, retErr);

    //set up the action set
    retErr = vr::VRInput()->GetActionSetHandle("/actions/cxr", &m_actionset);
    if (retErr == VRInputError_None)
        CXR_LOGI("Actionset handle: %d, Ret:%d", (int)m_actionset, (int)retErr);
    else
        CXR_LOGE("Actionset handle failed: %d, Ret:%d", (int)m_actionset, (int)retErr);

    m_controllers[0].Init(TrackedControllerRole_LeftHand, physicalControllerType);
    m_controllers[1].Init(TrackedControllerRole_RightHand, physicalControllerType);

    m_universeOrigin = VRCompositor()->GetTrackingSpace();

    VRCompositor()->WaitGetPoses(NULL, 0, NULL, 0);

    GetChaperone(&m_deviceDesc.chaperone);

    cxrError cError = ClientCommon::Init(options, outputPath);
    if (cError != cxrError_Success)
        return cError;

    m_init = true;
    return cxrError_Success;
}

void VisionClientOpenVR::GetTrackingState(cxrVRTrackingState *trackingState)
{
    memset(trackingState, 0, sizeof(cxrVRTrackingState));

    if (!m_init)
    {
        return;
    }

    const int NPOSES = k_unMaxTrackedDeviceCount;
    TrackedDevicePose_t renderPose[NPOSES];
    //VRCompositor()->WaitGetPoses(renderPose, k_unMaxTrackedDeviceCount, NULL, 0);

    VRSystem()->GetDeviceToAbsoluteTrackingPose(m_universeOrigin, 0, renderPose, NPOSES);
    cxrMatrix34 poseMatrix = {};
    memcpy(&poseMatrix, &(renderPose[0].mDeviceToAbsoluteTracking), sizeof(cxrMatrix34));
    cxrMatrixToVecQuat(&poseMatrix, &(trackingState->hmd.pose.position), &(trackingState->hmd.pose.rotation));
    memcpy(trackingState->hmd.pose.velocity.v, renderPose[0].vVelocity.v, sizeof(float) * 3);
    memcpy(trackingState->hmd.pose.angularVelocity.v, renderPose[0].vAngularVelocity.v, sizeof(float) * 3);
    //Enum values for cxrTrackingResult are the same as ETrackingResult
    trackingState->hmd.pose.trackingResult = (cxrTrackingResult) renderPose[0].eTrackingResult;
    trackingState->hmd.pose.poseIsValid = renderPose[0].bPoseIsValid ? cxrTrue : cxrFalse;
    trackingState->hmd.pose.deviceIsConnected = renderPose[0].bDeviceIsConnected ? cxrTrue : cxrFalse;
    trackingState->hmd.activityLevel = (cxrDeviceActivityLevel)VRSystem()->GetTrackedDeviceActivityLevel(vr::k_unTrackedDeviceIndex_Hmd);

    vr::VRActiveActionSet_t actionSet = { 0 };
    actionSet.ulActionSet = m_actionset;
    EVRInputError retErr = vr::VRInput()->UpdateActionState(&actionSet, sizeof(actionSet), 1);
    if (retErr != VRInputError_None)
    {
        CXR_LOGE("Unable to update active ActionSet state, error = 0x%0x", retErr);
        // TODO: should we early return here?  Possible to finish the fn w/o actionset?
    }

    for (uint32_t device = 1; device < k_unMaxTrackedDeviceCount; device++)
    {
        if (m_hmd->GetTrackedDeviceClass(device) != TrackedDeviceClass_Controller)
        {
            continue;
        }

        //role API deprecated but works so TBD in future
        ETrackedControllerRole role = m_hmd->GetControllerRoleForTrackedDeviceIndex(device);

        if ((role == TrackedControllerRole_LeftHand) || (role == TrackedControllerRole_RightHand))
        {
            int index = (role == TrackedControllerRole_LeftHand) ? 0 : 1;

            trackingState->controller[index].pose.trackingResult = (cxrTrackingResult) renderPose[device].eTrackingResult;
            trackingState->controller[index].pose.deviceIsConnected = renderPose[device].bDeviceIsConnected ? cxrTrue : cxrFalse;
            trackingState->controller[index].pose.poseIsValid = renderPose[device].bPoseIsValid  ? cxrTrue : cxrFalse;

            if (!renderPose[device].bDeviceIsConnected || !renderPose[device].bPoseIsValid)
            {
                continue;
            }

            memcpy(&poseMatrix, &(renderPose[device].mDeviceToAbsoluteTracking), sizeof(cxrMatrix34));
            cxrMatrixToVecQuat(&poseMatrix, &(trackingState->controller[index].pose.position), &(trackingState->controller[index].pose.rotation));
            memcpy(trackingState->controller[index].pose.velocity.v, renderPose[device].vVelocity.v, sizeof(float) * 3);
            memcpy(trackingState->controller[index].pose.angularVelocity.v, renderPose[device].vAngularVelocity.v, sizeof(float) * 3);

            cxrControllerTrackingState controllerState = trackingState->controller[index];
            if (!m_newControllers[index])
            {
                cxrControllerType type = m_controllers[index].m_type;
                cxrControllerDesc desc = {};
                desc.id = index; // TBD whether there is a UUID we should put here.
                desc.role = index ? "cxr://input/hand/right" : "cxr://input/hand/left";
                desc.controllerName = controllerNames[int(type)];
                if(type == cxrControllerType_ValveIndex)
                {
                    desc.inputCount = inputKnucklesLegacyCount;
                    desc.inputPaths = inputsKnucklesLegacy;
                    desc.inputValueTypes = inputValuesKnucklesLegacy;
                }
                else
                if(type == cxrControllerType_HtcVive)
                {
                    desc.inputCount = inputWandLegacyCount;
                    desc.inputPaths = inputsWandLegacy;
                    desc.inputValueTypes = inputValuesWandLegacy;
                }
                else // assume oculus, or oculus-type controller.
                {
                    desc.inputCount = inputTouchLegacyCount;
                    desc.inputPaths = inputsTouchLegacy;
                    desc.inputValueTypes = inputValuesTouchLegacy;
                }

                cxrError e = cxrAddController(m_receiver, &desc, &m_newControllers[index]);
                if (e != cxrError_Success)
                    CXR_LOGE("error: could not add controller.");
            }

            m_controllers[index].FireActions(m_receiver, m_newControllers[index], m_remapSysBtn);
        }
    }
}

void VisionClientOpenVR::TriggerHaptic(const cxrHapticFeedback *haptic)
{
    // Should really use map, ID->controller, and not assume
    // ID==0|1, then call trigger on that controller.
    // But for now, we do essentially hard-code OpenVR to 0+1.
    m_controllers[haptic->deviceID].TriggerHaptic(haptic);
}

cxrBool VisionClientOpenVR::RenderAudio(const cxrAudioFrame *audioFrame)
{
    m_sdl.QueueAudio(audioFrame->streamBuffer, audioFrame->streamSizeBytes);
    return cxrTrue;
}

void VisionClientOpenVR::GetChaperone(cxrChaperone *chaperone)
{
    chaperone->universe = (cxrUniverseOrigin)m_universeOrigin;

    IVRChaperoneSetup *setup = VRChaperoneSetup();

    chaperone->origin.m[0][0] = chaperone->origin.m[1][1] = chaperone->origin.m[2][2] = 1;
    chaperone->origin.m[0][1] = chaperone->origin.m[0][2] = chaperone->origin.m[0][3] = 0;
    chaperone->origin.m[1][0] = chaperone->origin.m[1][2] = chaperone->origin.m[1][3] = 0;
    chaperone->origin.m[2][0] = chaperone->origin.m[2][1] = chaperone->origin.m[2][3] = 0;

    setup->GetWorkingPlayAreaSize(&chaperone->playArea.v[0], &chaperone->playArea.v[1]);
}

void VisionClientOpenVR::GetGraphicsContext(cxrGraphicsContext *context, cxrGraphicsContextType type)
{
    m_platform.GetGraphicsContext(context, type);
}

cxrError VisionClientOpenVR::SubmitFrame(cxrFramesLatched* framesLatched)
{
    cxrError e = cxrError_Success;
    m_platform.MakeCurrent();

    VRTextureWithPose_t eyeVRTex;
    eyeVRTex.eColorSpace = ColorSpace_Gamma;
#ifdef _WIN32
    eyeVRTex.eType = TextureType_DirectX;
#else
    eyeVRTex.eType = TextureType_OpenGL;
#endif
    memcpy(&eyeVRTex.mDeviceToAbsoluteTracking, &framesLatched->poseMatrix, sizeof(cxrMatrix34));

    for (int i = 0; i < 2; i++)
    {
        eyeVRTex.handle = (void*)(uintptr_t)framesLatched->frames[i].texture;
        EVRCompositorError r = VRCompositorError_None;
        r = VRCompositor()->Submit((EVREye)i, &eyeVRTex, NULL, Submit_TextureWithPose );
        if (r != VRCompositorError_None)
        {
            CXR_LOGE("Error submitting frame, VRCompositor returned %d.", r);
            e = cxrError_Failed;
        }
    }

    VRCompositor()->PostPresentHandoff();

    return e;
}

bool VisionClientOpenVR::Update()
{
    vr::VREvent_t ev;
    while (m_hmd->PollNextEvent(&ev, sizeof(ev)))
    {
        switch (ev.eventType)
        {
        case vr::EVREventType::VREvent_Quit:
            m_hmd->AcknowledgeQuit_Exiting();
            CXR_LOGE("Received a quit message. Exiting...");
            return false;
        }
    }

    VRCompositor()->WaitGetPoses(NULL, 0, NULL, 0);

    return true;
}



void VisionController::Init(ETrackedControllerRole role, cxrControllerType type)
{
    m_role = role;
    m_type = type;

    EVRInputError retErr = VRInputError_None;
    memset(m_booleanInputs, 0, sizeof(m_booleanInputs));

    auto vrInput = vr::VRInput();

    if (m_role == TrackedControllerRole_RightHand)
    {
        /* Create action handle for each control in line with the IVRDriverInput

        These are organized as follows:

            Common inputs

            if (Trackpad controllers)
                Trackpad inputs

            if (Grip & Joystick controllers)
                Grip inputs
                Joystick inputs

            if (Quest)
                Quest only inputs

            and so on...

        */

        retErr = vrInput->GetActionHandle("/actions/cxr/in/right_system_press", &m_booleanInputs[cxrButton_System]);
        if (retErr != VRInputError_None)
        {
            CXR_LOGE("Failed to get an action handle: %d.", retErr);
        }

        retErr = vrInput->GetActionHandle("/actions/cxr/in/right_trigger_click", &m_booleanInputs[cxrButton_Trigger_Click]);
        retErr = vrInput->GetActionHandle("/actions/cxr/in/right_trigger_touch", &m_booleanInputs[cxrButton_Trigger_Touch]);
        retErr = vrInput->GetActionHandle("/actions/cxr/in/right_trigger_value", &m_analogInputs[cxrAnalog_Trigger]);

        if (type == cxrControllerType_HtcVive || type == cxrControllerType_ValveIndex)
        {
            retErr = vrInput->GetActionHandle("/actions/cxr/in/right_grip_press", &m_booleanInputs[cxrButton_Grip_Click]);

            retErr = vrInput->GetActionHandle("/actions/cxr/in/right_axis0_press", &m_booleanInputs[cxrButton_Touchpad_Click]);
            retErr = vrInput->GetActionHandle("/actions/cxr/in/right_axis0_touch", &m_booleanInputs[cxrButton_Touchpad_Touch]);
            retErr = vrInput->GetActionHandle("/actions/cxr/in/right_axis0_value", &m_analogInputs[cxrAnalog_TouchpadX]); // both x and y
        }

        if (type == cxrControllerType_OculusTouch || type == cxrControllerType_ValveIndex)
        {
            retErr = vrInput->GetActionHandle("/actions/cxr/in/right_grip_value", &m_analogInputs[cxrAnalog_Grip]);
            retErr = vrInput->GetActionHandle("/actions/cxr/in/right_grip_touch", &m_booleanInputs[cxrButton_Grip_Touch]);

            retErr = vrInput->GetActionHandle("/actions/cxr/in/right_joystick_press", &m_booleanInputs[cxrButton_Joystick_Click]);
            retErr = vrInput->GetActionHandle("/actions/cxr/in/right_joystick_touch", &m_booleanInputs[cxrButton_Joystick_Touch]);
            retErr = vrInput->GetActionHandle("/actions/cxr/in/right_joystick_value", &m_analogInputs[cxrAnalog_JoystickX]); // both x and y

            retErr = vrInput->GetActionHandle("/actions/cxr/in/right_a_press", &m_booleanInputs[cxrButton_A]);
            retErr = vrInput->GetActionHandle("/actions/cxr/in/right_b_press", &m_booleanInputs[cxrButton_B]);
        }

        if (type == cxrControllerType_HtcVive)
        {
            retErr = vrInput->GetActionHandle("/actions/cxr/in/right_applicationmenu_press", &m_booleanInputs[cxrButton_ApplicationMenu]);
        }
        else if (type == cxrControllerType_ValveIndex)
        {
            retErr = vrInput->GetActionHandle("/actions/cxr/in/right_grip_force", &m_analogInputs[cxrAnalog_Grip_Force]);
        }

        retErr = vrInput->GetActionHandle("/actions/cxr/out/right_haptic", &m_haptic);
    }
    else
    {
        retErr = vrInput->GetActionHandle("/actions/cxr/in/left_system_press", &m_booleanInputs[cxrButton_System]);
        if (retErr != VRInputError_None)
        {
            CXR_LOGE("Failed to get an action handle: %d.", retErr);
        }

        retErr = vrInput->GetActionHandle("/actions/cxr/in/left_trigger_click", &m_booleanInputs[cxrButton_Trigger_Click]);
        retErr = vrInput->GetActionHandle("/actions/cxr/in/left_trigger_touch", &m_booleanInputs[cxrButton_Trigger_Touch]);
        retErr = vrInput->GetActionHandle("/actions/cxr/in/left_trigger_value", &m_analogInputs[cxrAnalog_Trigger]);

        if (type == cxrControllerType_HtcVive || type == cxrControllerType_ValveIndex)
        {
            retErr = vrInput->GetActionHandle("/actions/cxr/in/left_grip_press", &m_booleanInputs[cxrButton_Grip_Click]);

            retErr = vrInput->GetActionHandle("/actions/cxr/in/left_axis0_press", &m_booleanInputs[cxrButton_Touchpad_Click]);
            retErr = vrInput->GetActionHandle("/actions/cxr/in/left_axis0_touch", &m_booleanInputs[cxrButton_Touchpad_Touch]);
            retErr = vrInput->GetActionHandle("/actions/cxr/in/left_axis0_value", &m_analogInputs[cxrAnalog_TouchpadX]); // both x and y
        }

        if (type == cxrControllerType_OculusTouch || type == cxrControllerType_ValveIndex)
        {
            retErr = vrInput->GetActionHandle("/actions/cxr/in/left_grip_value", &m_analogInputs[cxrAnalog_Grip]);
            retErr = vrInput->GetActionHandle("/actions/cxr/in/left_grip_touch", &m_booleanInputs[cxrButton_Grip_Touch]);

            retErr = vrInput->GetActionHandle("/actions/cxr/in/left_joystick_press", &m_booleanInputs[cxrButton_Joystick_Click]);
            retErr = vrInput->GetActionHandle("/actions/cxr/in/left_joystick_touch", &m_booleanInputs[cxrButton_Joystick_Touch]);
            retErr = vrInput->GetActionHandle("/actions/cxr/in/left_joystick_value", &m_analogInputs[cxrAnalog_JoystickX]); // both x and y
        }

        if (type == cxrControllerType_HtcVive)
        {
            retErr = vrInput->GetActionHandle("/actions/cxr/in/left_applicationmenu_press", &m_booleanInputs[cxrButton_ApplicationMenu]);
        }
        else if (type == cxrControllerType_OculusTouch)
        {
            retErr = vrInput->GetActionHandle("/actions/cxr/in/left_a_press", &m_booleanInputs[cxrButton_X]);
            retErr = vrInput->GetActionHandle("/actions/cxr/in/left_b_press", &m_booleanInputs[cxrButton_Y]);
        }
        else if (type == cxrControllerType_ValveIndex)
        {
            retErr = vrInput->GetActionHandle("/actions/cxr/in/left_grip_force", &m_analogInputs[cxrAnalog_Grip_Force]);

            retErr = vrInput->GetActionHandle("/actions/cxr/in/left_a_press", &m_booleanInputs[cxrButton_A]);
            retErr = vrInput->GetActionHandle("/actions/cxr/in/left_b_press", &m_booleanInputs[cxrButton_B]);
        }

        retErr = vrInput->GetActionHandle("/actions/cxr/out/left_haptic", &m_haptic);
    }
}

void VisionController::FireActions(cxrReceiverHandle m_receiver, cxrControllerHandle handle, bool remapSysBtn)
{
    cxrControllerEvent events[28] = {}; //[_countof(actionsGeneric)]
    int nextEvent = 0;
    
    auto vrInput = vr::VRInput();

    auto cxrButtonToKnucklesIndex = [](cxrButtonId id) -> int
    {
        switch(id)
        {
            #define match(x, y) case cxrButton_##x: return y
            match(System, 0);
            match(ApplicationMenu, 9);
            match(Grip_Touch, 10);
            match(Grip_Click, 11);
            match(Trigger_Touch, 2);
            match(Trigger_Click, 1);
            match(Touchpad_Touch, 6);
            match(Touchpad_Click, 5);
            match(Joystick_Touch, 14);
            match(Joystick_Click, 13);
            match(A, 17);
            match(B, 18);
            #undef match
        }
        return -1;
    };
    auto cxrButtonToWandIndex = [](cxrButtonId id) -> int
    {
        switch(id)
        {
            #define match(x, y) case cxrButton_##x: return y
            match(System, 0);
            match(ApplicationMenu, 9);
            match(Grip_Click, 4);
            match(Trigger_Touch, 2);
            match(Trigger_Click, 1);
            match(Touchpad_Touch, 6);
            match(Touchpad_Click, 5);
            #undef match
        }
        return -1;
    };
    auto cxrButtonToTouchIndex = [](cxrButtonId id) -> int
    {
        switch (id)
        {
#define match(x, y) case cxrButton_##x: return y
            match(System, 0);
            match(ApplicationMenu, 1);
            match(Grip_Touch, 6);
            match(Grip_Click, 5);
            match(Trigger_Touch, 3);
            match(Trigger_Click, 2);
            match(Joystick_Touch, 9);
            match(Joystick_Click, 8);
            match(A, 12);
            match(B, 13);
            match(X, 12);
            match(Y, 13);
            // we didn't have touch for abxy/thumb in legacy.
#undef match
        }
        return -1;
    };

    bool MenuButtonChanged = false;
    bool MenuButtonValue = false;
    // Button input
    for (int i = 0; i < cxrButton_Num; i++)
    {
        if (m_booleanInputs[(cxrButtonId)i] == 0)
            continue;

        vr::InputDigitalActionData_t actionData = { 0 };
        vrInput->GetDigitalActionData(m_booleanInputs[i], &actionData, sizeof(actionData), vr::k_ulInvalidInputValueHandle);

        if (actionData.bActive)
        {
            if (actionData.bChanged)
            {
                if((cxrButtonId)i == cxrButton_ApplicationMenu)
                {
                    MenuButtonChanged = true;
                    MenuButtonValue = actionData.bState;
                }
                auto& event = events[nextEvent++];
                if(m_type == cxrControllerType_ValveIndex)
                {
                    event.clientInputIndex = cxrButtonToKnucklesIndex((cxrButtonId)i);
                }
                if(m_type == cxrControllerType_HtcVive)
                {
                    event.clientInputIndex = cxrButtonToWandIndex((cxrButtonId)i);
                }
                else // oculus fallbak
                {
                    event.clientInputIndex = cxrButtonToTouchIndex((cxrButtonId)i);
                }
                event.inputValue.valueType = cxrInputValueType_boolean;
                event.inputValue.vBool = actionData.bState;
            }
        }
    }

    // remap left hand app menu button to system
    if (remapSysBtn && m_role == TrackedControllerRole_LeftHand)
    {
        if(MenuButtonChanged)
        {
            auto& event = events[nextEvent++];
            event.clientInputIndex = cxrButtonToKnucklesIndex(cxrButton_ApplicationMenu);
            event.inputValue.valueType = cxrInputValueType_boolean;
            event.inputValue.vBool = MenuButtonValue;
        }
    }

    // Analog input according to controller capability
    auto cxrAnalogToKnucklesIndex = [](cxrAnalogId id) -> int
    {
        switch(id)
        {
            #define match(x, y) case cxrAnalog_##x: return y
            match(Trigger, 3);
            match(TouchpadX, 7);
            match(TouchpadY, 8);
            match(JoystickX, 15);
            match(JoystickY, 16);
            match(Grip, 11);
            match(Grip_Force, 12);
            #undef match
        }
        return -1;
    };

    auto cxrAnalogToWandIndex = [](cxrAnalogId id) -> int
    {
        switch(id)
        {
            #define match(x, y) case cxrAnalog_##x: return y
            match(Trigger, 3);
            match(TouchpadX, 7);
            match(TouchpadY, 8);
            #undef match
        }
        return -1;
    };
    vr::InputAnalogActionData_t analogData = { 0 };
    if(m_type == cxrControllerType_ValveIndex)
    {
        if (vrInput->GetAnalogActionData(m_analogInputs[cxrAnalog_Trigger], &analogData, sizeof(analogData), vr::k_ulInvalidInputValueHandle) == vr::VRInputError_None && analogData.bActive && (analogData.deltaX != 0.f))
        {
            auto& event = events[nextEvent++];
            event.clientInputIndex = cxrAnalogToKnucklesIndex(cxrAnalog_Trigger);
            event.inputValue.valueType = cxrInputValueType_float32;
            event.inputValue.vF32 = analogData.x;
        }

        if (vrInput->GetAnalogActionData(m_analogInputs[cxrAnalog_TouchpadX], &analogData, sizeof(analogData), vr::k_ulInvalidInputValueHandle) == vr::VRInputError_None && analogData.bActive 
            && ((analogData.deltaX != 0.f) || (analogData.deltaY != 0.f)))
        {
            {
                auto& event = events[nextEvent++];
                event.clientInputIndex = cxrAnalogToKnucklesIndex(cxrAnalog_TouchpadX);
                event.inputValue.valueType = cxrInputValueType_float32;
                event.inputValue.vF32 = analogData.x;
            }
            {
                auto& event = events[nextEvent++];
                event.clientInputIndex = cxrAnalogToKnucklesIndex(cxrAnalog_TouchpadY);
                event.inputValue.valueType = cxrInputValueType_float32;
                event.inputValue.vF32 = analogData.y;
            }
        }

        if (vrInput->GetAnalogActionData(m_analogInputs[cxrAnalog_JoystickX], &analogData, sizeof(analogData), vr::k_ulInvalidInputValueHandle) == vr::VRInputError_None && analogData.bActive 
            && ((analogData.deltaX != 0.f) || (analogData.deltaY != 0.f)))
        {
            {
                auto& event = events[nextEvent++];
                event.clientInputIndex = cxrAnalogToKnucklesIndex(cxrAnalog_JoystickX);
                event.inputValue.valueType = cxrInputValueType_float32;
                event.inputValue.vF32 = analogData.x;
            }
            {
                auto& event = events[nextEvent++];
                event.clientInputIndex = cxrAnalogToKnucklesIndex(cxrAnalog_JoystickY);
                event.inputValue.valueType = cxrInputValueType_float32;
                event.inputValue.vF32 = analogData.y;
            }
        }

        if (vrInput->GetAnalogActionData(m_analogInputs[cxrAnalog_Grip], &analogData, sizeof(analogData), vr::k_ulInvalidInputValueHandle) == vr::VRInputError_None && analogData.bActive && (analogData.deltaX != 0.f))
        {
            auto& event = events[nextEvent++];
            event.clientInputIndex = cxrAnalogToKnucklesIndex(cxrAnalog_Grip);
            event.inputValue.valueType = cxrInputValueType_float32;
            event.inputValue.vF32 = analogData.x;
        }

        if (vrInput->GetAnalogActionData(m_analogInputs[cxrAnalog_Grip_Force], &analogData, sizeof(analogData), vr::k_ulInvalidInputValueHandle) == vr::VRInputError_None && analogData.bActive && (analogData.deltaX != 0.f))
        {
            auto& event = events[nextEvent++];
            event.clientInputIndex = cxrAnalogToKnucklesIndex(cxrAnalog_Grip_Force);
            event.inputValue.valueType = cxrInputValueType_float32;
            event.inputValue.vF32 = analogData.x;
        }
    }

    if(m_type == cxrControllerType_HtcVive)
    {
        #if 1
        if (vrInput->GetAnalogActionData(m_analogInputs[cxrAnalog_Trigger], &analogData, sizeof(analogData), vr::k_ulInvalidInputValueHandle) == vr::VRInputError_None && analogData.bActive && (analogData.deltaX != 0.f))
        {
            auto& event = events[nextEvent++];
            event.clientInputIndex = cxrAnalogToWandIndex(cxrAnalog_Trigger);
            event.inputValue.valueType = cxrInputValueType_float32;
            event.inputValue.vF32 = analogData.x;
            //CXR_LOGE("Sending modern event trigger action idx %d value %f", int(event.clientInputIndex), analogData.x);
        }
        #endif
        #if 1
        if (vrInput->GetAnalogActionData(m_analogInputs[cxrAnalog_TouchpadX], &analogData, sizeof(analogData), vr::k_ulInvalidInputValueHandle) == vr::VRInputError_None && analogData.bActive 
            && ((analogData.deltaX != 0.f) || (analogData.deltaY != 0.f)))
        {
            {
                auto& event = events[nextEvent++];
                event.clientInputIndex = cxrAnalogToWandIndex(cxrAnalog_TouchpadX);
                event.inputValue.valueType = cxrInputValueType_float32;
                event.inputValue.vF32 = analogData.x;
                //CXR_LOGE("Sending modern event touchpad X action idx %d value %f", int(event.clientInputIndex), analogData.x);
            }
            {
                auto& event = events[nextEvent++];
                event.clientInputIndex = cxrAnalogToWandIndex(cxrAnalog_TouchpadY);
                event.inputValue.valueType = cxrInputValueType_float32;
                event.inputValue.vF32 = analogData.y;
                //CXR_LOGE("Sending modern event touchpad Y action idx %d value %f", int(event.clientInputIndex), analogData.x);
            }
        }
        #endif
    }
    
    cxrError e = cxrFireControllerEvents(m_receiver, handle, events, nextEvent);
    assert(e == cxrError_Success);
}

void VisionController::TriggerHaptic(const cxrHapticFeedback *haptic)
{
    //CXR_LOGV("Triggering haptic %f %f %f", haptic.seconds, haptic.frequency, haptic.amplitude);

    vr::VRInput()->TriggerHapticVibrationAction(m_haptic, 0, haptic->seconds, haptic->frequency, haptic->amplitude, vr::k_ulInvalidInputValueHandle);
}
