/*
 * Copyright (c) 2019-2020, NVIDIA CORPORATION. All rights reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

#include "ClientCommon.h"

#include <algorithm>
#include "ClientWindow.h"

namespace
{   // switched to emulating Touch controller as there is PC+mobile variants, and has more inputs.
    static constexpr int inputTouchLegacyCount = 21;

    static const char* inputsTouchLegacy[inputTouchLegacyCount] =
    {
        "/input/system/click",
        "/input/application_menu/click", // this is carried over from old system and might be remove, it's not a button binding, more action.
        "/input/trigger/click",
        "/input/trigger/touch",
        "/input/trigger/value",
        "/input/grip/click",
        "/input/grip/touch",
        "/input/grip/value",
        "/input/thumbstick/click",
        "/input/thumbstick/touch",
        "/input/thumbstick/x",
        "/input/thumbstick/y",
        "/input/a/click",
        "/input/b/click",
        "/input/x/click", // Touch has X/Y on L controller, so we'll map the raw strings.
        "/input/y/click",
        "/input/a/touch",
        "/input/b/touch",
        "/input/x/touch",
        "/input/y/touch",
        "/input/thumb_rest/touch",
    };

    cxrInputValueType inputValuesTouchLegacy[inputTouchLegacyCount] =
    {
        cxrInputValueType_boolean, //input/system/click
        cxrInputValueType_boolean, //input/application_menu/click
        cxrInputValueType_boolean, //input/trigger/click
        cxrInputValueType_boolean, //input/trigger/touch
        cxrInputValueType_float32, //input/trigger/value
        cxrInputValueType_boolean, //input/grip/click
        cxrInputValueType_boolean, //input/grip/touch
        cxrInputValueType_float32, //input/grip/value
        cxrInputValueType_boolean, //input/thumbstick/click
        cxrInputValueType_boolean, //input/thumbstick/touch
        cxrInputValueType_float32, //input/thumbstick/x
        cxrInputValueType_float32, //input/thumbstick/y
        cxrInputValueType_boolean, //input/a/click
        cxrInputValueType_boolean, //input/b/click
        cxrInputValueType_boolean, //input/x/click
        cxrInputValueType_boolean, //input/y/click
        cxrInputValueType_boolean, //input/a/touch
        cxrInputValueType_boolean, //input/b/touch
        cxrInputValueType_boolean, //input/x/touch
        cxrInputValueType_boolean, //input/y/touch
        cxrInputValueType_boolean, //input/thumb_rest/touch
    };
}

cxrMatrix34 HmdMatrix34Identity()
{
    cxrMatrix34 m;
    memset(&m, 0, sizeof(m));
    m.m[0][0] = 1.0f;
    m.m[1][1] = 1.0f;
    m.m[2][2] = 1.0f;
    return m;
}

VisionClientWindow::VisionClientWindow(cxrGraphicsContextType type) : ClientCommon(),
    m_platform(type)
{
    m_width = 0;
    m_height = 0;
    m_frameCounter = 0;
    m_buttonDown = 0;
}

VisionClientWindow::~VisionClientWindow()
{
    m_platform.Destroy();
}

cxrError VisionClientWindow::Init(const CloudXR::ClientOptions& options, const std::string &outputPath)
{
    m_clientOptions = options; // cache for Connect to use some options later

    if (!m_sdl.InitVideo())
    {
        CXR_LOGE("Fatal error: could not initialize SDL video support.");
        return cxrError_Client_Setup_Failed;
    }

    // On Linux, need to create the window and gfx platform here inorder
    // to create the OpenGL rendering context before creating the decoder.
#ifndef _WIN32  
    m_width = 640;
    m_height = 480;
    if (!m_sdl.InitWindow("CloudXR Desktop View", m_width, m_height, true))
    {
        CXR_LOGE("Fatal error: could not initialize SDL window.");
        return cxrError_Client_Setup_Failed;
    }

    if (!m_platform.Create(m_sdl.GetDpy(), m_sdl.GetWin(),
          NULL, m_width, m_height))
    {
        CXR_LOGE("Fatal error: could not initialize graphics platform support.");
        return cxrError_Client_Setup_Failed;
    }
#endif

    bool receiveAudio = m_clientOptions.mReceiveAudio;
    bool sendAudio = m_clientOptions.mSendAudio;
    AudioRecordingCallback audioRecordingCallback = [](void* userdata, uint8_t* stream, int len) {
        reinterpret_cast<ClientCommon*>(userdata)->SendAudio(stream, len);
    };

    // TODO no error check here???
    m_sdl.InitAudio(CXR_AUDIO_SAMPLING_RATE, CXR_AUDIO_CHANNEL_COUNT, CXR_AUDIO_FRAME_LENGTH_MS, 
        receiveAudio, sendAudio, audioRecordingCallback, this);

    uint32_t fps = 0;
    uint32_t width = 0;
    uint32_t height = 0;
    m_sdl.FillMonitorParams(width, height, fps);

    m_deviceDesc.numVideoStreamDescs = options.mNumVideoStreams;
    for (uint32_t i = 0; i < m_deviceDesc.numVideoStreamDescs; i++)
    {
        m_deviceDesc.videoStreamDescs[i].format = m_clientOptions.mEnableAlpha ? cxrClientSurfaceFormat_RGBA : cxrClientSurfaceFormat_RGB;
        m_deviceDesc.videoStreamDescs[i].width = width / 2;
        m_deviceDesc.videoStreamDescs[i].height = height;
        m_deviceDesc.videoStreamDescs[i].fps = std::min(CXR_MAX_VIDEO_STREAM_FPS, (float)fps);
        m_deviceDesc.videoStreamDescs[i].maxBitrate = options.mMaxVideoBitrate;
    }
    m_deviceDesc.stereoDisplay = options.mNumVideoStreams == 2;

    m_deviceDesc.maxResFactor = 2.0f; // Try to avoid resampling of the game surfaces
    m_deviceDesc.ipd = 0.06f;
    m_deviceDesc.predOffset = 0;
    m_deviceDesc.proj[0][0] = m_deviceDesc.proj[0][2] = m_deviceDesc.proj[1][0] = m_deviceDesc.proj[1][2] = -1;
    m_deviceDesc.proj[0][1] = m_deviceDesc.proj[0][3] = m_deviceDesc.proj[1][1] = m_deviceDesc.proj[1][3] = 1;
    m_deviceDesc.receiveAudio = receiveAudio;
    m_deviceDesc.sendAudio = sendAudio;
    m_deviceDesc.embedInfoInVideo = false;
    m_deviceDesc.foveatedScaleFactor = (m_clientOptions.mFoveation < 100) ? m_clientOptions.mFoveation : 0;
    m_deviceDesc.posePollFreq = 0;
    m_deviceDesc.disablePosePrediction = true;
    m_deviceDesc.angularVelocityInDeviceSpace = false;
    m_deviceDesc.foveationModeCaps = 0;
    GetChaperone(&m_deviceDesc.chaperone);

    return ClientCommon::Init(options, outputPath);
}

void VisionClientWindow::GetTrackingState(cxrVRTrackingState *trackingState)
{
    memset(trackingState, 0, sizeof(cxrVRTrackingState));

    if (!trackingState->hmd.pose.poseIsValid)
    {
        trackingState->hmd.pose.poseIsValid = cxrTrue;
        trackingState->hmd.pose.deviceIsConnected = cxrTrue;
        trackingState->hmd.pose.trackingResult = cxrTrackingResult_Running_OK;
        trackingState->hmd.activityLevel = cxrDeviceActivityLevel_UserInteraction;
        trackingState->hmd.pose.rotation = cxrQuaternion{ 1, 0, 0, 0 };
    }

    for (int i = 0; i < CXR_NUM_CONTROLLERS; i++)
    {
        auto &c = trackingState->controller[i];
        c.pose.poseIsValid = cxrTrue;
        c.pose.deviceIsConnected = cxrTrue;
        c.pose.trackingResult = cxrTrackingResult_Running_OK;
        c.pose.position = cxrVector3{ 0, 0, 0 };
        c.pose.rotation = cxrQuaternion{ 1, 0, 0, 0 };
    }

    float x;
    float y;
#if _WIN32
    if (GetAsyncKeyState(VK_SHIFT))
    {
        // Center if SHIFT is held
        x = 0;
        y = 0;
    }
    else
#endif
    {
        // Get cursor position relative to center of window
        int cx, cy;
        m_sdl.GetMouseState(&cx, &cy);
        const float SENSITIVITY = 0.03f;
        x = SENSITIVITY * (cx - (int)m_width / 2);
        y = -SENSITIVITY * (cy - (int)m_height / 2);
    }
    trackingState->hmd.pose.position = cxrVector3{ x, 1.8f, -y };

    // Emulate a controller
    auto &ctrlLeft = trackingState->controller[0];
    ctrlLeft.pose = trackingState->hmd.pose;
    ctrlLeft.pose.position.v[0] -= 0.2f;
    ctrlLeft.pose.position.v[1] -= 0.5f;
    ctrlLeft.pose.position.v[2] -= 1.0f;
    ctrlLeft.pose.poseIsValid = ctrlLeft.pose.deviceIsConnected = cxrTrue;

    auto& ctrlRight = trackingState->controller[1];
    ctrlRight.pose = trackingState->hmd.pose;
    ctrlRight.pose.position.v[0] += 0.2f;
    ctrlRight.pose.position.v[1] -= 0.5f;
    ctrlRight.pose.position.v[2] -= 1.0f;
    ctrlRight.pose.poseIsValid = ctrlRight.pose.deviceIsConnected = cxrTrue;

    for (int i = 0; i < CXR_NUM_CONTROLLERS; i++)
    {
        auto &c = trackingState->controller[i];
        if (!m_newControllers[i])
        {
            cxrControllerDesc desc = {};
            desc.id = i;
            desc.role = i ? "cxr://input/hand/right" : "cxr://input/hand/left";
            desc.controllerName = "Oculus Touch";
            desc.inputCount = inputTouchLegacyCount;
            desc.inputPaths = inputsTouchLegacy;
            desc.inputValueTypes = inputValuesTouchLegacy;
            cxrError e = cxrAddController(m_receiver, &desc, &m_newControllers[i]);
            if(e != cxrError_Success)
            {
                CXR_LOGE("error: could not add controller.");
            }
        }
    }

#if _WIN32
	#define TOTAL_INPUT_CHARS    11
    static bool buttonDown[TOTAL_INPUT_CHARS] = {};

    float triggerVal = m_triggerVal;
    for(int i = 0; i < 10; ++i)
    {
        char c = '0' + i;
        if((GetAsyncKeyState(c) & 0x8000) == 0x8000)
        {
            triggerVal = float(i) / 9.0f;
        }
    }

    // this is a cheapo implementation of mapping keyboard key to a given input index,
    // including support for which controller it is on, and support for two-sided float axis binding.
    const char inputChars[TOTAL_INPUT_CHARS] = {'A', 'B', 'X', 'Y', 'G', 'R', 'T', 'I', 'J', 'K', 'L'};
    const int inputIndex[TOTAL_INPUT_CHARS] = { 12, 13, 14, 15, 7, -10, 10, 11, -10, -11, 10 };
    const int conIndex[TOTAL_INPUT_CHARS] = { 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1};

    cxrControllerEvent events[CXR_NUM_CONTROLLERS][16] = {};
    uint32_t eventCount[CXR_NUM_CONTROLLERS] = {};

    for (int i = 0; i < TOTAL_INPUT_CHARS; i++)
    {
        const short keystate = GetAsyncKeyState(inputChars[i]);
        bool down = (keystate & 0x8000) == 0x8000;
        if (down != buttonDown[i])
        { // send event.
            cxrControllerEvent& e = events[conIndex[i]][eventCount[conIndex[i]]++];
            e.clientTimeNS = 1234;
            const int inin = abs(int(inputIndex[i]));
            e.clientInputIndex = inin;
            if (inputValuesTouchLegacy[inin] == cxrInputValueType_boolean)
            {
                e.inputValue.valueType = cxrInputValueType_boolean;
                e.inputValue.vBool = down ? cxrTrue : cxrFalse;
            }
            else // float
            {
                e.inputValue.valueType = cxrInputValueType_float32;
                e.inputValue.vF32 = down ? 0.5f : 0.0f;
                if (inputIndex[i]<0) e.inputValue.vF32 *= -1.0f; // handle 2d axes
            }
            fprintf(stdout, "btn [%c] %s\n", inputChars[i], down ? "down" : "up");
        }
    }

    if (triggerVal != m_triggerVal)
    {
        // hard coding this as LEFT controller, index 0.
        cxrControllerEvent& e = events[0][eventCount[0]++];
        e.clientTimeNS = 1234+100;
        e.clientInputIndex = 4;
        e.inputValue.valueType = cxrInputValueType_float32;
        e.inputValue.vF32 = triggerVal;
        m_triggerVal = triggerVal;
        fprintf(stdout, "trigger: %f", triggerVal);
    }

    for (int i=0; i<CXR_NUM_CONTROLLERS; i++)
    {
        if (eventCount[i])
        {
            cxrError err = cxrFireControllerEvents(m_receiver, m_newControllers[i], events[i], eventCount[i]);
            if (err != cxrError_Success)
            {
                fprintf(stderr, "ERROR: cxrFireControllerEvents: '%s'", cxrErrorString(err));
            }
        }
    }
#endif
}

cxrBool VisionClientWindow::RenderAudio(const cxrAudioFrame *audioFrame)
{
    m_sdl.QueueAudio(audioFrame->streamBuffer, audioFrame->streamSizeBytes);
    return cxrTrue;
}

void VisionClientWindow::GetChaperone(cxrChaperone *chaperone)
{
    chaperone->universe = cxrUniverseOrigin_Standing;
    chaperone->origin.m[0][0] = chaperone->origin.m[1][1] = chaperone->origin.m[2][2] = 1;
    chaperone->origin.m[0][1] = chaperone->origin.m[0][2] = chaperone->origin.m[0][3] = 0;
    chaperone->origin.m[1][0] = chaperone->origin.m[1][2] = chaperone->origin.m[1][3] = 0;
    chaperone->origin.m[2][0] = chaperone->origin.m[2][1] = chaperone->origin.m[2][3] = 0;

    chaperone->playArea.v[0] = 4.0f;
    chaperone->playArea.v[1] = 4.0f;
}

void VisionClientWindow::GetGraphicsContext(cxrGraphicsContext *context, cxrGraphicsContextType type)
{
    m_platform.GetGraphicsContext(context, type);
}

// NOTE: This sample only currently supports frames of IDENTICAL size,
//  and renders all frames at once.
cxrError VisionClientWindow::SubmitFrame(cxrFramesLatched *framesLatched)
{
    if (framesLatched == NULL)
    {
        CXR_LOGE("Error: SubmitFrame called with no latched frames.");
        return cxrError_Frame_Not_Latched;
    }

    const uint32_t numFrames = framesLatched->count;
    if (numFrames == 0 || numFrames > CXR_MAX_NUM_VIDEO_STREAMS)
    {
        CXR_LOGE("Error: SubmitFrame called with %d latched frames, expecting [1-%d].", numFrames, CXR_MAX_NUM_VIDEO_STREAMS);
        return cxrError_Invalid_Number_Of_Streams; // seemed better to differentiate error code here vs above.
    }

    // TODO: I'm not so sure we should be using incoming frames in all
    //  cases here. Some should be predefined at clientDesc setup.

    uint32_t w = framesLatched->frames[0].width;
    uint32_t h = framesLatched->frames[0].height;
    for (uint32_t i = 1; i < numFrames; ++i)
    {
        w += framesLatched->frames[i].width;
        h = std::max(h, framesLatched->frames[i].height);
    }
    // On Windows, if the width and height are 0, this is the first
    // frame so create the window and graphics platform here. 
    if (!m_width || !m_height)
    {
#ifdef _WIN32      
        if (!m_sdl.InitWindow("CloudXR Desktop View", w, h, false))
        {
            CXR_LOGE("Fatal Error: Unable to initialize SDL window.");
            return cxrError_Failed; // TODO: something slightly more specific?
        }

        if (!m_platform.Create(m_sdl.GetHDC(), m_sdl.GetHWND(),
            framesLatched->frames[0].texture, w, h, false))
        {
            CXR_LOGE("Fatal Error: Unable to initialize platform graphics support.");
            return cxrError_Failed; // TODO: something slightly more specific?
        }

        m_width = w;
        m_height = h;
#endif
    }
    else if (w != m_width || h != m_height)
    {
        // TODO: do these need error checks?
        m_sdl.Resize(w, h);
        m_platform.Resize(w, h);

        m_width = w;
        m_height = h;
    }

    m_frameCounter++;

    // TODO should be handling return code here.
    m_platform.Draw(w, h, framesLatched, cxrFrameMask_All);

    return cxrError_Success;
}

bool VisionClientWindow::Update()
{
    m_sdl.PollEvents();

    return !m_sdl.ShouldQuit();
}
