/*
 * Copyright (c) 2019-2020, NVIDIA CORPORATION. All rights reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

#include "ClientCommon.h"

#include "GfxPlatform.h"

#ifdef _WIN32
#include <wrl/client.h>
#endif

#ifdef WITH_CUDA
#ifdef _WIN32
#include <cudaD3D11.h>
#else
#include <cudaGL.h>
#include <cuda_gl_interop.h>
#endif
#endif

#ifdef __aarch64__
#define DISABLE_CUDA_GL_INTEROP 1
#endif

#include <thread>
#include <chrono>
#include <sstream>
#include <iostream>

using std::chrono::milliseconds;
using std::this_thread::sleep_for;

#ifdef _WIN32
using Microsoft::WRL::ComPtr;
#endif

#if defined(__linux__)
#define CHECK_GL_ERROR() {                                              \
    GLenum err = glGetError();                                          \
    if (err != GL_NO_ERROR) {                                           \
        printf("GL Error 0x%x at %s:%d\n", err, __FILE__, __LINE__);    \
    }                                                                   \
}
#endif


#ifdef WITH_CUDA
static inline void check_drv(CUresult e, int iLine, const char *szFile)
{
    if (e != CUDA_SUCCESS)
    {
        const char* err;
        cuGetErrorString(e, &err);
        std::ostringstream errorLog;
        errorLog << "CUDA error " << e << " at line " << iLine << " in file " << szFile
            << ": " << err;
        std::cerr << errorLog.str() << std::endl;
        throw std::runtime_error(errorLog.str());
    }
}

#define CUDA_CALL(call) check_drv(call, __LINE__, __FILE__)
#endif


GfxPlatform::GfxPlatform(cxrGraphicsContextType type)
{
#ifdef _WIN32
    m_pSwapChain = NULL;
#else
    m_dpy = 0;
    m_win = 0;
    m_EGLdpy = 0;
    m_EGLctx = 0;
    m_EGLsurf = 0;
    m_shaderProgram = 0;
    pos_location = 0;
    tc_location = 0;
#endif
    m_type = type;
}

GfxPlatform::~GfxPlatform()
{

}

#define VENDORID_NVIDIA  0x10DE
#ifdef _WIN32
bool  GfxPlatform::InitD3D()
{
    if (m_device)
    {
        return true;
    }
    HRESULT hr = NULL;

    // Init DX
    CreateDXGIFactory1(__uuidof(IDXGIFactory1), (void**)&m_dxgiFactory1);
    if (!m_dxgiFactory1)
    {
        printf("Failed to create m_pDXGIFactory1");
        return false;
    }

    IDXGIAdapter *pAdapter = nullptr;
    DXGI_ADAPTER_DESC adapterDesc;

    for (UINT iAdapter = 0; m_dxgiFactory1->EnumAdapters(iAdapter, &pAdapter) != DXGI_ERROR_NOT_FOUND; ++iAdapter)
    {
        pAdapter->GetDesc(&adapterDesc);
#if _DEBUG
        printf("Adapter %d: Description='%S', VendorId=0x%x, DedicatedVideoMemory=%.1f MB, DedicatedSystemMemory=%.1f MB, SharedSystemMemory=%.1f MB, AdapterLuid=0x%08x%08x\n",
            iAdapter,
            adapterDesc.Description,
            adapterDesc.VendorId,
            (float)adapterDesc.DedicatedVideoMemory / (1024 * 1024),
            (float)adapterDesc.DedicatedSystemMemory / (1024 * 1024),
            (float)adapterDesc.SharedSystemMemory / (1024 * 1024),
            (unsigned int)adapterDesc.AdapterLuid.HighPart,
            (unsigned int)adapterDesc.AdapterLuid.LowPart);
#else
        printf("Adapter %d: Description='%S'", iAdapter, adapterDesc.Description);
#endif
        if (adapterDesc.VendorId == VENDORID_NVIDIA)
        {
            break;
        }
    }

    if (!pAdapter)
    {
        printf("No discrete adapter found.  Exiting.");
        return false;
    }

    D3D_FEATURE_LEVEL featureLevels[] =
    {
        D3D_FEATURE_LEVEL_11_0,
        D3D_FEATURE_LEVEL_10_1,
        D3D_FEATURE_LEVEL_10_0,
    };
    UINT numFeatureLevels = ARRAYSIZE(featureLevels);

#ifdef _DEBUG // attempt to create a debug device (not all systems have the libraries installed)
    hr = D3D11CreateDevice(pAdapter, D3D_DRIVER_TYPE_UNKNOWN, NULL, D3D11_CREATE_DEVICE_DEBUG, featureLevels, numFeatureLevels, D3D11_SDK_VERSION, &m_device, NULL, &m_immCtx);
#endif

    if (!m_device)
    {
        hr = D3D11CreateDevice(pAdapter, D3D_DRIVER_TYPE_UNKNOWN, NULL, 0, featureLevels, numFeatureLevels, D3D11_SDK_VERSION, &m_device, NULL, &m_immCtx);
    }

    if (FAILED(hr))
    {
        printf("D3D11CreateDevice failed with result: %#X", hr);
        return false;
    }

    return true;

}
#endif

bool
GfxPlatform::Create(NativeDisplayType dpy, NativeWindowType win,
                    cxrTextureHandle tex, uint32_t width, uint32_t height, bool fullscreen)
{
#ifdef _WIN32

    if (m_type == cxrGraphicsContext_Cuda)
    {
        InitD3D();
    }
    else
    {
        ID3D11Texture2D* d3dTex = TEXH_TO_D11TEX(tex);
        d3dTex->GetDevice(&m_device);
        m_device->GetImmediateContext(&m_immCtx);
    }

    //D3D11_TEXTURE2D_DESC desc;
    //d3dTex->GetDesc(&desc);

    // Create swap chain
    DXGI_SWAP_CHAIN_DESC swapChainDesc;
    memset(&swapChainDesc, 0, sizeof(swapChainDesc));

    swapChainDesc.BufferDesc.Width = width;
    swapChainDesc.BufferDesc.Height = height;
    swapChainDesc.BufferDesc.Format = DXGI_FORMAT_R8G8B8A8_UNORM_SRGB;
    swapChainDesc.SampleDesc.Count = 1;
    swapChainDesc.BufferUsage = DXGI_USAGE_RENDER_TARGET_OUTPUT;
    swapChainDesc.BufferCount = 1;
    swapChainDesc.OutputWindow = win;
    swapChainDesc.Windowed = fullscreen ? 0 : 1;

    ComPtr<IDXGIFactory1> dxgiFactory1;
    CreateDXGIFactory1(__uuidof(IDXGIFactory1), &dxgiFactory1);

    HRESULT hr = dxgiFactory1->CreateSwapChain(m_device, &swapChainDesc, &m_pSwapChain);
    if (FAILED(hr))
    {
        CXR_LOGE("Couldn't create debug view swapchain.");
        return false;
    }

#else
    if (!InitEGL(dpy, win))
    {
        CXR_LOGE("Failed to initialize EGL");
        return false;
    }
#endif
#ifdef WITH_CUDA
    if (m_type == cxrGraphicsContext_Cuda)
    {
        if (!InitCuda())
        {
            CXR_LOGE("Failed to initialize CUDA");
        }
    }
#endif
    return true;
}


// TODO: we want to be able to say latch all frames in one call, but then
//  individually render single or paired frames using mask.  Right now, we
//  do interop for all frames, but properly mask rendering.  So partway there...
bool GfxPlatform::Draw(uint32_t width, uint32_t height, cxrFramesLatched* framesLatched, uint32_t frameMask)
{
#ifdef WITH_CUDA
    if (m_type == cxrGraphicsContext_Cuda)
        SetupInterop(framesLatched, frameMask);
#endif

#ifdef _WIN32
    if (m_pSwapChain)
    {
        ComPtr<ID3D11Texture2D> dstTex;
        m_pSwapChain->GetBuffer(0, __uuidof(ID3D11Texture2D), &dstTex);

        uint32_t offsetX = 0;
        uint32_t framesDrawn = 0;
        for (uint32_t i = 0; i < CXR_MAX_NUM_VIDEO_STREAMS; i++)
        {
            if (0 == (frameMask & (1 << i)))
                continue; // skip, frame not in mask.

            ID3D11Texture2D* d3dTex = MapVideoFrameToD3D11Texture(&(framesLatched->frames[i]));
            m_immCtx->CopySubresourceRegion(*&dstTex, 0, offsetX, 0, 0, d3dTex, 0, NULL);
            offsetX += framesLatched->frames[i].width;

            if (++framesDrawn == framesLatched->count)
                break;
        }
        m_pSwapChain->Present(1, 0);
    }
    else
    {
        sleep_for(milliseconds(5));
    }
    return true;
#else
    return(DrawEGL(width, height, framesLatched, frameMask));
#endif
}

void GfxPlatform::Resize(uint32_t width, uint32_t height)
{
#ifdef _WIN32
    m_pSwapChain->ResizeBuffers(0, 0, 0, DXGI_FORMAT_UNKNOWN, 0);
#else
#endif
}

bool
GfxPlatform::Destroy()
{
#ifdef WITH_CUDA
    for (int i = 0; i < CXR_MAX_NUM_VIDEO_STREAMS; i++)
    {
        if (m_cuGfxResource[i])
        {
            cuCtxPushCurrent(m_cvContext);
            cuGraphicsUnregisterResource(m_cuGfxResource[i]);
            cuMemFreeHost(m_cudaHostPtr[i]);
            cuCtxPopCurrent(NULL);
        }
    }
#endif


#ifdef _WIN32
    for (int i = 0; i < CXR_MAX_NUM_VIDEO_STREAMS; i++)
    {
        if (m_pCudaInteropTexture[i])
        {
            m_pCudaInteropTexture[i]->Release();
            m_pCudaInteropTexture[i] = nullptr;
        }
    }

    if (m_pSwapChain)
    {
        m_pSwapChain->SetFullscreenState(false, NULL);
        m_pSwapChain->Release();
        m_pSwapChain = NULL;
    }

    if (m_immCtx)
    {
        m_immCtx->Release();
        m_immCtx = nullptr;
    }

    if (m_device)
    {
        m_device->Release();
        m_device = nullptr;
    }

    if (m_dxgiFactory1)
    {
        m_dxgiFactory1->Release();
        m_dxgiFactory1 = nullptr;
    }
    return true;
#else
    return(CleanupEGL());
#endif
}

void
GfxPlatform::GetGraphicsContext(cxrGraphicsContext *context, cxrGraphicsContextType type)
{
    if (!context)
    {
        CXR_LOGE("GraphicsContext is null");
        return;
    }

    context->type = type;
    switch (type)
    {
#if defined(__linux__)
    case cxrGraphicsContext_GLES:
        context->egl.display = m_EGLdpy;
        context->egl.context = m_EGLctx;
        break;
#endif
#ifdef WITH_CUDA
    case cxrGraphicsContext_Cuda:
        context->cuda.device = m_cvDevice;
        context->cuda.ctx = m_cvContext;
        break;
#endif
    default:
        CXR_LOGE("GetGraphicsContext of unknown type");
    }
}

#ifdef WITH_CUDA
void GfxPlatform::CopyCudaFrame(cxrVideoFrame* videoFrame)
{
    uint32_t streamIdx = videoFrame->streamIdx;
    CUDA_CALL(cuCtxPushCurrent(m_cvContext));
#if defined(DISABLE_CUDA_GL_INTEROP)
    CUDA_MEMCPY2D memcpy2D = { 0 };
    memcpy2D.srcMemoryType = CU_MEMORYTYPE_DEVICE;
    memcpy2D.srcDevice = (CUdeviceptr)videoFrame->texture;
    memcpy2D.srcPitch = videoFrame->pitch;
    memcpy2D.dstMemoryType = CU_MEMORYTYPE_HOST;
    memcpy2D.dstHost = m_cudaHostPtr[streamIdx];
    memcpy2D.dstPitch = videoFrame->width * 4;
    memcpy2D.WidthInBytes = videoFrame->width * 4;
    memcpy2D.Height = videoFrame->height;
    CUDA_CALL(cuMemcpy2D(&memcpy2D));
    glBindTexture(GL_TEXTURE_2D, m_tex[streamIdx]);
    glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, videoFrame->width, videoFrame->height, GL_RGBA, GL_UNSIGNED_BYTE, m_cudaHostPtr[streamIdx]);
#else
    cuGraphicsMapResources(1, &m_cuGfxResource[streamIdx], 0);
    CUarray dstArray;
    CUDA_CALL(cuGraphicsSubResourceGetMappedArray(&dstArray, m_cuGfxResource[streamIdx], 0, 0));
    CUDA_MEMCPY2D memcpy2D = { 0 };
    memcpy2D.srcMemoryType = CU_MEMORYTYPE_DEVICE;
    memcpy2D.srcDevice = (CUdeviceptr)videoFrame->texture;
    memcpy2D.srcPitch = videoFrame->pitch;
    memcpy2D.dstMemoryType = CU_MEMORYTYPE_ARRAY;
    memcpy2D.dstArray = dstArray;
    memcpy2D.WidthInBytes = videoFrame->width * 4;
    memcpy2D.Height = videoFrame->height;
    CUDA_CALL(cuMemcpy2D(&memcpy2D));
    CUDA_CALL(cuGraphicsUnmapResources(1, &m_cuGfxResource[streamIdx], 0));
#endif
    CUDA_CALL(cuCtxPopCurrent(NULL));
}

bool GfxPlatform::InitCuda()
{
    CUDA_CALL(cuInit(0));
#if 0
    CUdevice dev;
    uint32_t deviceCount = 0;
    CUDA_CALL(cuGLGetDevices(&deviceCount, &dev, 1, CU_GL_DEVICE_LIST_ALL));
#endif
    CUDA_CALL(cuDeviceGet(&m_cvDevice, 0));
    CUDA_CALL(cuDevicePrimaryCtxRetain(&m_cvContext, m_cvDevice));
    for(uint32_t i = 0; i < CXR_MAX_NUM_VIDEO_STREAMS; ++i)
    {
        m_cuGfxResource[i] = nullptr;
        m_cudaHostPtr[i] = nullptr;
    }
    return true;
}

// TODO: take frameMask as well?
void GfxPlatform::SetupInterop(cxrFramesLatched* framesLatched, uint32_t frameMask)
{
    if (m_bCudaInteropBuffersInitialized)
        return;
    CUDA_CALL(cuCtxPushCurrent(m_cvContext));
    MakeCurrent();

    // TODO: use mask to only set up interop for 'requested' frames.
    //  but for the moment for simplicity, we prepare all frames.
    // TODO: for per optimization with rendering individual frames or
    //  many subsets of a large number of streams, need to properly implement!
    for (uint32_t i = 0; i < framesLatched->count; i++)
    {
        cxrVideoFrame &vf = framesLatched->frames[i];
        const uint32_t streamIdx = vf.streamIdx;

#ifndef _WIN32
        glGenTextures(1, &m_tex[streamIdx]);

        glBindTexture(GL_TEXTURE_2D, m_tex[streamIdx]);

        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);

        glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, vf.width, vf.height, 0,
            GL_RGBA, GL_UNSIGNED_BYTE, nullptr);
        CHECK_GL_ERROR();
        CXR_LOGI("SetupInterop : StreamIdx = %d, GLTexture = %d, Wdith = %d, Height = %d", streamIdx,  m_tex[streamIdx], vf.width, vf.height);
#if defined(DISABLE_CUDA_GL_INTEROP)
        CUDA_CALL(cuMemAllocHost(&m_cudaHostPtr[streamIdx], vf.width * vf.height * 4));
#else
        CUDA_CALL(cuGraphicsGLRegisterImage(&m_cuGfxResource[streamIdx], m_tex[streamIdx], GL_TEXTURE_2D, CU_GRAPHICS_REGISTER_FLAGS_WRITE_DISCARD));
#endif
#else
        if (streamIdx >= CXR_MAX_NUM_VIDEO_STREAMS)
            return;
        D3D11_TEXTURE2D_DESC desc;
        memset(&desc, 0, sizeof(desc));
        desc.Width = vf.width;
        desc.Height = vf.height;
        desc.Format = DXGI_FORMAT_R8G8B8A8_UNORM_SRGB;
        desc.MipLevels = desc.ArraySize = 1;
        desc.SampleDesc.Count = 1;
        desc.SampleDesc.Quality = 0;
        desc.Usage = D3D11_USAGE_DEFAULT;
        desc.BindFlags = D3D11_BIND_SHADER_RESOURCE | D3D11_BIND_RENDER_TARGET;
        desc.CPUAccessFlags = 0;

        HRESULT hr;

        hr = m_device->CreateTexture2D(&desc, NULL, &m_pCudaInteropTexture[streamIdx]);
        CUDA_CALL(cuGraphicsD3D11RegisterResource(&m_cuGfxResource[streamIdx], m_pCudaInteropTexture[streamIdx], CU_GRAPHICS_REGISTER_FLAGS_NONE));
#endif
    }
    CUDA_CALL(cuCtxPopCurrent(NULL));
    m_bCudaInteropBuffersInitialized = true;
}

#ifdef __linux__
GLuint GfxPlatform::CudaToGlTex(cxrVideoFrame* frame)
{
    CopyCudaFrame(frame);
    return m_tex[frame->streamIdx];
}

#endif
#endif

#ifdef __linux__
GLint GfxPlatform::MapVideoFrameToGlTexture(cxrVideoFrame* videoFrame)
{
    // For Cuda gfx type copy it to a staging buffer and return the staging buffer
#ifdef WITH_CUDA
    if (m_type == cxrGraphicsContext_Cuda)
    {
        return CudaToGlTex(videoFrame);
    }
    else
#endif
    {
        return TEXH_TO_GLINT(videoFrame->texture);
    }
}
#endif

#ifdef _WIN32
ID3D11Texture2D* GfxPlatform::MapVideoFrameToD3D11Texture(cxrVideoFrame* videoFrame)
{
    // For Cuda gfx type copy it to a staging buffer and return the staging buffer
#ifdef WITH_CUDA
    if (m_type == cxrGraphicsContext_Cuda)
    {
        CopyCudaFrame(videoFrame);
        return m_pCudaInteropTexture[videoFrame->streamIdx];
    }
    else
#endif
    {
        return TEXH_TO_D11TEX(videoFrame->texture);
    }
}
#endif

#ifndef _WIN32
//
// G L S L  S H A D E R S
//

static const float kVertices[] = {
  -1.f, -1.f,
  1.f, -1.f,
  -1.f, 1.f,
  1.f, 1.f,
};

static const float kTextureCoords[] = {
  0.0f, 1.0f,
  1.0f, 1.0f,
  0.0f, 0.0f,
  1.0f, 0.0f,
};

static const char kVertexShader[] = "varying vec2 interp_tc;\n"
  "attribute vec4 in_pos;\n"
  "attribute vec2 in_tc; \n"
  "void main() { \n"
  "interp_tc = in_tc; \n" "gl_Position = in_pos; \n" "}\n";

static const char kFragmentShader[] =
  "precision mediump float;\n" "varying vec2 interp_tc; \n"
  "uniform sampler2D tex; \n" "void main() {\n"
  "gl_FragColor = texture2D(tex, interp_tc);\n" "}\n";

bool GfxPlatform::CreateShader(GLuint program, GLenum type,
              const char *source, int size)
{

    char log[4096];
    int result = GL_FALSE;

    GLuint shader = glCreateShader(type);

    glShaderSource(shader, 1, &source, &size);
    glCompileShader(shader);
    glGetShaderiv(shader, GL_COMPILE_STATUS, &result);
    if (!result)
    {

        glGetShaderInfoLog(shader, sizeof(log), NULL, log);
        return false;
    }
    glAttachShader(program, shader);

    glDeleteShader(shader);

    if (glGetError() != GL_NO_ERROR)
    {
      return false;
    }

    return true;
}

//
// Initialize EGL
//
bool GfxPlatform::InitEGL(NativeDisplayType dpy, NativeWindowType win)
{
    // Initialize EGL
    m_dpy = dpy;
    m_win = win;
    m_EGLdpy = eglGetPlatformDisplay(EGL_PLATFORM_X11_EXT, m_dpy, NULL);
    if (m_EGLdpy == EGL_NO_DISPLAY)
    {
        CXR_LOGE("InitEGL: eglGetPlatformDisplay returned EGL_NO_DISPLAY.\n");
        return false;
    }

    if (!eglInitialize(m_EGLdpy, NULL, NULL))
    {
        CXR_LOGE("InitEGL: eglInitialize failed.\n");
        return false;
    }

    // Choose EGL config
    EGLint attr[] = {
        EGL_RED_SIZE, 8,
        EGL_GREEN_SIZE, 8,
        EGL_BLUE_SIZE, 8,
        EGL_DEPTH_SIZE, 8,
        EGL_RENDERABLE_TYPE, EGL_OPENGL_BIT,
        EGL_SURFACE_TYPE, EGL_PBUFFER_BIT,
        EGL_NONE
    };

    EGLConfig ecfg;
    EGLint num_configs;
    if (!eglChooseConfig(m_EGLdpy, attr, &ecfg, 1, &num_configs))
    {
        CXR_LOGE("InitEGL: eglChooseConfig failed.\n");
        return false;
    }

    if (num_configs != 1)
    {
        CXR_LOGE("InitEGL: eglChooseConfig returned more than one config.\n");
        return false;
    }

    // Use the OpenGL API.
#ifdef __aarch64__
    eglBindAPI(EGL_OPENGL_ES_API);
    const EGLint contextAttrs[] =
    {
        EGL_CONTEXT_CLIENT_VERSION, 3,
        EGL_NONE,
    };


    m_EGLctx = eglCreateContext(m_EGLdpy, ecfg, EGL_NO_CONTEXT, contextAttrs);
#else
    eglBindAPI(EGL_OPENGL_API);
    m_EGLctx = eglCreateContext(m_EGLdpy, ecfg, EGL_NO_CONTEXT, nullptr);
#endif
    if (m_EGLctx == EGL_NO_CONTEXT)
    {
        CXR_LOGE("InitEGL: eglCreateContext returned EGL_NO_CONTEXT.\n");
        return false;
    }

    // Create Window Surface
    m_EGLsurf = eglCreateWindowSurface(m_EGLdpy, ecfg,
                       (EGLNativeWindowType)m_win, NULL);
    if (m_EGLsurf == EGL_NO_SURFACE)
    {
        CXR_LOGE("InitEGL: eglCreateWindowSurface returned EGL_NO_SURFACE\n");
        return false;
    }

    // Associate the EGL context with the EGL surface
    if (!eglMakeCurrent(m_EGLdpy, m_EGLsurf, m_EGLsurf, m_EGLctx))
    {
        CXR_LOGE("InitEGL: eglMakeCurrent failed.\n");
        return false;
    }




    // Initialize EGL rendering state
    glDisable(GL_DITHER);
    glDisable(GL_BLEND);
    glDisable(GL_DEPTH_TEST);

    // Create shader program for output image texture from FBO
    m_shaderProgram = glCreateProgram();

    // Create vertex shader
    if (!CreateShader(m_shaderProgram, GL_VERTEX_SHADER,
              kVertexShader, sizeof(kVertexShader)))
    {
        CXR_LOGE("InitEGL: CreateShader() failed to create vertex shader\n");
        return false;
    }

    // Create fragment shader
    if (!CreateShader(m_shaderProgram, GL_FRAGMENT_SHADER,
              kFragmentShader, sizeof(kFragmentShader)))
    {
        CXR_LOGE("InitEGL: CreateShader() failed to create fragment shader\n");
        return false;
    }

    // Link shader program
    glLinkProgram(m_shaderProgram);

    // Check for link errors
    GLint val;
    glGetProgramiv(m_shaderProgram, GL_LINK_STATUS, &val);
    if (!val) {
        GLint infoLength;
        glGetProgramiv(m_shaderProgram, GL_INFO_LOG_LENGTH, &infoLength);
        if (infoLength) {
            char *buf;
            buf = (char *)malloc(infoLength);
            if (buf) {
            glGetProgramInfoLog(m_shaderProgram, infoLength, NULL, buf);
            free(buf);
            }
        }
        CXR_LOGE("InitEGL: glLinkProgram failed to link GLSL shader program.\n");
        return false;
    }

    glUseProgram(m_shaderProgram);
    if (glGetError() != GL_NO_ERROR)
    {
        CXR_LOGE("InitEGL: glUseProgram failed.\n");
        return false;
    }

    pos_location = glGetAttribLocation(m_shaderProgram, "in_pos");

    glEnableVertexAttribArray(pos_location);
    glVertexAttribPointer(pos_location, 2, GL_FLOAT, GL_FALSE, 0, kVertices);
    tc_location = glGetAttribLocation(m_shaderProgram, "in_tc");

    glEnableVertexAttribArray(tc_location);
    glVertexAttribPointer(tc_location, 2, GL_FLOAT, GL_FALSE, 0,
              kTextureCoords);

    glActiveTexture(GL_TEXTURE0);
    glUniform1i(glGetUniformLocation(m_shaderProgram, "tex"), 0);

    if (glGetError() != GL_NO_ERROR)
    {
        CXR_LOGE("InitEGL: glGetError() returned error.\n");
        return false;
    }

    //glViewport(0, 0, m_width, m_height);

    return true;
}

bool GfxPlatform::MakeCurrent()
{
    return eglMakeCurrent(m_EGLdpy, m_EGLsurf, m_EGLsurf, m_EGLctx);
}

//
// Draw EGL scene.
//
bool GfxPlatform::DrawEGL(uint32_t width, uint32_t height, cxrFramesLatched* framesLatched, uint32_t frameMask)
{
    // Make EGL Context Current
    if (!eglMakeCurrent(m_EGLdpy, m_EGLsurf, m_EGLsurf, m_EGLctx))
    {
        return false;
    }

    uint32_t offsetX = 0;
    uint32_t framesDrawn = 0;
    for (uint32_t i = 0; i < CXR_MAX_NUM_VIDEO_STREAMS; ++i)
    {
        if (0 == (frameMask & (1 << i)))
            continue; // skip, frame not in mask.

        // Draw a single stream
        glActiveTexture(GL_TEXTURE0);
        glBindTexture(GL_TEXTURE_2D, MapVideoFrameToGlTexture(&(framesLatched->frames[i])));
        glViewport(offsetX, 0, framesLatched->frames[i].width, framesLatched->frames[i].height);
        glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
        offsetX += framesLatched->frames[i].width;

        if (++framesDrawn == framesLatched->count)
            break;
    }
    // Swap buffers
    if (!eglSwapBuffers(m_EGLdpy, m_EGLsurf))
    {
        return false;
    }

    return true;
}

//
// Destroy EGL objects.
//
bool GfxPlatform::CleanupEGL()
{
    glDeleteProgram(m_shaderProgram);

    if (!eglDestroySurface(m_EGLdpy, m_EGLsurf))
    {
        return false;
    }

    if (!eglDestroyContext(m_EGLdpy, m_EGLctx))
    {
        return false;
    }

    if (!eglTerminate(m_EGLdpy))
    {
      return false;
    }

    return true;
}

#else
bool GfxPlatform::MakeCurrent()
{
    return true;
}
#endif
