/*
 * Copyright (c) 2019-2020, NVIDIA CORPORATION. All rights reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

#pragma once
#include "CloudXRClient.h"

#ifdef WITH_CUDA
#include <cuda.h>
#endif

#ifndef _WIN32
#include <EGL/egl.h>
#include <EGL/eglext.h>
#ifdef __aarch64__
#include <GLES3/gl3.h>
#else
#include <GL/gl.h>
#endif
#else
#define NativeDisplayType HDC
#define NativeWindowType HWND
#endif


class GfxPlatform
{
public:
    GfxPlatform(cxrGraphicsContextType type = cxrGraphicsContextType::cxr_GraphicsContext_Max);
    ~GfxPlatform();

    bool Create(NativeDisplayType dpy, NativeWindowType win,
                 cxrTextureHandle tex, uint32_t width, uint32_t height, bool fullscreen = true);
    bool Draw(uint32_t width, uint32_t height, cxrFramesLatched* framesLatched, uint32_t frameMask);
    void Resize(uint32_t width, uint32_t height);
    bool MakeCurrent();
    bool Destroy();

    void GetGraphicsContext(cxrGraphicsContext *context, cxrGraphicsContextType type);

private:
#ifdef WITH_CUDA
    bool InitCuda();
    void SetupInterop(cxrFramesLatched* framesLatched, uint32_t frameMask);
    void CopyCudaFrame(cxrVideoFrame* videoFrame);
#endif

#ifdef _WIN32
    ID3D11Texture2D* MapVideoFrameToD3D11Texture(cxrVideoFrame* videoFrame);
    bool InitD3D();
    ID3D11Device* m_device = NULL;
    ID3D11DeviceContext *m_immCtx = NULL;
    IDXGISwapChain *m_pSwapChain;
    ID3D11Texture2D* m_pCudaInteropTexture[CXR_MAX_NUM_VIDEO_STREAMS] = {};
    IDXGIFactory1* m_dxgiFactory1 = nullptr;
#else
    GLint MapVideoFrameToGlTexture(cxrVideoFrame* videoFrame);
    bool CreateShader(GLuint program, GLenum type,
              const char *source, int size);
    bool InitEGL(NativeDisplayType dpy,
         NativeWindowType win);
    bool DrawEGL(uint32_t width, uint32_t height, cxrFramesLatched* framesLatched, uint32_t frameMask);
    bool CleanupEGL();
    bool CreateShader();
#ifdef WITH_CUDA
    GLuint CudaToGlTex(cxrVideoFrame* frame);
#endif

    Display *m_dpy;
    Window  m_win;
    EGLDisplay m_EGLdpy;
    EGLContext m_EGLctx;
    EGLSurface m_EGLsurf;

    GLuint m_shaderProgram;
    uint32_t pos_location;
    uint32_t tc_location;
    GLuint m_tex[CXR_MAX_NUM_VIDEO_STREAMS];
#endif

#ifdef WITH_CUDA
    CUgraphicsResource m_cuGfxResource[CXR_MAX_NUM_VIDEO_STREAMS] = {};
    CUdevice m_cvDevice{};
    CUcontext m_cvContext{};
    void* m_cudaHostPtr[CXR_MAX_NUM_VIDEO_STREAMS]{};
#endif
    cxrGraphicsContextType m_type;
    bool m_bCudaInteropBuffersInitialized = false;
};
