/*
 * Copyright (c) 2019-2020, NVIDIA CORPORATION. All rights reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

#include "ClientCommon.h"

#include "SDLWrapper.h"

#include <SDL.h>
#include <SDL_syswm.h>

#include <thread>
#include <chrono>

using std::chrono::milliseconds;
using std::this_thread::sleep_for;

SDLWrapper::~SDLWrapper()
{
    if (m_playbackDevice)
    {
        SDL_CloseAudioDevice(m_playbackDevice);
    }
    if (m_recordingDevice)
    {
        SDL_CloseAudioDevice(m_recordingDevice);
    }
    SDL_DestroyWindow(m_window);
    SDL_Quit();
}

bool SDLWrapper::InitWindow(const char* title, int width, int height, bool isOpenGL)
{
    if (m_window)
    {
        return true;
    }

    uint32_t flags = SDL_WINDOW_SHOWN | SDL_WINDOW_RESIZABLE;

    if (isOpenGL)
    {
        flags |= SDL_WINDOW_OPENGL;
    }

    m_window = SDL_CreateWindow(title, SDL_WINDOWPOS_CENTERED, SDL_WINDOWPOS_CENTERED, width, height, flags);
    if (!m_window)
    {
        CXR_LOGE("Failed to create SDL window: %s", SDL_GetError());
        return false;
    }

    return true;
}

bool SDLWrapper::InitVideo()
{
    if (SDL_Init(SDL_INIT_VIDEO) < 0)
    {
        CXR_LOGE("Unable to initialize SDL video: %s", SDL_GetError());
        return false;
    }
    return true;
}

void SDLWrapper::InitAudio(int frequency, uint8_t channels, unsigned bufferMs, bool& playback, bool& record, 
    AudioRecordingCallback audioRecordingCallback, void* audioRecordingContext)
{
    if (!record && !playback)
    {
        return;
    }

    if (SDL_Init(SDL_INIT_AUDIO) < 0)
    {
        CXR_LOGE("Unable to initialize SDL audio: %s", SDL_GetError());
        record = playback = false;
        return;
    }

    unsigned samplesPerMs = frequency / 1000;
    unsigned samples = samplesPerMs * bufferMs;
    m_bytesPerMs = samplesPerMs * CXR_AUDIO_SAMPLE_SIZE * channels;

    if (playback && m_playbackDevice == 0)
    {
        // Create the playback device
        SDL_AudioSpec audio_spec{};
        audio_spec.freq = frequency;
        audio_spec.channels = channels;
        audio_spec.samples = samples;
        audio_spec.format = AUDIO_S16LSB;

        // Open default audio device
        m_playbackDevice = SDL_OpenAudioDevice(nullptr, SDL_FALSE, &audio_spec, nullptr, 0);
        if (m_playbackDevice == 0)
        {
            CXR_LOGE("Unable to open audio device: %s", SDL_GetError());
            playback = false;
        }
        else
        {
            // Start audio playback
            SDL_PauseAudioDevice(m_playbackDevice, 0);
        }

    }

    if (record && m_recordingDevice == 0)
    {

        // Create the recording device
        SDL_AudioSpec recording_spec{};
        recording_spec.freq = frequency;
        recording_spec.channels = channels;
        recording_spec.samples = samples;
        recording_spec.format = AUDIO_S16LSB;
        recording_spec.callback = audioRecordingCallback;
        recording_spec.userdata = audioRecordingContext;

        // Open default recording device
        m_recordingDevice = SDL_OpenAudioDevice(nullptr, SDL_TRUE, &recording_spec, nullptr, 0);
        if (m_recordingDevice == 0)
        {
            CXR_LOGE("Unable to open recording device: %s", SDL_GetError());
            record = false;
        }
        else
        {
            // Start audio recording
            SDL_PauseAudioDevice(m_recordingDevice, 0);
        }
    }
}

void SDLWrapper::Resize(int width, int height)
{
    SDL_SetWindowSize(m_window, width, height);
}

#ifdef _WIN32
HDC SDLWrapper::GetHDC()
{
    SDL_SysWMinfo info;
    SDL_VERSION(&info.version);

    if (!SDL_GetWindowWMInfo(m_window, &info))
    {
        CXR_LOGE("Unable to get WM info: %s", SDL_GetError());
        return nullptr;
    }

    if (info.subsystem != SDL_SYSWM_WINDOWS)
    {
        CXR_LOGE("Expected SDL to report we're on Windows, it reported subsystem %d", (uint32_t)(info.subsystem));
        return nullptr;
    }

    return info.info.win.hdc;
}

HWND SDLWrapper::GetHWND()
{
    SDL_SysWMinfo info;
    SDL_VERSION(&info.version);

    if (!SDL_GetWindowWMInfo(m_window, &info))
    {
        CXR_LOGE("Unable to get WM info: %s", SDL_GetError());
        return nullptr;
    }

    if (info.subsystem != SDL_SYSWM_WINDOWS)
    {
        CXR_LOGE("Expected Windows subsystem");
        return nullptr;
    }

    return info.info.win.window;
}
#else
Display *SDLWrapper::GetDpy()
{
    SDL_SysWMinfo info;
    SDL_VERSION(&info.version);

    if (!SDL_GetWindowWMInfo(m_window, &info))
    {
        CXR_LOGE("Unable to get WM info: %s", SDL_GetError());
        return nullptr;
    }

    if (info.subsystem != SDL_SYSWM_X11)
    {
        CXR_LOGE("Expected X11 subsystem");
        return nullptr;
    }

    return info.info.x11.display;
}

Window SDLWrapper::GetWin()
{
    SDL_SysWMinfo info;
    SDL_VERSION(&info.version);

    if (!SDL_GetWindowWMInfo(m_window, &info))
    {
        CXR_LOGE("Unable to get WM info: %s", SDL_GetError());
        return 0;
    }

    if (info.subsystem != SDL_SYSWM_X11)
    {
        CXR_LOGE("Expected X11 subsystem");
        return 0;
    }

    return info.info.x11.window;
}
#endif

unsigned SDLWrapper::GetMouseState(int *x, int *y)
{
    return SDL_GetMouseState(x, y);
}

void SDLWrapper::PollEvents()
{
    SDL_Event event;
    while (SDL_PollEvent(&event))
    {
        if (event.type == SDL_KEYDOWN)
        {
            const auto sym = event.key.keysym.sym;

            switch (sym)
            {
            case SDLK_ESCAPE:
                m_shouldQuit = true;
                break;
            }
        }
        else if (event.type == SDL_QUIT)
        {
            m_shouldQuit = true;
        }
    }
}

void SDLWrapper::QueueAudio(const void *data, uint32_t len)
{
    if (SDL_QueueAudio(m_playbackDevice, data, len) != 0)
    {
        CXR_LOGE("SDL error: %s", SDL_GetError());
    }

    // Queue two frames of audio before blocking
    const uint32_t maxQueue = 2 * len;
    const uint32_t frameLenMs = len / m_bytesPerMs;
    while (SDL_GetQueuedAudioSize(m_playbackDevice) >= maxQueue)
    {
        sleep_for(milliseconds(frameLenMs));
    }
}

void SDLWrapper::FillMonitorParams(uint32_t &width, uint32_t &height, uint32_t &hz)
{
    int display_index = 0;
    SDL_DisplayMode mode = { SDL_PIXELFORMAT_UNKNOWN, 0, 0, 0, 0 };

    if (SDL_GetCurrentDisplayMode(display_index, &mode) != 0) {
        SDL_Log("SDL_GetCurrentDisplayMode failed: %s", SDL_GetError());
        width = 1024;
        height = 768;
        hz = 60;
    }
    else
    {
        width = mode.w;
        height = mode.h;
        hz = mode.refresh_rate ? mode.refresh_rate : 60;
    }
}
