/*
 * Copyright (c) 2019-2020, NVIDIA CORPORATION. All rights reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

#define LOG_TAG "main"
#include "ClientCommon.h"

#include <iostream>
#include "CloudXRClient.h"
#include "CloudXRInputEvents.h"
#include "CloudXRClientOptions.h"
#include "ClientOpenVR.h"
#include "ClientWindow.h"

#ifdef _WIN32
#include <shlobj.h>
HINSTANCE  g_instance;
#else
#include <string.h>
#define MAX_PATH 256
#endif

int main(int argc, char **argv)
{
    CXR_LOGI("CloudXR SDK Version: %s", CLOUDXR_VERSION);

#ifdef _WIN32
    g_instance = GetModuleHandle(NULL);
#endif

    // first, decide where input and output files go
    // TODO: might at some point prefer appBaseDir to be exe path or cwd.
    std::string appBaseDir = "";
    std::string appOutputDir = "";
#if defined(WIN32)
    char userDir[CXR_MAX_PATH + 1] = "";
    // we use the windows API to get the current user local appdata path to store our files.
    HRESULT result = SHGetFolderPath(NULL, CSIDL_LOCAL_APPDATA, NULL, SHGFP_TYPE_CURRENT, userDir);
    if (!SUCCEEDED(result) && !GetTempPath(sizeof(userDir), userDir))
        strcpy(userDir, "C:\\Temp"); // super-fallback case.
    appBaseDir = userDir;
    // then we append a subpath within the given directory where we want OUR files to appear
    // for the moment, we use our legacy directory, as some tests expect it.
    appBaseDir += "\\NVIDIA\\CloudXR\\SampleClient\\";
#elif defined (__linux__)
    appBaseDir = std::string(getenv("HOME")) + "/.CloudXR/SampleClient/";
#endif
    CXR_LOGI("App base directory set to: %s", appBaseDir.c_str());

#if defined(WIN32)
    appOutputDir = appBaseDir + "logs\\";
#else
    appOutputDir = appBaseDir + "logs/";
#endif
    CXR_LOGI("Output directory set to: %s", appOutputDir.c_str());

    CloudXR::ClientOptions options;
    // first, try to read launch options from text file
    std::string optionsPath = appBaseDir;
    optionsPath.append("CloudXRLaunchOptions.txt"); // path should already be slash terminated
    CXR_LOGI("Attempting to load launch options from: %s", optionsPath.c_str());
    ParseStatus status = options.ParseFile(optionsPath.c_str());
    if (status != ParseStatus_Success)
    { // TODO handle the individual parse errors.  fail, fnf, var, exit?
        if (status == ParseStatus_ExitRequested)
        { // this only occurs for -help or similar command.
            return 1;
        }
        if (status == ParseStatus_FileNotFound)
            CXR_LOGW("No server launch option file found at %s.", optionsPath.c_str());
        else
            CXR_LOGE("Error %d occurred while parsing server launch options file.", status);
    }


    // next, process actual 'commandline' args -- overrides any file-set values
    CXR_LOGI("Processing any command-line arguments.");
    status = options.ParseArgs(argc, argv);
    if (status == ParseStatus_ExitRequested)
    {
        return 1;
    }

    if (options.mServerIP.empty())
    {
        std::cout << "Enter server address: ";
        std::cin >> options.mServerIP;
    }

    // set any logger options PRIOR to init call.
    if (options.mDebugFlags & cxrDebugFlags_LogQuiet) // quiet takes precedence
        g_logFile.setLogLevel(cxrLL_Silence);
    else if (options.mDebugFlags & cxrDebugFlags_LogVerbose)
        g_logFile.setLogLevel(cxrLL_Verbose);
    else
        g_logFile.setLogLevel(cxrLL_Debug); // otherwise defaults to Info.

    g_logFile.setPrivacyEnabled((options.mDebugFlags & cxrDebugFlags_LogPrivacyDisabled) ? 0 : 1);
    g_logFile.setMaxSizeKB(options.mLogMaxSizeKB);
    g_logFile.setMaxAgeDays(options.mLogMaxAgeDays);

    // start the log file as early as we can...
    std::string filePrefix = "CloudXR Client";
    g_logFile.init(appOutputDir, filePrefix);

    ClientCommon *client = nullptr;
    if (options.mWindowed)
    {
        client = new VisionClientWindow(options.mGfxType);
    }
    else
    {
        client = new VisionClientOpenVR(options.mBtnRemap);
    }

    cxrError err = cxrError_Success;
    err = client->Init(options, appOutputDir);
    if (err == cxrError_Success)
    {
        err = client->Connect();
        // connect runs the main loop.  we could break that out into another fn if needed.
    } 

    delete client;

    if (cxrError_Success != err)
    {
        CXR_LOGE("Error preparing client: %s.  Exiting.", cxrErrorString(err));
        return 1;
    }

    return 0;
}
