/*
 * Copyright (c) 2023, NVIDIA CORPORATION. All rights reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

import Foundation

class CXRLogger {
    private let LOG_TAG = "CXRLogger"
    public var appBaseDirURL : URL
    public var appLogDirURL : URL
    private var logfileHandle : FileHandle?
    private var logPrefix : String
    
    init(filePrefix : String) {
        print("CXRLogger:init")
        
        logPrefix = filePrefix
        
        appBaseDirURL = FileManager.default.urls(for: .documentDirectory, in: .userDomainMask).first!
        appLogDirURL = appBaseDirURL.appendingPathComponent("logs")
        do {
            try FileManager.default.createDirectory(at: appLogDirURL, withIntermediateDirectories: true, attributes: nil)
        } catch {
            logMsg(level: cxrLL_Error, tag: LOG_TAG, msg: "Failed to create the app logs directory, logging to console only.")
            return
        }
        
        startLogFile()
    }
    
    func startLogFile()
    {
        if logfileHandle != nil { // then we already have log file open...
            logMsg(level: cxrLL_Warning, tag: LOG_TAG, msg: "Log file already opened, ignoring startLogFile")
            return
        }
        
        let dateFmt = DateFormatter()
        dateFmt.locale = Locale(identifier: "en_US_POSIX")
        dateFmt.dateFormat = "yyyy-MM-dd HH.mm.ss"
        let fileName = String(format: "%@ Log %@.txt", logPrefix, dateFmt.string(from: Date.init(timeIntervalSinceNow: 0)))

        let logfileUrl = appLogDirURL.appendingPathComponent(fileName)
        guard FileManager.default.createFile(atPath: logfileUrl.path,
                                             contents: nil)
        else {
            // should we really throw here, or just log this error and move on?
            // systen is still 'functional' without writing to file, just annoying.
            logMsg(level: cxrLL_Error, tag: LOG_TAG, msg: "Unable to open log file: " + logfileUrl.absoluteString)
            return // so long as not other setup below but opening the file...
        }

        logfileHandle = FileHandle(forWritingAtPath: logfileUrl.path)
        if logfileHandle == nil {
            logMsg(level: cxrLL_Error, tag: LOG_TAG, msg: "Failed to open file handle to log file, logging to console only.")
        } else {
            logMsg(level: cxrLL_Info, tag: LOG_TAG, msg: "Log file opened and writing...")
        }
        logfileHandle?.seekToEndOfFile()
    }

    // Temp commented out Category, as we're not yet really differentiating in our logging.
    func logMsg(level: cxrLogLevel, /*category: cxrMessageCategory,*/ tag: String, msg: String)
    {
        let timeFmt = DateFormatter()
        timeFmt.dateFormat = "HH:mm:ss.SSS"
        var formattedMsg = String(format: "[%@]  %c (%@)  %@", timeFmt.string(from: Date.init(timeIntervalSinceNow: 0)), cxrLLToChar(level), tag, msg)

        // always print to console for now.
        print(formattedMsg)
        // cool snippet we could apply in the future
        //if isatty(STDERR_FILENO) != 1 { // if no terminal is attached to stderr, don't bother printing?

        if logfileHandle != nil {
            formattedMsg += "\n" // need a linefeed for the file write line ending, but AFTER we print which adds one for us.
            if let data = formattedMsg.data(using: .utf8) {
                // do we really need to seek every time if we're not closing like dumb examples?
                //logfileHandle?.seekToEndOfFile()
                logfileHandle?.write(data)
            }
        }
    }
    
    func closeLogFile() {
        if logfileHandle != nil {
            logMsg(level: cxrLL_Info, tag: LOG_TAG, msg: "Closing log file.")
            logfileHandle?.closeFile()
        }
        logfileHandle = nil
    }
    
    deinit {
        closeLogFile()
    }
}
