/*
 * Copyright (c) 2021-2023, NVIDIA CORPORATION. All rights reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

import UIKit
import Metal
import MetalKit
import ARKit

// TODO: by static initializing here, we have no way to close and reopen
// probably needs to move up into one of the other files so we can actually create/destroy or at least add a 'reopen'
// method to restart logging when the app goes fg -> bg -> fg.
let cLogger = CXRLogger(filePrefix: "iOS Client")

let cameraImageCacheSize:Int = 32      // Ringbuffer size of previous frames & hmd poses
let maxTimeSamples:Int = 10            // Used for rough estimate of FPS calculation

// Shader data used to render the ground plane as a grid
struct shader_data
{
    var viewMatrix:simd_float4x4
    var projectionMatrix:simd_float4x4
}

struct line_seg
{
    var start:simd_float3
    var end:simd_float3
}

func intersect(_ l:line_seg,_ planeP:simd_float3 , _ planeN:simd_float3) -> simd_float3
{
    let dir = simd_normalize(l.end - l.start);
    let d = -simd_dot(planeP, planeN);
    let dist = -(simd_dot(l.start, planeN) + d) / (simd_dot(dir, planeN));
    let r = l.start + (dir * dist);
    return r;
}

func distance(_ a:simd_float3,_ b:simd_float3) -> Float
{
    let v = b - a;
    return sqrt(simd_dot(v, v));
}

func pointInPlaneRect(_ p:ARPlaneAnchor, _ x:simd_float3) -> Bool
{
    let t = p.transform;
    let tInverse = simd_inverse(t);
    let c = p.center;
    let xAnchor = simd_make_float3(simd_mul(tInverse, simd_make_float4(x, 1))) // ? not sure
    let e = p.extent / 2.0;
    let xIn:Bool = (xAnchor.x >= (c.x - e.x)) && (xAnchor.x <= (c.x + e.x));
    let zIn:Bool = (xAnchor.z >= (c.z - e.z)) && (xAnchor.z <= (c.z + e.z));
    return xIn && zIn;
}

// Convert cxrError into a throwable Swift object
public enum CXRError: Error {
    case failed
    case timeout
    case notConnected
    case notStreaming
    case notImplemented
    case requiredParameter
    case moduleLoadFailed
    case invalidAPIVersion
    case invalidNumberOfStreams
    case invalidDeviceDescriptor
    case invalidGraphicsContext
    case invalidVideoFormat
    case invalidVideoWidth
    case invalidVideoHeight
    case invalidVideoFps
    case invalidVideoMaxBitrate
    case invalidFrameMask
    case serverSetupFailed
    case serverVersionOld
    case serverUntrustedCertificate
    case serverInitiatedDisconnect
    case serverHandshakeFailed
    case serverStreamConfigurationFailed
    case serverFirewallConfigurationFailed
    case serverFeatureDisabledHEVC
    case serverFeatureDisabledVVSync
    case serverFeatureDisabledPosePrediction
    case serverFeatureDisabledSendAudio
    case serverFeatureDisabledReceiveAudio
    case clientSetupFailed
    case clientVersionOld
    case clientUnauthorized
    case frameNotReleased
    case frameNotLatched
    case frameNotReady
    case decoderSetupFailed
    case decoderNoTexture
    case decoderFrameNotReady
    case poseCallbackProvided
    case nameTooLong
    case tooManyInputs
    case controllerIdInUse
    case roleTooLong
    case dataTooLarge
}

func CXRConvertError(_ e: cxrError) -> CXRError {
    switch e {
    case cxrError_Failed:
        return .failed
    case cxrError_Timeout:
        return .timeout
    case cxrError_Not_Connected:
        return .notConnected
    case cxrError_Not_Streaming:
        return .notStreaming
    case cxrError_Not_Implemented:
        return .notImplemented
    case cxrError_Required_Parameter:
        return .requiredParameter
    case cxrError_Module_Load_Failed:
        return .moduleLoadFailed
    case cxrError_Invalid_API_Version:
        return .invalidAPIVersion
    case cxrError_Invalid_Number_Of_Streams:
        return .invalidNumberOfStreams
    case cxrError_Invalid_Device_Descriptor:
        return .invalidDeviceDescriptor
    case cxrError_Invalid_Graphics_Context:
        return .invalidGraphicsContext
    case cxrError_Invalid_Video_Format:
        return .invalidVideoFormat
    case cxrError_Invalid_Video_Width:
        return .invalidVideoWidth
    case cxrError_Invalid_Video_Height:
        return .invalidVideoHeight
    case cxrError_Invalid_Video_Fps:
        return .invalidVideoFps
    case cxrError_Invalid_Video_Max_Bitrate:
        return .invalidVideoMaxBitrate
    case cxrError_Invalid_Frame_Mask:
        return .invalidFrameMask
    case cxrError_Server_Setup_Failed:
        return .serverSetupFailed
    case cxrError_Server_Version_Old:
        return .serverVersionOld
    case cxrError_Server_Untrusted_Certificate:
        return .serverUntrustedCertificate
    case cxrError_Server_Initiated_Disconnect:
        return .serverInitiatedDisconnect
    case cxrError_Server_Handshake_Failed:
        return .serverHandshakeFailed
    case cxrError_Server_Stream_Configuration_Failed:
        return .serverStreamConfigurationFailed
    case cxrError_Server_Firewall_Configuration_Failed:
        return .serverFirewallConfigurationFailed
    case cxrError_Server_Feature_Disabled_HEVC:
        return .serverFeatureDisabledHEVC
    case cxrError_Server_Feature_Disabled_VVSync:
        return .serverFeatureDisabledVVSync
    case cxrError_Server_Feature_Disabled_Pose_Prediction:
        return .serverFeatureDisabledPosePrediction
    case cxrError_Server_Feature_Disabled_Send_Audio:
        return .serverFeatureDisabledSendAudio
    case cxrError_Server_Feature_Disabled_Receive_Audio:
        return .serverFeatureDisabledReceiveAudio
    case cxrError_Client_Setup_Failed:
        return .clientSetupFailed
    case cxrError_Client_Version_Old:
        return .clientVersionOld
    case cxrError_Client_Unauthorized:
        return .clientUnauthorized
    case cxrError_Frame_Not_Released:
        return .frameNotReleased
    case cxrError_Frame_Not_Latched:
        return .frameNotLatched
    case cxrError_Frame_Not_Ready:
        return .frameNotReady
    case cxrError_Decoder_Setup_Failed:
        return .decoderSetupFailed
    case cxrError_Decoder_No_Texture:
        return .decoderNoTexture
    case cxrError_Decoder_Frame_Not_Ready:
        return .decoderFrameNotReady
    case cxrError_Pose_Callback_Provided:
        return .poseCallbackProvided
    case cxrError_Name_Too_Long:
        return .nameTooLong
    case cxrError_Too_Many_Inputs:
        return .tooManyInputs
    case cxrError_Controller_Id_In_Use:
        return .controllerIdInUse
    case cxrError_Role_Too_Long:
        return .roleTooLong
    case cxrError_Data_Too_Large:
        return .dataTooLarge
    default:
        return .failed
    }
}

enum TimerIndex:Int {
    case TIMER_METAL = 0
    case TIMER_CXR   = 1
    case TIMER_ARKIT = 2
    case TIMER_COUNT = 3
}

func MatrixToVecQuat(mat: cxrMatrix34) -> (cxrVector3, cxrQuaternion) {
    var q = cxrQuaternion()
   
    q.w = sqrt(fmax(0.0, 1.0 + mat.m.0.0   + mat.m.1.1 + mat.m.2.2)) / 2.0;
    q.x = sqrt(fmax(0.0, 1.0 + mat.m.0.0 - mat.m.1.1 - mat.m.2.2)) / 2.0;
    q.y = sqrt(fmax(0.0, 1.0 - mat.m.0.0 + mat.m.1.1 - mat.m.2.2)) / 2.0;
    q.z = sqrt(fmax(0.0, 1.0 - mat.m.0.0 - mat.m.1.1 + mat.m.2.2)) / 2.0;
    
    q.x = copysign(q.x, mat.m.2.1 - mat.m.1.2);
    q.y = copysign(q.y, mat.m.0.2 - mat.m.2.0);
    q.z = copysign(q.z, mat.m.1.0 - mat.m.0.1);

    let pos = cxrVector3(v: (mat.m.0.3, mat.m.1.3,mat.m.2.3))
    return (pos,q)
}

class CXRViewController: UIViewController, MTKViewDelegate, ARSessionDelegate {
    
    // CloudXR
    fileprivate var receiver : cxrReceiverHandle?
    var receiverHandlePtr : UnsafeMutablePointer<cxrReceiverHandle?>?
    var framesLatched = 0
    var latchedFrame:UnsafeMutablePointer<cxrFramesLatched>?
    var latchedPose:simd_float3x4 = simd_float3x4.init(1.0)
    var latchedTextureY : MTLTexture?
    var latchedTextureCbCr : MTLTexture?
    var latchedTextureAlphaY : MTLTexture?
    var latchedTextureAlphaCbCr : MTLTexture?
    var stats:UnsafeMutablePointer<cxrConnectionStats>?
    var framesUntilStats = 100
    var clientState : cxrClientState?
    var clientError : cxrError?
    // Metal
    var device : MTLDevice?
    var commandQ : MTLCommandQueue?
    var textureCache : CVMetalTextureCache?
    var psoCameraOnly : MTLRenderPipelineState?
    var psoCompositeXR : MTLRenderPipelineState?
    var imagePlaneVertexBuffer: MTLBuffer?
    let kImagePlaneVertexData:[Float] =  [             // Vertex data for an image plane
        1.0, 1.0,
        0.0, 1.0,
        0.0, 0.0,
        //
        1.0, 1.0,
        0.0, 0.0,
        1.0, 0.0,
    ]
    // Metal: Render debug lines
    var psoVisPlane:MTLRenderPipelineState?            // Pipeline for drawing debug lines
    var constantBuffer:MTLBuffer?                      // Shader constants (projMat, viewMat, etc)
    var lineVertBuffer:MTLBuffer?                      // Vertices (endpoints) of the line segments
    var lineVertCount:UInt = 0
    // Metal: Render ground planes
    var psoVisTris:MTLRenderPipelineState?              // Pipeline used to draw the grid on the detected ground planes
    var triVertBuffer:MTLBuffer?                        // Vertices of the triangles that represent the ground planes
    var triVertCount:UInt = 0
    var gridTex:MTLTexture?
    var tapPoint:CGPoint? = CGPoint(x: 0.0,y: 0.0)      // Screen space coordinates of the finger tap location
    // Metal: Compute shader for blit operation
    var computePipelineState:MTLComputePipelineState?   // Pipeline used to copy ARKit camera images into the cache
    var copyTextureKernel:MTLFunction?
    var computeThreadgroupSize:MTLSize? = MTLSize()
    var computeThreadgroupCount:MTLSize? = MTLSize()
    // ARKit
    var session = ARSession()
    var trackingConfig:ARWorldTrackingConfiguration? = ARWorldTrackingConfiguration()
    var seenAnchors:[ARAnchor] = []
    var cameraTextureY : MTLTexture?
    var cameraTextureCbCr : MTLTexture?
    var lastTransform = simd_float4x4.init(diagonal: SIMD4<Float>(repeating: 1.0))
    var planeSelected: Bool = false
    var scale:Float = 1.0                                   // Used to align the real and virtual worlds
    var rotationAngleDegrees:Float = 0.0                    // Used to align the real and virtual worlds
    var originPos:simd_float3 = simd_float3(repeating: 0.0) // Used to align the real and virtual worlds
    var trackingState = cxrVRTrackingState()
    // ARKit: Client data cache
    // The camera images received from ARKit need to be cached because they are only valid for a small amount of time.
    // The network round trip takes time and by the time it finishes, the ARKit camera image that is needed for
    // composition will (likely) have been overwritten. In addition to the camera images being cached, the ARKit poses
    // that correlate with those images are also cached.
    var cameraTextureCacheY    = [MTLTexture?]    (repeating: nil, count: cameraImageCacheSize)
    var cameraTextureCacheCbCr = [MTLTexture?]    (repeating: nil, count: cameraImageCacheSize)
    var cameraPoseMatrixCache  = [simd_float3x4?] (repeating: nil, count: cameraImageCacheSize)
    
    var cameraImageCacheIndexLock:NSLock? = NSLock()
    var cameraImageCacheWriteIndex:Int = 0
    var cameraImageCacheReadIndex:Int = 0
    
    // Timer queries. These are different from the data captured with CAPTURE_PERF.
    // These queries are low precision and only used to estimate FPS of each individual
    // thread (render, cxr, arkit).
    var startTimes     = [NSDate?](repeating: nil, count: TimerIndex.TIMER_COUNT.rawValue)
    var endTimes       = [NSDate?](repeating: nil, count: TimerIndex.TIMER_COUNT.rawValue)
    var timerSamples   = Array(repeating: Array(repeating: 0.0, count: maxTimeSamples), count: TimerIndex.TIMER_COUNT.rawValue)
    var timerIndex     = [Int](repeating:0, count: TimerIndex.TIMER_COUNT.rawValue)
    var timerFirstTime = [Bool](repeating:true, count: TimerIndex.TIMER_COUNT.rawValue)
       
    public var address : String = ""    // this is set right before we segue into this view controller

    @IBOutlet var mtlView: MTKView!
    @IBOutlet var rotationSlider: UISlider!
    @IBOutlet var scaleSlider: UISlider!
    
    @IBOutlet var framesLatchedLabel: UILabel!
    @IBOutlet var statusLabel: UILabel!
    @IBOutlet var scaleLabel: UILabel!
    @IBOutlet var rotationLabel: UILabel!
    @IBOutlet var cxrFPSLabel: UILabel!
    @IBOutlet var arkitFPSLabel: UILabel!
    @IBOutlet var metalFPSLabel: UILabel!    
    @IBOutlet var framesBehindLabel: UILabel!
    
    private let LOG_TAG = String("CXRViewController")
    var connAttemptCount = 0

    // MARK: UIViewController delegates

    override func viewDidLoad() {
        super.viewDidLoad()
        UIApplication.shared.isIdleTimerDisabled = true;
        self.modalPresentationStyle = .fullScreen
        self.statusLabel.text = ""
        
        device = MTLCreateSystemDefaultDevice()
        commandQ = device?.makeCommandQueue()
        mtlView.delegate = self
        mtlView.device = device
        mtlView.clearColor = MTLClearColorMake(0.0, 1.0, 0.0, 1.0)
        mtlView.enableSetNeedsDisplay = false
        mtlView.preferredFramesPerSecond = 60; // default is 60 according to iOS docs, but hard code anyways
        let pipelineDescriptor = MTLRenderPipelineDescriptor()
        let library = device!.makeDefaultLibrary()
        pipelineDescriptor.vertexFunction = library!.makeFunction(name: "vertexShader")
        pipelineDescriptor.fragmentFunction = library!.makeFunction(name: "fragmentShader")
        pipelineDescriptor.colorAttachments[0].pixelFormat = mtlView.colorPixelFormat
        pipelineDescriptor.depthAttachmentPixelFormat = mtlView.depthStencilPixelFormat
        do {
            try psoCameraOnly = device!.makeRenderPipelineState(descriptor: pipelineDescriptor)
            pipelineDescriptor.fragmentFunction = library!.makeFunction(name: "fragmentShaderCompositeXR")
            try psoCompositeXR = device!.makeRenderPipelineState(descriptor: pipelineDescriptor)
        }
        catch {
            cLogger.logMsg(level: cxrLL_Error, tag: LOG_TAG, msg: "Failed to create PSO")
            return
        }
        imagePlaneVertexBuffer = device?.makeBuffer(length: MemoryLayout<Float>.size * 12, options: MTLResourceOptions.storageModeShared)
        //
        pipelineDescriptor.vertexFunction = library!.makeFunction(name: "basicVS")
        pipelineDescriptor.fragmentFunction = library!.makeFunction(name: "solidColor")
        do {
            try psoVisPlane = device!.makeRenderPipelineState(descriptor:pipelineDescriptor)
        }
        catch {
            cLogger.logMsg(level: cxrLL_Error, tag: LOG_TAG, msg: "Failed to create psoVisPlane")
            return
        }
        constantBuffer = device?.makeBuffer(length:MemoryLayout<shader_data>.size, options:MTLResourceOptions.storageModeShared)
        lineVertBuffer = device?.makeBuffer(length:1024, options:MTLResourceOptions.storageModeShared)
        //
        // Compute shader for blit operations
        copyTextureKernel = library!.makeFunction(name: "copyTextureKernel")
        do {
            try computePipelineState = device?.makeComputePipelineState(function: copyTextureKernel!)
        }
        catch {
            cLogger.logMsg(level: cxrLL_Error, tag: LOG_TAG, msg: "Failed to create copyTextureKernel")
            return
        }
        
        //
        pipelineDescriptor.vertexFunction = library!.makeFunction(name: "triVS")
        pipelineDescriptor.fragmentFunction = library!.makeFunction(name: "triFS")
        //
        pipelineDescriptor.colorAttachments[0].isBlendingEnabled = true;
        pipelineDescriptor.colorAttachments[0].rgbBlendOperation = MTLBlendOperation.add
        pipelineDescriptor.colorAttachments[0].alphaBlendOperation = MTLBlendOperation.add;
        pipelineDescriptor.colorAttachments[0].sourceRGBBlendFactor = MTLBlendFactor.sourceAlpha
        pipelineDescriptor.colorAttachments[0].sourceAlphaBlendFactor = MTLBlendFactor.sourceAlpha;
        pipelineDescriptor.colorAttachments[0].destinationRGBBlendFactor = MTLBlendFactor.oneMinusSourceAlpha;
        pipelineDescriptor.colorAttachments[0].destinationAlphaBlendFactor = MTLBlendFactor.oneMinusSourceAlpha;
        //
        do {
            try psoVisTris = device!.makeRenderPipelineState(descriptor: pipelineDescriptor)
        }
        catch {
        }
        triVertBuffer = device!.makeBuffer(length: 4096, options: MTLResourceOptions.storageModeShared)
        let loader:MTKTextureLoader = MTKTextureLoader(device: device!)
        let url:URL? = Bundle.main.url(forResource: "grid", withExtension: "png") as URL?
        do {
            try gridTex = loader.newTexture(URL: url!, options: nil)
        }
        catch {
            cLogger.logMsg(level: cxrLL_Error, tag: LOG_TAG, msg: "Failed to load grid texture.")
            return
        }
        //
        let r = CVMetalTextureCacheCreate(kCFAllocatorDefault, nil, device!, nil, &textureCache)
        if r != kCVReturnSuccess {
            cLogger.logMsg(level: cxrLL_Error, tag: LOG_TAG, msg: "Error creating metal texture cache")
            return
        }
        session.delegate = self
        
        //
        self.mtlView.isUserInteractionEnabled = true
        let tapGestureRecognizer = UITapGestureRecognizer(target: self, action: #selector(self.handleTap(_:)))
        self.mtlView.addGestureRecognizer(tapGestureRecognizer)
        
        let longPressGestureRecognizer = UILongPressGestureRecognizer(target: self, action: #selector(self.handleLongPress(_:)))
        self.mtlView.addGestureRecognizer(longPressGestureRecognizer)
        
        // ARKit cache of iamges and poses
        for i in 0...cameraImageCacheSize-1 {
            cameraTextureCacheY[i] = nil;
            cameraTextureCacheCbCr[i] = nil;
            cameraPoseMatrixCache[i] = simd_float3x4(0.0)
        }
        
        // Thread FPS timer related
        for i in 0...TimerIndex.TIMER_COUNT.rawValue-1 {
            timerFirstTime[i] = true;
            timerIndex[i] = 0;
            for j in 0...maxTimeSamples-1 {
                timerSamples[i][j] = 0.0
            }
        }
        latchedFrame = UnsafeMutablePointer<cxrFramesLatched>.allocate(capacity: 1)
        stats = UnsafeMutablePointer<cxrConnectionStats>.allocate(capacity: 1)
        //
        trackingConfig!.planeDetection = ARWorldTrackingConfiguration.PlaneDetection.horizontal
        session.run(trackingConfig!)
        
        receiverHandlePtr = withUnsafeMutablePointer(to: &receiver, { $0 })
        
        clientState = cxrClientState_ReadyToConnect
    }
    
    override func viewWillDisappear(_ animated: Bool) {
        session.pause();
        if (self.receiver != nil)
        {
            cxrDestroyReceiver(self.receiver!)
        }
        cLogger.closeLogFile();
    }
    
    // MARK: CloudXR

    func CLOUDXR_VERSION_DWORD(major: Int32, minor:Int32, patch:Int32) -> UInt32
    {
        UInt32(major << 16 | minor << 8 | patch)
    }
    
    func connectToServer(_ projectionMatrix: simd_float4x4) -> Bool {

        var desc: cxrReceiverDesc = cxrReceiverDesc()

        	cLogger.startLogFile()
        
        desc.requestedVersion = CLOUDXR_VERSION_DWORD(major: CLOUDXR_VERSION_MAJOR,
                                                     minor: CLOUDXR_VERSION_MINOR,
                                                     patch: CLOUDXR_VERSION_PATCH)

        desc.deviceDesc.numVideoStreamDescs = 1;
        desc.deviceDesc.videoStreamDescs.0.format = cxrClientSurfaceFormat_RGBA;
        desc.deviceDesc.videoStreamDescs.0.width = UInt32(self.mtlView.bounds.size.width);
        desc.deviceDesc.videoStreamDescs.0.height = UInt32(self.mtlView.bounds.size.height);
        desc.deviceDesc.videoStreamDescs.0.fps = 60.0;
        desc.deviceDesc.videoStreamDescs.0.maxBitrate = UInt32(CXR_DEFAULT_VIDEO_STREAM_BITRATE);
        desc.deviceDesc.stereoDisplay = UInt8(cxrFalse);
        
        let projTanHalfAngles = projectionMatrixToTangentHalfAngles(projectionMatrix) // l,r,t,b

        desc.deviceDesc.maxResFactor = 1.0;
        desc.deviceDesc.ipd = 0.00;
        desc.deviceDesc.proj.0.0 = projTanHalfAngles.0  // left;
        desc.deviceDesc.proj.0.1 = projTanHalfAngles.1  // right;
        desc.deviceDesc.proj.0.2 = projTanHalfAngles.2  // top;
        desc.deviceDesc.proj.0.3 = projTanHalfAngles.3  // bottom;
        desc.deviceDesc.proj.1.0 = projTanHalfAngles.0  // left;
        desc.deviceDesc.proj.1.1 = projTanHalfAngles.1  // right;
        desc.deviceDesc.proj.1.2 = 0.0; // disable right eye rendering on server
        desc.deviceDesc.proj.1.3 = 0.0; // disable right eye rendering on server
        desc.deviceDesc.predOffset = 0.0;
        desc.deviceDesc.receiveAudio = UInt8(cxrFalse);
        desc.deviceDesc.sendAudio = UInt8(cxrFalse);
        desc.deviceDesc.embedInfoInVideo = UInt8(cxrFalse);
        desc.deviceDesc.disablePosePrediction = UInt8(cxrTrue);
        desc.deviceDesc.angularVelocityInDeviceSpace = UInt8(cxrFalse);
        desc.deviceDesc.foveatedScaleFactor = 0;
        desc.deviceDesc.posePollFreq = 0; //using push pose mode
        desc.deviceDesc.foveationModeCaps = 0;
        desc.deviceDesc.chaperone.universe = cxrUniverseOrigin_Standing;
        var identity:cxrMatrix34 = cxrMatrix34()
        identity.m.0.0 = 1.0
        identity.m.1.1 = 1.0
        identity.m.2.2 = 1.0
        desc.deviceDesc.chaperone.origin = identity;
        desc.deviceDesc.chaperone.playArea.v.0 = 0.0
        desc.deviceDesc.chaperone.playArea.v.1 = 0.0;
        desc.deviceDesc.maxClientQueueSize = 8.0; // Client-side dejitter buffer size
        desc.deviceDesc.disableVVSync = UInt8(cxrTrue);
        
        var clientProxy = cxrClientCallbacks()
        clientProxy.UpdateClientState = { context, state, error in
            let viewController = Unmanaged<CXRViewController>.fromOpaque(context!).takeUnretainedValue()
            viewController.clientState = state
            viewController.clientError = error
        }
        
        clientProxy.LogMessage = { context, level, category, extra, tag, messageText in
            // so Swift has some problems with auto-checking for nil strings.
            // here, we skip shorthand and do it out long, create a String for each C string,
            // and handle the nil cases.
            let tagString:String
            let msgString:String
            if tag != nil {
                tagString = String(cString:tag!)
            } else {
                tagString = "Unknown"
            }
            if messageText != nil {
                msgString = String(cString:messageText!)
            } else {
                msgString = "Unknown"
            }
            
            // we then pass to our logger object.  This could just as easily be a static helper func.
            cLogger.logMsg(level: level, tag: tagString, msg: msgString);
        }
        
        // noting this chage: client context is now owned by the Callbacks, not the device descriptor.
        clientProxy.clientContext = Unmanaged.passUnretained(self).toOpaque()

        desc.clientCallbacks = clientProxy
        desc.debugFlags = 0;
        desc.logMaxSizeKB = 65536;
        desc.logMaxAgeDays = 30;
        let nsstring = cLogger.appLogDirURL.path as NSString
        if let cstr = nsstring.utf8String {
            let _ = withUnsafeMutablePointer(to: &desc.appOutputPath) {
                $0.withMemoryRebound(to: Int8.self, capacity: Int(CXR_MAX_PATH-1)) { ptr in
                    strncpy( ptr, cstr, Int(CXR_MAX_PATH-1))
                }
            }
        }
        
        // Update these values to adapt QoS to different network configurations
        var connectionDesc: cxrConnectionDesc = cxrConnectionDesc()
        connectionDesc.async = UInt8(cxrTrue);
        connectionDesc.useL4S = UInt8(cxrFalse); // TODO: this will want to be set as appropriate for the network.
        connectionDesc.clientNetwork = cxrNetworkInterface_Unknown;
        connectionDesc.topology = cxrNetworkTopology_LAN;

        let r = cxrCreateReceiver(&desc, receiverHandlePtr)
        if r != cxrError_Success {
            cLogger.logMsg(level: cxrLL_Critical, tag: LOG_TAG, msg: (String(format: "Failed to create reciever with error: %s", cxrErrorString(r))))
            return false
        }
        
        // Convert from Swift String object all the way to a pointer-based C string
        let addrBytes = self.address.data(using: String.Encoding.ascii, allowLossyConversion:false)!
        let bufSize = addrBytes.count + 1
        let buf = UnsafeMutablePointer<Int8>.allocate(capacity: bufSize)
        defer {
            buf.deallocate()
        }
        buf.assign(repeating: 0, count: bufSize)
        addrBytes.withUnsafeBytes( {
                let bytes = $0.baseAddress!.assumingMemoryBound(to: Int8.self)
                        buf.assign(from: bytes, count: addrBytes.count)
                    } )
        
        let e = cxrConnect(self.receiver!, UnsafePointer(buf), &connectionDesc)
        if e != cxrError_Success {
            cLogger.logMsg(level: cxrLL_Critical, tag: LOG_TAG, msg: (String(format: "Failed to connect to [%s] with error: %s", UnsafePointer(buf), cxrErrorString(r))))
            return false
        }
       
        return true;
    }
    
    // MARK: Render Thread & UI
    
    func draw(in view: MTKView) {
        
        if (clientState != cxrClientState_StreamingSessionInProgress) {
            self.computeElapsedTime(TimerIndex.TIMER_METAL)
            return;
        }
        
        var e:cxrError = cxrError_Failed;
        var frameLatched:Bool = false;
        
        e = cxrLatchFrame(receiver, latchedFrame, 0xFFFFFFFF, 0);
        switch e {

        case cxrError_Success:
            if(latchedFrame!.pointee.count > 0)
            {
                framesLatched+=1;
                frameLatched = true;
            
                if let cxrPixbuf = latchedFrame!.pointee.frames.0.texture, let cxrPixbufAlpha = latchedFrame!.pointee.frames.1.texture {
                    // RGB texture (left eye)
                    let cvPixBuf = Unmanaged<CVPixelBuffer>.fromOpaque(cxrPixbuf).takeUnretainedValue()
                    let cxrTexY : CVMetalTexture? = createTexture(fromPixelBuffer: cvPixBuf, pixelFormat: .r8Unorm, planeIndex: 0)
                    let cxrTexCbCr : CVMetalTexture? = createTexture(fromPixelBuffer: cvPixBuf, pixelFormat: .rg8Unorm, planeIndex: 1)
                    latchedTextureY = CVMetalTextureGetTexture(cxrTexY!)
                    latchedTextureCbCr = CVMetalTextureGetTexture(cxrTexCbCr!)
                            
                    // Alpha texture (right eye)
                    let cvPixBufA = Unmanaged<CVPixelBuffer>.fromOpaque(cxrPixbufAlpha).takeUnretainedValue()
                    let cxrTexAlphaY : CVMetalTexture? = createTexture(fromPixelBuffer: cvPixBufA, pixelFormat: .r8Unorm, planeIndex: 0)
                    let cxrTexAlphaCbCr : CVMetalTexture? = createTexture(fromPixelBuffer: cvPixBufA, pixelFormat: .rg8Unorm, planeIndex: 1)
                    latchedTextureAlphaY = CVMetalTextureGetTexture(cxrTexAlphaY!)
                    latchedTextureAlphaCbCr = CVMetalTextureGetTexture(cxrTexAlphaCbCr!)
                    
                    let c0 = simd_make_float4(latchedFrame!.pointee.poseMatrix.m.0.0,
                                              latchedFrame!.pointee.poseMatrix.m.0.1,
                                              latchedFrame!.pointee.poseMatrix.m.0.2,
                                              latchedFrame!.pointee.poseMatrix.m.0.3)
                    
                    let c1 = simd_make_float4(latchedFrame!.pointee.poseMatrix.m.1.0,
                                              latchedFrame!.pointee.poseMatrix.m.1.1,
                                              latchedFrame!.pointee.poseMatrix.m.1.2,
                                              latchedFrame!.pointee.poseMatrix.m.1.3)
                    
                    let c2 = simd_make_float4(latchedFrame!.pointee.poseMatrix.m.2.0,
                                              latchedFrame!.pointee.poseMatrix.m.2.1,
                                              latchedFrame!.pointee.poseMatrix.m.2.2,
                                              latchedFrame!.pointee.poseMatrix.m.2.3)
                    
                    // Pose
                    latchedPose = simd_matrix(c0,c1,c2)
                }
            }
            break;
        case cxrError_Frame_Not_Ready:
            if (planeSelected) {
                // short circuit, reuse previous frame
                computeElapsedTime(TimerIndex.TIMER_METAL)
                return;
            }
            break;
        default:
            cLogger.logMsg(level: cxrLL_Critical, tag: LOG_TAG, msg: (String(format: "Failed to latch a frame with error: %s", cxrErrorString(e))))
            self.dismiss(animated: true)
            break;
        }
        
        let commandBuf = commandQ?.makeCommandBuffer()
        let onscreenDescriptor = view.currentRenderPassDescriptor!
        let encoder = commandBuf!.makeRenderCommandEncoder(descriptor: onscreenDescriptor)
        encoder?.setVertexBuffer(imagePlaneVertexBuffer, offset: 0, index: 0);
        if planeSelected {
            // Try to find the ARKit camera image pose that matches the latched (decoded) frame
            var startIdx = cameraImageCacheReadIndex + 1;
            if !(startIdx < cameraImageCacheSize) {
                startIdx = 0;
            }
            
            cameraImageCacheReadIndex = determineArrayOffset(latchedPose, startIdx)
            
            // Handle case where no match between decoded frames and arkit frames was found
            if cameraImageCacheReadIndex == -1 {
                cameraImageCacheReadIndex = cameraImageCacheWriteIndex - 1;
                if cameraImageCacheReadIndex < 0 {
                    cameraImageCacheReadIndex = cameraImageCacheSize - 1;
                }
            }
            
            // set up shader bindings
            encoder!.setRenderPipelineState(psoCompositeXR!)
            encoder!.setFragmentTexture(cameraTextureCacheY[cameraImageCacheReadIndex], index:0)
            encoder!.setFragmentTexture(cameraTextureCacheCbCr[cameraImageCacheReadIndex], index:1)
            encoder!.setFragmentTexture(latchedTextureY, index:2)
            encoder!.setFragmentTexture(latchedTextureCbCr, index:3)
            encoder!.setFragmentTexture(latchedTextureAlphaY, index:4)
            encoder!.setFragmentTexture(latchedTextureAlphaCbCr, index:5)
            
        } else {
            // No plane selected, so return the latched frame and just display the camera feed (no composition with CloudXR feed)
            if frameLatched {
                cxrReleaseFrame(receiver, latchedFrame);
                computeElapsedTime(TimerIndex.TIMER_CXR)
                frameLatched = false;
            }
            
            // Use the most recent camera image
            var writeIdx = cameraImageCacheWriteIndex-1;
            if writeIdx < 0 {
                writeIdx = cameraImageCacheSize-1;
            }
            
            // setup shader bindings
            encoder!.setRenderPipelineState(psoCameraOnly!)
            encoder!.setFragmentTexture(cameraTextureCacheY[writeIdx], index:0)
            encoder!.setFragmentTexture(cameraTextureCacheCbCr[writeIdx], index:1)
        }
        
        encoder!.drawPrimitives(type:MTLPrimitiveType.triangle, vertexStart:0, vertexCount:6)
      
        // Draw the ground planes as a grid
        if !planeSelected {
            encoder!.setRenderPipelineState(psoVisTris!)
            encoder!.setVertexBuffer(constantBuffer, offset:0, index:0)
            encoder!.setVertexBuffer(triVertBuffer, offset:0, index:1)
            encoder!.setFragmentTexture(gridTex, index:0)
            encoder!.drawPrimitives(type:MTLPrimitiveType.triangle, vertexStart:0, vertexCount:Int(triVertCount))
        }
        
        //
        encoder!.endEncoding()
        commandBuf!.present(mtlView.currentDrawable!)
        commandBuf!.commit()
        commandBuf!.waitUntilCompleted()
        
        if (frameLatched) {
            cxrReleaseFrame(receiver, latchedFrame)
            computeElapsedTime(TimerIndex.TIMER_CXR)
            framesUntilStats-=1;
        }
     
        updateUILabels()
        computeElapsedTime(TimerIndex.TIMER_METAL)
       
    }
    
    func updateUILabels() {
        
        framesLatchedLabel.text = "Frames Latched: \(framesLatched)"
        
        var averageTime:Float  = 0.0;
        
        if (timerIndex[TimerIndex.TIMER_METAL.rawValue] == maxTimeSamples-1) {
            for i in 0...maxTimeSamples-1 {
                averageTime += Float(timerSamples[TimerIndex.TIMER_METAL.rawValue][i]);
            }
            averageTime /= Float(maxTimeSamples);
            self.metalFPSLabel.text = String.init(format:"Metal FPS: %3.4f", 1.0 / averageTime)
        }
        
        if (timerIndex[TimerIndex.TIMER_CXR.rawValue] == maxTimeSamples - 1) {
            for i in 0...maxTimeSamples-1 {
                averageTime += Float(timerSamples[TimerIndex.TIMER_CXR.rawValue][i])
            }
            averageTime /= Float(maxTimeSamples);
            self.cxrFPSLabel.text = String.init(format: "CXR FPS: %3.4f", 1.0/averageTime)
        }
        
        if (timerIndex[TimerIndex.TIMER_ARKIT.rawValue] == maxTimeSamples-1) {
            for i in 0...maxTimeSamples-1 {
                averageTime += Float(timerSamples[TimerIndex.TIMER_ARKIT.rawValue][i])
            }
            averageTime /= Float(maxTimeSamples)
            self.arkitFPSLabel.text = String.init(format: "ARKit FPS: %3.4f", 1.0/averageTime)
        }
        
        if (planeSelected) {
            let latencyInFrames:Int = (cameraImageCacheReadIndex >= cameraImageCacheWriteIndex) ?
                                  (cameraImageCacheSize - cameraImageCacheReadIndex + cameraImageCacheWriteIndex) :
                                  (cameraImageCacheWriteIndex - cameraImageCacheReadIndex)
            
            self.framesBehindLabel.text = String.init(format:"Frames behind: %i", latencyInFrames)
        }
        else
        {
            statusLabel.text = "Searching for planes...";
        }
        
        // Update connection stats every 3 seconds
        let kStatsIntervalSec = 3
        if (framesUntilStats <= 0 &&
            cxrGetConnectionStats(receiver, stats) == cxrError_Success)
        {
            // Capture the key connection statistics
            var statsString = String.init(format:"Bitrate (kbps): %5d    Latency (ms): %3d    ",
                                          stats!.pointee.bandwidthUtilizationKbps,
                                          stats!.pointee.roundTripDelayMs)
            
            // Turn the connection quality into a visual representation along the lines of a signal strength bar
            statsString += ("Connection quality: [" +
                (stats!.pointee.quality.rawValue >= cxrConnectionQuality_Bad.rawValue  ? "#" : "_") +
                (stats!.pointee.quality.rawValue >= cxrConnectionQuality_Poor.rawValue ? "#" : "_") +
                (stats!.pointee.quality.rawValue >= cxrConnectionQuality_Fair.rawValue ? "#" : "_") +
                (stats!.pointee.quality.rawValue >= cxrConnectionQuality_Good.rawValue ? "#" : "_") +
                (stats!.pointee.quality == cxrConnectionQuality_Excellent ? "#" : "_") +
                "]    ")
            
            // There could be multiple reasons for low quality however we log only the most impactful to the end user here
            if (stats!.pointee.quality.rawValue <= cxrConnectionQuality_Fair.rawValue)
            {
                if (stats!.pointee.qualityReasons == cxrConnectionQualityReason_EstimatingQuality.rawValue)
                {
                    statsString += "Reason: Estimating quality"
                }
                else if (stats!.pointee.qualityReasons & cxrConnectionQualityReason_HighLatency.rawValue != 0)
                {
                    statsString += String.init(format:"Reason: High Latency (ms): %3d", stats!.pointee.roundTripDelayMs)
                }
                else if (stats!.pointee.qualityReasons & cxrConnectionQualityReason_LowBandwidth.rawValue != 0)
                {
                    statsString += String.init(format:"Reason: Low Bandwidth (kbps): %5d", stats!.pointee.bandwidthAvailableKbps)
                }
                else if (stats!.pointee.qualityReasons & cxrConnectionQualityReason_HighPacketLoss.rawValue != 0)
                {
                    if (stats!.pointee.totalPacketsLost == 0)
                    {
                        statsString += "Reason: High Packet Loss (Recoverable)";
                    }
                    else
                    {
                        statsString += String.init(format:"Reason: High Packet Loss (%%): %3.1f",
                                                   100.0 * Double(stats!.pointee.totalPacketsLost) /
                                                   Double(stats!.pointee.totalPacketsReceived))
                    }
                }
            }
            
            statusLabel.text = statsString;
            cLogger.logMsg(level: cxrLL_Debug, tag: LOG_TAG, msg: statsString)
            framesUntilStats = Int(stats!.pointee.framesPerSecond) * kStatsIntervalSec;
        }
     
    }
    
    // The pose of each ARKit camera image can be thought of as an ID, and the ID is
    // used to correlate the ARKit images with the CXR images for composition.
    func determineArrayOffset(_ poseFromLatchedFrame:simd_float3x4,_ queueIndex:Int) -> Int {
        var curIdx = queueIndex;
        for _ in 0...cameraImageCacheSize-1 {
            // use "almost equal" since network roundtrip we go from matrix to vec/quat and then back to matrix
            if (simd_almost_equal_elements(cameraPoseMatrixCache[curIdx]!, poseFromLatchedFrame, 0.001)) {
                return curIdx;
            }
            curIdx = (curIdx + 1) % cameraImageCacheSize;
        }
        return -1;
    }
    
    func mtkView(_ view: MTKView, drawableSizeWillChange size: CGSize) {
    }
    
    func createTexture(fromPixelBuffer pixelBuffer: CVPixelBuffer, pixelFormat: MTLPixelFormat, planeIndex: Int) -> CVMetalTexture? {
        let width = CVPixelBufferGetWidthOfPlane(pixelBuffer, planeIndex)
        let height = CVPixelBufferGetHeightOfPlane(pixelBuffer, planeIndex)
        
        var texture: CVMetalTexture? = nil
        let status = CVMetalTextureCacheCreateTextureFromImage(nil, textureCache!, pixelBuffer, nil, pixelFormat, width, height, planeIndex, &texture)
        
        if status != kCVReturnSuccess {
            texture = nil
        }
        
        return texture
    }
    
    // MARK: ARKit
    
    func cachePoseMatrix(_ currentPoseMatrix:simd_float4x4) {
        // Transpose since the cached value will be compared to the pose
        // receieved back from the server (row-major layout, e.g. d3d)
        
        cameraPoseMatrixCache[cameraImageCacheWriteIndex]!.columns.0.x = currentPoseMatrix.columns.0.x;
        cameraPoseMatrixCache[cameraImageCacheWriteIndex]!.columns.0.y = currentPoseMatrix.columns.1.x;
        cameraPoseMatrixCache[cameraImageCacheWriteIndex]!.columns.0.z = currentPoseMatrix.columns.2.x;
        cameraPoseMatrixCache[cameraImageCacheWriteIndex]!.columns.0.w = currentPoseMatrix.columns.3.x;
        
        cameraPoseMatrixCache[cameraImageCacheWriteIndex]!.columns.1.x = currentPoseMatrix.columns.0.y;
        cameraPoseMatrixCache[cameraImageCacheWriteIndex]!.columns.1.y = currentPoseMatrix.columns.1.y;
        cameraPoseMatrixCache[cameraImageCacheWriteIndex]!.columns.1.z = currentPoseMatrix.columns.2.y;
        cameraPoseMatrixCache[cameraImageCacheWriteIndex]!.columns.1.w = currentPoseMatrix.columns.3.y;
        
        cameraPoseMatrixCache[cameraImageCacheWriteIndex]!.columns.2.x = currentPoseMatrix.columns.0.z;
        cameraPoseMatrixCache[cameraImageCacheWriteIndex]!.columns.2.y = currentPoseMatrix.columns.1.z;
        cameraPoseMatrixCache[cameraImageCacheWriteIndex]!.columns.2.z = currentPoseMatrix.columns.2.z;
        cameraPoseMatrixCache[cameraImageCacheWriteIndex]!.columns.2.w = currentPoseMatrix.columns.3.z;
    }
    
    func session(_ session: ARSession, didUpdate frame: ARFrame) {
        
        let matView:simd_float4x4  = frame.camera.viewMatrix(for: UIInterfaceOrientation.landscapeRight)
        let matProj:simd_float4x4  = frame.camera.projectionMatrix(for: UIInterfaceOrientation.landscapeRight, viewportSize: self.mtlView.bounds.size, zNear: 0.01, zFar: 1000.0)

        if (clientState == cxrClientState_ReadyToConnect) {
            if(!connectToServer(matProj)) {
                self.dismiss(animated: true)
                return
            }
        }
        else if (clientState == cxrClientState_ConnectionAttemptInProgress) {
            // Wait for the connection to occur
            statusLabel.text = "Connecting...";
            connAttemptCount += 1
            if connAttemptCount % 60 == 0 {
                cLogger.logMsg(level: cxrLL_Info, tag: LOG_TAG, msg: "Waiting for server connection...")
            }
            return
        }
        else if (clientState == cxrClientState_ConnectionAttemptFailed) {
            cLogger.logMsg(level: cxrLL_Error, tag: LOG_TAG, msg: String(format: "Failed to connect with error: %s", cxrErrorString(clientError!)))
            clientState = cxrClientState_Exiting
            self.dismiss(animated: true)
            return
        }
        else if (clientState == cxrClientState_Disconnected)
        {
            cLogger.logMsg(level: cxrLL_Error, tag: LOG_TAG, msg: String(format: "Disconnected with error: %s", cxrErrorString(clientError!)))
            clientState = cxrClientState_Exiting
            self.dismiss(animated: true)
            return
        }
        else if (clientState == cxrClientState_Exiting)
        {
            cLogger.logMsg(level: cxrLL_Info, tag: LOG_TAG, msg: "Client is exiting...")
            // Wait for this view to be dismissed
            statusLabel.text = "Exiting...";
            return
        }

        updateImagePlaneWithFrame(frame: frame)
        lastTransform = frame.camera.transform;

        // Apply translation to the point the user selected
        lastTransform.columns.3.x -= originPos.x
        lastTransform.columns.3.y -= originPos.y
        lastTransform.columns.3.z -= originPos.z
        
        // Apply rotation from UI slider value
        let radians = rotationAngleDegrees * (Float.pi/180.0);
        let y_axis:simd_float3  = simd_make_float3(0.0, 1.0, 0.0);
        let quat:simd_quatf  = simd_quaternion(radians, y_axis);
        let yRotationMatrix:simd_float4x4  = simd_matrix4x4(quat);
        lastTransform = simd_mul(yRotationMatrix, lastTransform);

        // Transpose the pose matrix from column major (ARKit) to row major (CloudXR)
        let transposed:simd_float4x4  = simd_transpose(lastTransform);
        var mat:cxrMatrix34 = cxrMatrix34()
        mat.m.0.0 = transposed.columns.0.x
        mat.m.0.1 = transposed.columns.0.y
        mat.m.0.2 = transposed.columns.0.z
        mat.m.0.3 = transposed.columns.0.w
        
        mat.m.1.0 = transposed.columns.1.x
        mat.m.1.1 = transposed.columns.1.y
        mat.m.1.2 = transposed.columns.1.z
        mat.m.1.3 = transposed.columns.1.w
        
        mat.m.2.0 = transposed.columns.2.x
        mat.m.2.1 = transposed.columns.2.y
        mat.m.2.2 = transposed.columns.2.z
        mat.m.2.3 = transposed.columns.2.w

        let decomposed = MatrixToVecQuat(mat: mat)
        trackingState.hmd.pose.position = decomposed.0
        trackingState.hmd.pose.rotation = decomposed.1
   
        trackingState.hmd.pose.deviceIsConnected = UInt8(cxrTrue)
        trackingState.hmd.pose.trackingResult = cxrTrackingResult_Running_OK;
        trackingState.hmd.pose.poseIsValid = UInt8(cxrTrue)
        trackingState.hmd.activityLevel = cxrDeviceActivityLevel_UserInteraction;
        trackingState.poseTimeOffset = 0.042;
        trackingState.hmd.flags = UInt64(cxrHmdTrackingFlags_HasProjection.rawValue) | // enable dynamic projection matrix updates
                                  UInt64(cxrHmdTrackingFlags_HasScaling.rawValue)
        
        let projTanHalfAngles = projectionMatrixToTangentHalfAngles(matProj)
        
        // only need to update the left eye values for iOS client
        trackingState.hmd.proj.0.0 = projTanHalfAngles.0 // l
        trackingState.hmd.proj.0.1 = projTanHalfAngles.1 // r
        trackingState.hmd.proj.0.2 = projTanHalfAngles.2 // t
        trackingState.hmd.proj.0.3 = projTanHalfAngles.3 // b

        // Send scale from the UI slider value to the server
        trackingState.hmd.scaling = self.scale
                
        let err = cxrSendPose(receiver, &trackingState);
        if err != cxrError_Success
        {
            // Receiver
            cLogger.logMsg(level: cxrLL_Error, tag: LOG_TAG, msg: String(format: "Failed to send pose with error: %s", cxrErrorString(err)))
            
            if err == cxrError_Pose_Callback_Provided
            {
                cLogger.logMsg(level: cxrLL_Info, tag: LOG_TAG, msg: "Check that receiverDesc.clientCallbacks.GetTrackingState is NOT defined")
            }
        }
        
        // Update the data for the draw call
        let rawPointerConstantBuffer = constantBuffer?.contents()
        guard let shaderDataValues = rawPointerConstantBuffer?.bindMemory(
            to: shader_data.self,
            capacity: 1
        ) else {
            return
        }
        
        shaderDataValues.pointee.projectionMatrix = matProj
        shaderDataValues.pointee.viewMatrix = matView
        
        // Ray-plane intersection based on finger tap point
        if((tapPoint!.x != 0) && (tapPoint!.y != 0)) {
            let  start = simd_make_float4(frame.camera.transform.columns.3.x,
                                          frame.camera.transform.columns.3.y,
                                          frame.camera.transform.columns.3.z,
                                          1.0)
            
            let rawPointerLineVertBuffer = lineVertBuffer?.contents()
            guard let lineVertBufferContents = rawPointerLineVertBuffer?.bindMemory(
                to: simd_float4.self,
                capacity: 64
            ) else {
                return
            }
            
            var candidate = simd_make_float3(Float.greatestFiniteMagnitude)
            for anchor in seenAnchors {
                let p = anchor as! ARPlaneAnchor
                // Use ARKit api to unproject
                let unprojPoint = frame.camera.unprojectPoint(tapPoint!,
                                                              ontoPlane: p.transform,
                                                              orientation: UIInterfaceOrientation.landscapeRight,
                                                              viewportSize: self.mtlView.bounds.size)!

                let planeCenter  = simd_mul(p.transform, simd_make_float4(p.center, 1));
                let startFloat3 = simd_make_float3(start.x,start.y,start.z)
                let ray:line_seg  = line_seg.init(start: startFloat3, end: unprojPoint)
                let planeNormal  = simd_make_float3(0, 1, 0);
                let planeCenterFloat3 = simd_make_float3(planeCenter.x,planeCenter.y,planeCenter.z)
                let intersection = intersect(ray, planeCenterFloat3, planeNormal);
                lineVertBufferContents[Int(lineVertCount)]  = simd_make_float4(intersection, 1);
                lineVertCount+=1
                lineVertBufferContents[Int(lineVertCount)]  = simd_make_float4(intersection + planeNormal, 1);
                lineVertCount+=1
                let isCloser = distance(startFloat3, intersection) < distance(startFloat3, candidate);
                let isInRect = pointInPlaneRect(p, intersection);
                if(isCloser && isInRect) {
                    candidate = intersection;
                }
            }
            
            tapPoint?.x = 0.0
            tapPoint?.y = 0.0
            
            if candidate.x != Float.greatestFiniteMagnitude {
                planeSelected = true;
                trackingConfig!.planeDetection =  []
                session.run(trackingConfig!)
                originPos = candidate;
                lineVertBufferContents[Int(lineVertCount)] = start;
                lineVertCount+=1
                lineVertBufferContents[Int(lineVertCount)] = simd_make_float4(candidate, 1);
                lineVertCount+=1
                let originString = String.init(format: "Origin: [%f, %f, %f]", originPos.x, originPos.y, originPos.z);
                statusLabel.text = originString;
                cLogger.logMsg(level: cxrLL_Info, tag: LOG_TAG, msg: originString)
            }
        }
        
        // Convert camera image from ARKit into a Metal texture
        let cvTexY = createTexture(fromPixelBuffer: frame.capturedImage, pixelFormat: MTLPixelFormat.r8Unorm, planeIndex:  0);
        let cvTexCbCr = createTexture(fromPixelBuffer: frame.capturedImage, pixelFormat: MTLPixelFormat.rg8Unorm, planeIndex:1);
        cameraTextureY = CVMetalTextureGetTexture(cvTexY!)
        cameraTextureCbCr = CVMetalTextureGetTexture(cvTexCbCr!);

        // Cache the camera texture and current pose
        if (cameraTextureCacheY[cameraImageCacheWriteIndex] == nil ||
                cameraTextureY!.width != cameraTextureCacheY[cameraImageCacheWriteIndex]!.width ||
                cameraTextureY!.height != cameraTextureCacheY[cameraImageCacheWriteIndex]!.height) {
            
            let texDescY:MTLTextureDescriptor = MTLTextureDescriptor.texture2DDescriptor(pixelFormat: MTLPixelFormat.r8Unorm,
                                                                                         width: cameraTextureY!.width,
                                                                                         height: cameraTextureY!.height,
                                                                                         mipmapped: false)
            texDescY.storageMode = MTLStorageMode.private
            texDescY.usage = MTLTextureUsage(rawValue:MTLTextureUsage.shaderRead.rawValue | MTLTextureUsage.shaderWrite.rawValue | MTLTextureUsage.renderTarget.rawValue | MTLTextureUsage.pixelFormatView.rawValue)
            cameraTextureCacheY[cameraImageCacheWriteIndex] = device?.makeTexture(descriptor: texDescY)
        }
        
        if cameraTextureCacheCbCr[cameraImageCacheWriteIndex] == nil ||
           cameraTextureCbCr!.width != cameraTextureCacheCbCr[cameraImageCacheWriteIndex]!.width ||
           cameraTextureCbCr!.height != cameraTextureCacheCbCr[cameraImageCacheWriteIndex]!.height  {
            
            let texDescCbCr:MTLTextureDescriptor =  MTLTextureDescriptor.texture2DDescriptor(pixelFormat: MTLPixelFormat.rg8Unorm, width: cameraTextureCbCr!.width, height: cameraTextureCbCr!.height, mipmapped: false)
            texDescCbCr.usage = MTLTextureUsage(rawValue:MTLTextureUsage.shaderRead.rawValue | MTLTextureUsage.shaderWrite.rawValue | MTLTextureUsage.renderTarget.rawValue | MTLTextureUsage.pixelFormatView.rawValue)
            
            texDescCbCr.storageMode = MTLStorageMode.private;
            
            cameraTextureCacheCbCr[cameraImageCacheWriteIndex] = device?.makeTexture(descriptor: texDescCbCr)
        }
        
        let commandBuf     = commandQ?.makeCommandBuffer()
        let commandEncoder = commandBuf?.makeComputeCommandEncoder()
        
        // The ARKit images will be copied into the cache via a compute shader
        computeThreadgroupSize = MTLSizeMake(16, 16, 1);
        computeThreadgroupCount!.width  = (cameraTextureCacheY[cameraImageCacheWriteIndex]!.width  + computeThreadgroupSize!.width -  1) / computeThreadgroupSize!.width
        computeThreadgroupCount!.height = (cameraTextureCacheY[cameraImageCacheWriteIndex]!.height + computeThreadgroupSize!.height - 1) / computeThreadgroupSize!.height;
        computeThreadgroupCount!.depth  = 1;
        
        commandEncoder!.setComputePipelineState(computePipelineState!)
        // Blit Y (luma)
        commandEncoder!.setTexture(cameraTextureY, index: 0)
        commandEncoder!.setTexture(cameraTextureCacheY[cameraImageCacheWriteIndex],index:1)
        commandEncoder!.dispatchThreadgroups(computeThreadgroupCount!, threadsPerThreadgroup: computeThreadgroupSize!)
        
        // Blit CbCr (chroma)
        computeThreadgroupCount!.width  = (cameraTextureCacheCbCr[cameraImageCacheWriteIndex]!.width  + computeThreadgroupSize!.width -  1) / computeThreadgroupSize!.width;
        computeThreadgroupCount!.height = (cameraTextureCacheCbCr[cameraImageCacheWriteIndex]!.height + computeThreadgroupSize!.height - 1) / computeThreadgroupSize!.height;
        
        commandEncoder!.setTexture(cameraTextureCbCr, index: 0)
        commandEncoder!.setTexture(cameraTextureCacheCbCr[cameraImageCacheWriteIndex], index: 1)
        commandEncoder!.dispatchThreadgroups(computeThreadgroupCount!, threadsPerThreadgroup: computeThreadgroupSize!)
        
        commandEncoder!.endEncoding()
        commandBuf!.commit()
        cameraTextureY = nil;
        cameraTextureCbCr = nil;
        cachePoseMatrix(lastTransform) // Note: a transpose happens in this function
        
        cameraImageCacheIndexLock!.lock()
        cameraImageCacheWriteIndex = (cameraImageCacheWriteIndex + 1) % cameraImageCacheSize;
        cameraImageCacheIndexLock!.unlock();
        
        computeElapsedTime(TimerIndex.TIMER_ARKIT)
    }
    
    func session(_ session: ARSession, didAdd anchors: [ARAnchor]) {
        for  i in 0...anchors.count-1 {
            if(!(anchors[i] is ARPlaneAnchor)) {
                continue;
            }
            let p = anchors[i] as! ARPlaneAnchor
            seenAnchors.append(p)
            let t:simd_float4x4  = p.transform;
            let c:simd_float3    = p.center;
            let e:simd_float3    = p.extent / 2.0;
            var i:simd_float4    = simd_make_float4(c.x + e.x, c.y, c.z + e.z, 1);
            var j:simd_float4    = simd_make_float4(c.x - e.x, c.y, c.z + e.z, 1);
            var k:simd_float4    = simd_make_float4(c.x - e.x, c.y, c.z - e.z, 1);
            var l:simd_float4    = simd_make_float4(c.x + e.x, c.y, c.z - e.z, 1);
            i = simd_mul(t, i);
            j = simd_mul(t, j);
            k = simd_mul(t, k);
            l = simd_mul(t, l);
            
    #if PLANE_VIS_LINES
            var rawPointer = lineVertBuffer?.contents()
            guard let values = rawPointer?.bindMemory(
                to: Float.self,
                capacity: lineVertBuffer.count
            ) else { return }

            lineVertBufferContents[Int(lineVertCount)]  = i
            lineVertCount+=1
    
            lineVertBufferContents[Int(lineVertCount)]  = j
            lineVertCount+=1
            
            lineVertBufferContents[Int(lineVertCount)]  = j
            lineVertCount+=1
            
            lineVertBufferContents[Int(lineVertCount)]  = k
            lineVertCount+=1
    
            lineVertBufferContents[Int(lineVertCount)]  = k
            lineVertCount+=1
            
            lineVertBufferContents[Int(lineVertCount)]  = l
            lineVertCount+=1
            
            lineVertBufferContents[Int(lineVertCount)]  = l
            lineVertCount+=1
            
            lineVertBufferContents[Int(lineVertCount)]  = i
            lineVertCount+=1
    
    #endif
            // Add the plane geometry (two triangles) to the GPU buffer
            
            let rawPointerTriVertBuffer = triVertBuffer?.contents()
            guard let rawPointerTriVertBufferContents = rawPointerTriVertBuffer?.bindMemory(
                to: simd_float4.self,
                capacity: triVertBuffer!.length
            ) else { return }
            
            var nextVert:Int = Int(triVertCount);
            rawPointerTriVertBufferContents[nextVert] = i
            nextVert+=1
            
            rawPointerTriVertBufferContents[nextVert] = j
            nextVert+=1
            
            rawPointerTriVertBufferContents[nextVert] = k
            nextVert+=1
            
            rawPointerTriVertBufferContents[nextVert] = i
            nextVert+=1
            
            rawPointerTriVertBufferContents[nextVert] = k
            nextVert+=1
            
            rawPointerTriVertBufferContents[nextVert] = l
            nextVert+=1

            triVertCount += 6;
        }
    }
    
    // MARK: Gestures/UI
    
    @objc func handleTap(_ sender: UITapGestureRecognizer) {
        tapPoint = sender.location(in: self.mtlView!)
    }
    
    @objc private func handleLongPress(_ sender: UIGestureRecognizer) {
        if (sender.state == UIGestureRecognizer.State.began) {
            planeSelected = false;
            originPos = simd_make_float3(0, 0, 0);
            lineVertCount = 0;
        }
    }
    
    @IBAction func rotationSliderChanged(_ sender: UISlider) {
        self.rotationAngleDegrees = sender.value
        self.rotationLabel.text = String.init(format: "Rotation: %1f", sender.value)
    }
    
    @IBAction func scaleSliderChanged(_ sender: UISlider) {
        self.scale = sender.value
        self.scaleLabel.text = String.init(format: "Scale: %.1f", sender.value)
    }
    
    // MARK: Misc

    // This function is used to calculate the FPS of each of the different
    // components (CloudXR, Metal, ARKit). Only a rough estimate.
    func computeElapsedTime(_ index:TimerIndex) {
        if(timerFirstTime[index.rawValue]) {
            startTimes[index.rawValue] = NSDate()
            timerFirstTime[index.rawValue] = false;
        } else {
            endTimes[index.rawValue] = NSDate()
            let executionTime:TimeInterval = (endTimes[index.rawValue]?.timeIntervalSince(startTimes[index.rawValue]! as Date))! //timeIntervalSinceDate:startTimes[index.rawValue]];
            timerSamples[index.rawValue][timerIndex[index.rawValue]] = Double(executionTime)
            startTimes[index.rawValue] = endTimes[index.rawValue];
            timerIndex[index.rawValue] = (timerIndex[index.rawValue] + 1) % maxTimeSamples;
        }
    }
        
    func updateImagePlaneWithFrame(frame:ARFrame) {
        // Update the texture coordinates of our image plane to aspect fill the viewport
        let displayToCameraTransform:CGAffineTransform = frame.displayTransform(for: UIInterfaceOrientation.landscapeRight, viewportSize: self.mtlView.bounds.size).inverted()
        
        let rawPointer = imagePlaneVertexBuffer?.contents()
        guard let values = rawPointer?.bindMemory(
            to: Float.self,
            capacity: kImagePlaneVertexData.count
        ) else { return }
 
        for index in 0...5 {
            let textureCoordIndex:Int     = 2 * index;
            let textureCoord:CGPoint      = CGPoint(x: CGFloat(kImagePlaneVertexData[textureCoordIndex]), y:CGFloat(kImagePlaneVertexData[textureCoordIndex + 1]));
            let transformedCoord:CGPoint  = textureCoord.applying(displayToCameraTransform);
            let offsetX = values + textureCoordIndex
            offsetX.pointee = Float(transformedCoord.x)
            let offsetY = values + textureCoordIndex + 1
            offsetY.pointee = Float(transformedCoord.y)
        }
    }
    
    func projectionMatrixToTangentHalfAngles(_ projectionMatrix: simd_float4x4) -> (Float,Float,Float,Float) {
    
        // Extract the 4 tangents of the half-angles from the center view axis
        // of the projection matrix
        var bottom: Float = 0.0;
        var top:    Float = 0.0;
        var left:   Float = 0.0;
        var right:  Float = 0.0;
        
        if (abs((projectionMatrix).columns.2.x) > 0.0001) {
          // Non-symmetric projection
          let oneOver00: Float = 1.0/(projectionMatrix).columns.0.x;
          let l: Float         = -(1.0 - (projectionMatrix).columns.2.x) * oneOver00;
          let r: Float         = 2.0 * oneOver00 + l;

          let oneOver11:Float = 1.0 / (projectionMatrix).columns.1.y;
          let b:Float         = -(1.0 - (projectionMatrix).columns.2.y) * oneOver11;
          let t:Float         = 2.0 * oneOver11 + b;

          left   = l;
          right  = r;
          top    = -t;
          bottom = -b;
        } else {
          // Symmetric projection
          left   = -1.0/(projectionMatrix).columns.0.x;
          right  = -left;
          top    = -1.0/(projectionMatrix).columns.1.y;
          bottom = -top;
        }
        
        return (left,right,top,bottom)
    }
}
