// Copyright NVIDIA Corporation 2012 - 2021
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL
// NVIDIA OR ITS SUPPLIERS BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR
// CONSEQUENTIAL DAMAGES WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR
// LOSS OF BUSINESS PROFITS, BUSINESS INTERRUPTION, LOSS OF BUSINESS
// INFORMATION, OR ANY OTHER PECUNIARY LOSS) ARISING OUT OF THE USE OF OR
// INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS BEEN ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGES.

/// \file Result.h
/// Defines result codes.

#pragma once

#include "Util.h"

#if defined(__cplusplus)
extern "C"
{
#endif

    /// \addtogroup Results Stream Result structs
    /// @{

#define NVST_SUCCESS 0x00000000
#define NVST_ERROR_BITMASK 0x80000000
#define NVST_ERROR_GROUP_START(group) ((group) << 16 | NVST_ERROR_BITMASK)

    // 0               1               2               3
    // 0 1 2 3 4 5 6 7 0 1 2 3 4 5 6 7 0 1 2 3 4 5 6 7 0 1 2 3 4 5 6 7
    // +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    // |E|Reserved Bits|      EG       |      Status Code(SC)          |
    // +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    //
    // Error Bit (E): 1 bit
    //                 If 1 then it's an error, otherwise success.
    //
    // Reserved Bits(RC): 7 bits
    //                 These bits are reserved for possible future expansion.
    //
    // Error Group(EG): 8 bits
    //                 This will include various errors groups like NVSC,
    //                 Decoder, ENET, JNI, JAVA, PGC, RTSP, ENET, etc.
    //                 2^8 is large enough to hold any number of modules that we define.
    //
    // Status Code(SC): 16 bits
    //                 This will include module specific error codes.
    //                 These are defined in their respective module.
    //                 For example, from NVSC these will include disconnect reasons,
    //                 network error codes, connection attempt error codes, etc.

    /// Error groups.
    /// Last 16 bits of the module code should be zero.
    /// THESE MAPPINGS ARE FIXED AND SHOULD NOT BE CHANGED.
    /// IF REQUIRED YOU CAN APPEND NEW ERROR GROUP AFTER THE LAST.
    typedef enum NvstErrorGroups_t
    {
        /// Generic errors unrelated to a particular module.
        NEG_NVST_GENERIC = 0x00,
        // NVSC related errors.
        // This will start with error group 0x01.
        // Reserving 0x01-0x1F for NVSC reasons.
        /// First NVSC error group.
        NEG_NVSC_FIRST = 0x01,
        /// Errors during initial handshake.
        NEG_NVSC_CONNECTION_ATTEMPT = 0x02,
        /// during streaming.
        NEG_NVSC_SERVER_DISCONNECT = 0x03,
        /// Network errors during streaming.
        NEG_NVSC_NETWORK = 0x04,
        /// Audio QoS errors.
        NEG_NVSC_AUDIO_QOS = 0x05,
        /// Video QoS errors.
        NEG_NVSC_VIDEO_QOS = 0x06,
        /// Enet releated errors.
        NEG_NVSC_ENET = 0x07,
        /// RTSP related errors.
        NEG_NVSC_RTSP = 0x08,
        /// RTP related errors.
        NEG_NVSC_RTP = 0x09,
        /// FEC errors.
        NEG_NVSC_FEC = 0x0A,
        /// NVSC generic errors.
        NEG_NVSC_GENERIC = 0x0B,
        /// Errors in the server control module.
        NEG_NVSC_SERVER_CONTROL = 0x0C,
        /// Errors due to remote input.
        NEG_NVSC_REMOTE_INPUT = 0x0D,
        /// Errors due to video frame provider, encoder, adapter, pipeline.
        NEG_NVSC_SERVER_VIDEO_UNINTENDED = 0x0E,
        /// Errors due to audio pipeline on server.
        NEG_NVSC_SERVER_AUDIO_UNINTENDED = 0x0F,
        /// Errors due to audio pipeline on client.
        NEG_NVSC_CLIENT_AUDIO_UNINTENDED = 0x10,
        /// Intended client disconnection reasons.
        NEG_NVSC_CLIENT_INTENDED = 0x11,
        /// Native RTC peer disconnect codes.
        NEG_NATIVE_RTC_PEER_TRANSPORT_DISCONNECT = 0x12,
        /// Native RTC network error codes.
        NEG_NATIVE_RTC_NETWORK_ERROR = 0x13,
        /// Last NVSC error group.
        NEG_NVSC_LAST = 0x1F,

        // Reserved for Gamestream Services
        /// Reserved for services (first).
        NEG_RESERVED_FOR_SERVICES_FIRST = 0x20,
        /// Reserved for services (last).
        NEG_RESERVED_FOR_SERVICES_LAST = 0x2F,

        // Reserved for clients to NVSC.
        /// Reserved for clients to NVSC (first).
        NEG_NVSC_RESERVED_FOR_CLIENT_FIRST = 0x30,
        /// Reserved for client to NVSC (last).
        NEG_NVSC_RESERVED_FOR_CLIENT_LAST = 0x3F,

        // NVSS related errors.
        /// The NVSS group range goes from 0x40 to 0x4F.
        NEG_NVSS_FIRST = 0x40,
        /// Initialization related errors.
        NEG_NVSS_INIT = 0x40,
        /// Placeholder for the next error group.
        NEG_NVSS_PLACEHOLDER = 0x41,
        /// Errors during initial handshake.
        NEG_NVSS_CONNECTION_ATTEMPT = 0x42,
        /// Errors during streaming.
        NEG_NVSS_CLIENT_DISCONNECT = 0x43,
        /// Network errors during streaming.
        NEG_NVSS_NETWORK = 0x44,
        /// Audio QOS Errors.
        NEG_NVSS_AUDIO_QOS = 0x45,
        /// Video QOS Errors.
        NEG_NVSS_VIDEO_QOS = 0x46,
        /// Enet releated errors.
        NEG_NVSS_ENET = 0x47,
        /// RTSP related errors.
        NEG_NVSS_RTSP = 0x48,
        // Please replace NEG_NVSS_PLACEHOLDER (0x41) before using 0x49
        // then delete this comment.
        /// Last NVSS error group.
        NEG_NVSS_LAST = 0x4F,
    } NvstErrorGroups;

    /// Result codes for functions in the NVIDIA Streamer SDK API.
    /// This does not cover every error code that could be produced
    /// \note From Andrija: We have a contract with the rest of GFN to not change existing codes.
    typedef enum NvstResult_t
    {
        // Generic result codes.
        NVST_R_SUCCESS = 0x00000000,
        NVST_R_GENERIC_ERROR = 0x800B0000,
        NVST_R_INVALID_OBJECT = 0x800B0001,
        NVST_R_INVALID_ENUM = 0x800B0002,
        NVST_R_INVALID_VALUE = 0x800B0003,
        NVST_R_INVALID_PARAM = 0x800B0004,
        NVST_R_INVALID_OPERATION = 0x800B0005,
        NVST_R_INVALID_HARDWARE = 0x800B0006,
        NVST_R_NO_IMPLEMENTATION = 0x800B0007,

        // Network protocol specific error codes.
        NVST_R_NETWORK_ERROR = 0x800B0008,
        NVST_R_NETWORK_TIMEOUT = 0x800B0009,
        NVST_R_INCOMPATIBLE_SERVER = 0x800B000A,

        // Initial connection errors.
        NVST_R_ERROR_INVALID_SDP = 0x800B000B,
        NVST_R_ERROR_UNEXPECTED_DISCONNECTION = 0x800B000C,
        NVST_R_ERROR_INVALID_MEDIA_ATTRIBUTES = 0x800B000D,
        NVST_R_ERROR_INVALID_CONTROL_ID = 0x800B000E,
        NVST_R_ERROR_INVALID_STREAM_COUNT = 0x800B000F,
        NVST_R_ERROR_SERVER_CONTROL_ALLOCATION_FAILED = 0x800B0010,
        NVST_R_ERROR_PIPELINE_INITIALIZATION_FAILED = 0x800B0011,
        NVST_R_ERROR_INVALID_STREAM_TYPE = 0x800B0012,
        NVST_R_ERROR_INVALID_SPS_PPS_KEY = 0x800B0013,
        NVST_R_ERROR_INVALID_SPS_PPS_FORMAT = 0x800B0014,
        NVST_R_ERROR_UDP_RTP_SOURCE_CREATE_FAILED = 0x800B0015,
        NVST_R_ERROR_UDP_RTP_SOURCE_OPEN_FAILED_UDP_BUFFER_SIZE = 0x800B0016,
        NVST_R_ERROR_UDP_RTP_SOURCE_CONNECT_FAILED = 0x800B0017,
        NVST_R_ERROR_TRANSPORT_PARSE_FAILED = 0x800B0018,
        /// RTSP STUNing failed.
        NVST_R_ERROR_RTSP_STUN_FAILED = 0x800B0019,

        /// Operation failed because the system was busy handling something else.
        NVST_R_BUSY = 0x800B001A,
        /// Failed to get local IP addresses through network enumerator.
        NVST_R_ERROR_GET_LOCAL_ADDRESS_FAILED = 0x800B001B,
        /// Frame is dropped by ProcessStreamData.
        NVST_R_FRAME_DROPPED = 0x800B001C,
        /// Library and Header Versions differ.
        NVST_R_API_MISMATCH = 0x800B001D,
        // No available local port in the RTP port probe range
        NVST_R_ERROR_UDP_RTP_SOURCE_OPEN_FAILED_NO_PORTS_AVAILABLE = 0x800B001E,
        /// Failed to set local description.
        NVST_R_ERROR_SET_LOCAL_DESCRIPTION_FAILED = 0x800B001F,
        /// Failed to set remote description. Usually caused by bad options being sent by the client.
        NVST_R_ERROR_SET_REMOTE_DESCRIPTION_FAILED = 0x800B0020,
        /// Operation failed due to a security concern.
        NVST_R_SECURITY = 0x800B0021,

        // Additional generic result codes
        /// An operation was attempted that is not valid in the object's current state.
        NVST_R_INVALID_STATE = 0x800B1000,
        /// An expected object is not present.
        NVST_R_NOT_FOUND = 0x800B1001,
        /// An unexpected internal server error.
        NVST_R_INTERNAL_ERROR = 0x800B1002,
        /// A configuration value is invalid.
        NVST_R_INVALID_CONFIGURATION = 0x800B1003,
        // Initial connection creation failed for signaling
        /// Creation of the WebSocket for signaling failed.
        NVST_R_ERROR_UNEXPECTED_DISCONNECTION_INITIAL = 0x800B1004,
        /// Failed to create an object
        NVST_R_CREATION_FAILED = 0x800B1005,

        // Server control failures.
        NVST_R_SERVER_CONTROL_FAILED = 0x800C0000,
        NVST_R_SERVER_CONTROL_CONNECTION_FAILURE = 0x800C0001,
        NVST_R_SERVER_CONTROL_RECEIVE_PACKET_FAILURE = 0x800C0002,
        NVST_R_SERVER_CONTROL_FRAME_INVALIDATION_FAILED = 0x800C0003,
        NVST_R_SERVER_CONTROL_FRAME_RATE_CHANGE_REQUEST_FAILED = 0x800C0004,
        NVST_R_SERVER_CONTROL_QOS_STATS_SEND_FAILED = 0x800C0005,
        NVST_R_SERVER_CONTROL_AUDIO_STATS_SEND_FAILED = 0x800C0006,
        NVST_R_SERVER_CONTROL_LAST_RECEIVED_F_NUM_FAILED = 0x800C0007,
        NVST_R_SERVER_CONTROL_PICTURE_REQUEST_FAILED = 0x800C0008,
        NVST_R_SERVER_CONTROL_STATS_RECORD_NOTIFICATION_FAILED = 0x800C0009,
        NVST_R_SERVER_CONTROL_NETWORK_CAPTURE_NOTIFICATION_FAILED = 0x800C000A,
        NVST_R_SERVER_CONTROL_ETW_NOTIFICATION_FAILED = 0x800C000B,
        NVST_R_SERVER_CONTROL_MOUSE_CAPTURE_FAILED = 0x800C000C,
        NVST_R_SERVER_CONTROL_MIMIC_REMOTE_CURSOR_FAILED = 0x800C000D,
        NVST_R_SERVER_CONTROL_GAME_RECORD_FAILED = 0x800C000E,
        NVST_R_SERVER_CONTROL_MOUSE_SETTING_FAILED = 0x800C000F,
        NVST_R_SERVER_CONTROL_RELAY_RI_FAILED = 0x800C0010,
        NVST_R_SERVER_CONTROL_PING_ACK_FAILED = 0x800C0011,
        NVST_R_SERVER_CONTROL_AUDIO_SURROUND_INFO_FAILED = 0x800C0012,
        NVST_R_SERVER_CONTROL_STUN_FAILURE_DEPRECATED = 0x800C0013,
        NVST_R_SERVER_CONTROL_HOLE_PUNCH_FAILURE = 0x800C0014,
        NVST_R_SERVER_CONTROL_HAPTICS_STATE_CHANGE_FAILED = 0x800C0015,
        NVST_R_SERVER_CONTROL_MAX_BITRATE_CHANGE_FAILED = 0x800C0016,
        NVST_R_SERVER_CONTROL_DRC_STATE_CHANGE_FAILED = 0x800C0017,
        NVST_R_SERVER_CONTROL_DFC_STATE_CHANGE_FAILED = 0x800C0018,
        NVST_R_SERVER_CONTROL_DATA_CHANNEL_CREATION_FAILURE = 0x800C0019,

        /// Passthrough group for SessionTerminationReason due to video, defined in ProtocolDefines.h.
        NVSTE_GROUP_SERVER_TERMINATED_VIDEO = 0x800E0000,
        // Server terminated due to video passthrough error codes.

        /// Client timed-out while waiting for server to setup video pipeline
        NVST_DISCONN_CLIENT_SERVER_VIDEO_SETUP_TIMEOUT = 0x800E8009,
        /// Client timed-out while waiting for server to initialize video adapter
        NVST_DISCONN_CLIENT_SERVER_VIDEO_ADAPTER_INIT_TIMEOUT = 0x800E800A,
        /// Client timed-out while waiting for server to initialize video frame provider
        NVST_DISCONN_CLIENT_SERVER_VIDEO_FRAMEPROVIDER_INIT_TIMEOUT = 0x800E800B,
        /// Client timed-out while waiting for server to initialize video encoder
        NVST_DISCONN_CLIENT_SERVER_VIDEO_ENCODER_INIT_TIMEOUT = 0x800E800C,
        /// Server failed to send the first frame.
        NVST_DISCONN_SERVER_VIDEO_FIRST_FRAME_SEND_FAILED = 0x800E810A,
        /// Server failed to send a frame besides the first.
        NVST_DISCONN_SERVER_VIDEO_NEXT_FRAME_SEND_FAILED = 0x800E810B,

        /// Video Frame Provider Initialization Disconnection Reasons
        NVST_DISCONN_SERVER_VFP_GENERIC = 0x800E8300,
        NVST_DISCONN_SERVER_VFP_INIT_DLL_LOAD_FAILED = 0x800E8301,
        NVST_DISCONN_SERVER_VFP_INIT_ADAPTER_NOT_INITIALIZED = 0x800E8302,
        NVST_DISCONN_SERVER_VFP_INIT_UNSUPPORTED_FORMAT = 0x800E8303,
        NVST_DISCONN_SERVER_VFP_INIT_NOT_READY = 0x800E8304,
        NVST_DISCONN_SERVER_VFP_INIT_NOT_INITIALIZED = 0x800E8305,
        NVST_DISCONN_SERVER_VFP_INIT_NOT_RUNNING = 0x800E8306,
        NVST_DISCONN_SERVER_VFP_INIT_HDR_NOT_SUPPORTED = 0x800E8308,
        NVST_DISCONN_SERVER_VFP_INIT_CREATE_FAILED = 0x800E8309,
        NVST_DISCONN_SERVER_VFP_INIT_NULL_OBJECT = 0x800E830A,
        NVST_DISCONN_SERVER_VFP_INIT_ALREADY_RUNNING = 0x800E830B,
        NVST_DISCONN_SERVER_VFP_INIT_PID_ALREADY_INITIALIZED = 0x800E830C,
        NVST_DISCONN_SERVER_VFP_INIT_PID_CAPTURE_NOT_SUPPORTED = 0x800E830D,
        NVST_DISCONN_SERVER_VFP_INIT_GETSTATUS_FAILURE = 0x800E8312,
        NVST_DISCONN_SERVER_VFP_INIT_CAPTURE_NOT_POSSIBLE = 0x800E8313,
        NVST_DISCONN_SERVER_VFP_INIT_MAX_BUFFERSIZE_UNKNOWN = 0x800E8314,
        NVST_DISCONN_SERVER_VFP_INIT_CUDA_SETUP_PARAMS_FAILED = 0x800E8315,
        NVST_DISCONN_SERVER_VFP_INIT_WAC_ERROR_GENERIC = 0x800E8318,
        NVST_DISCONN_SERVER_VFP_INIT_WAC_ERROR_INVALID_PARAM = 0x800E8319,
        NVST_DISCONN_SERVER_VFP_INIT_WAC_ERROR_CAPTURE_FAILURE = 0x800E831A,
        NVST_DISCONN_SERVER_VFP_INIT_WAC_ERROR_INVALID_CALL = 0x800E831B,
        NVST_DISCONN_SERVER_VFP_INIT_WAC_ERROR_SYSTEM_ERROR = 0x800E831C,
        NVST_DISCONN_SERVER_VFP_INIT_WAC_ERROR_INVALID_WINDOW_HANDLE = 0x800E831D,
        NVST_DISCONN_SERVER_VFP_INIT_WAC_ERROR_WINDOW_Z_ORDER_CHANGED = 0x800E831E,
        NVST_DISCONN_SERVER_VFP_INIT_WAC_ERROR_OPEN_PROCESS_FAILED_ACCESS_DENIED = 0x800E831F,
        NVST_DISCONN_SERVER_VFP_INIT_WAC_ERROR_OPEN_PROCESS_FAILED = 0x800E8320,
        NVST_DISCONN_SERVER_VFP_INIT_WAC_ERROR_MAG_STRATEGY_INTERNAL_POINTERS_FAILURE = 0x800E8321,
        NVST_DISCONN_SERVER_VFP_INIT_WAC_ERROR_MAG_STRATEGY_UPDATE_FAILURE = 0x800E8322,
        NVST_DISCONN_SERVER_VFP_INIT_DDA_ERROR_INVALID_PARAM = 0x800E8323,
        NVST_DISCONN_SERVER_VFP_INIT_DDA_ERROR_SETUP_FAILED = 0x800E8324,

        NVST_DISCONN_SERVER_VIDEO_ENCODER_GENERIC = 0x800E8400,
        NVST_DISCONN_SERVER_VIDEO_ENCODER_INIT_DLL_LOAD_FAILED = 0x800E8401,
        NVST_DISCONN_SERVER_VIDEO_ENCODER_INIT_GET_PROC_ADDRESS_FAILED = 0x800E8402,
        NVST_DISCONN_SERVER_VIDEO_ENCODER_INIT_GET_API_FAILED = 0x800E8403,
        NVST_DISCONN_SERVER_VIDEO_ENCODER_INIT_NOT_READY = 0x800E8404,
        NVST_DISCONN_SERVER_VIDEO_ENCODER_INIT_NOT_SUPPORTED = 0x800E8405,
        NVST_DISCONN_SERVER_VIDEO_ENCODER_INIT_UNSUPPORTED_FORMAT = 0x800E8406,
        NVST_DISCONN_SERVER_VIDEO_ENCODER_INIT_HEVC_NOT_SUPPORTED = 0x800E8407,
        NVST_DISCONN_SERVER_VIDEO_ENCODER_INIT_HDR_NOT_SUPPORTED = 0x800E8408,
        NVST_DISCONN_SERVER_VIDEO_ENCODER_INIT_NOT_INITIALIZED = 0x800E8409,
        NVST_DISCONN_SERVER_VIDEO_ENCODER_INIT_ALREADY_RUNNING = 0x800E840A,
        NVST_DISCONN_SERVER_VIDEO_ENCODER_INIT_NOT_RUNNING = 0x800E840B,
        NVST_DISCONN_SERVER_VIDEO_ENCODER_INIT_CUDA_ENCODE_OPEN_FAILED = 0x800E840C,
        NVST_DISCONN_SERVER_VIDEO_ENCODER_INIT_UPDATE_PARAMS_FAILED = 0x800E840D,
        NVST_DISCONN_SERVER_VIDEO_ENCODER_INIT_ADAPTER_NOT_INITIALIZED = 0x800E840E,
        NVST_DISCONN_SERVER_VIDEO_ENCODER_INIT_FAILED = 0x800E840F,
        NVST_DISCONN_SERVER_VIDEO_ENCODER_INIT_ALLOCATE_IO_FAILED = 0x800E8410,
        NVST_DISCONN_SERVER_VIDEO_ENCODER_INIT_SPS_PPS_GET_FAILED = 0x800E8411,
        NVST_DISCONN_SERVER_VIDEO_ENCODER_INIT_CREATE_FRAME_FAILED = 0x800E8412,
        NVST_DISCONN_SERVER_VIDEO_ENCODER_INIT_REGISTER_RESOURCE_FAILED = 0x800E8414,
        NVST_DISCONN_SERVER_VIDEO_ENCODER_INIT_CREATE_BITSTREAM_BUFFER_FAILED = 0x800E8415,

        NVST_DISCONN_SERVER_VIDEO_ADAPTER_GENERIC = 0x800E8500,
        NVST_DISCONN_SERVER_VIDEO_ADAPTER_NOT_INITIALIZED = 0x800E8501,
        NVST_DISCONN_SERVER_VIDEO_ADAPTER_ALREADY_INITIALIZED = 0x800E8502,
        NVST_DISCONN_SERVER_VIDEO_ADAPTER_CREATE_WINDOW_FAILED = 0x800E8503,
        NVST_DISCONN_SERVER_VIDEO_ADAPTER_INITIALIZE_D3D_FAILED = 0x800E8504,
        NVST_DISCONN_SERVER_VIDEO_ADAPTER_CREATE_D3D_DEVICE_FAILED = 0x800E8505,
        NVST_DISCONN_SERVER_VIDEO_ADAPTER_CREATE_SCHEDULER_FAILED = 0x800E8506,
        NVST_DISCONN_SERVER_VIDEO_ADAPTER_GET_CO_PROC_INFO_FAILED = 0x800E8507,
        NVST_DISCONN_SERVER_VIDEO_ADAPTER_CUDA_ALREADY_INITIALIZED = 0x800E8508,
        NVST_DISCONN_SERVER_VIDEO_ADAPTER_CUDA_CREATE_CONTEXT_FAILED = 0x800E8509,
        NVST_DISCONN_SERVER_VIDEO_ADAPTER_CUDA_INITIALIZE_API_FAILED = 0x800E850A,
        NVST_DISCONN_SERVER_VIDEO_ADAPTER_GET_DRIVER_VERSION_FAILED = 0x800E850B,
        NVST_DISCONN_SERVER_VIDEO_ADAPTER_CREATE_DXVA_FAILED = 0x800E850C,
        NVST_DISCONN_SERVER_VIDEO_ADAPTER_D3D11_ALREADY_INITIALIZED = 0x800E850D,

        /// Server failed to start because the video pipeline was not initialized.
        NVST_DISCONN_SERVER_VIDEO_PIPELINE_NOT_INITIALIZED = 0x800E8600,
        /// Server failed to start because the pipeline was already running.
        NVST_DISCONN_SERVER_VIDEO_PIPELINE_ALREADY_RUNNING = 0x800E8601,
        NVST_DISCONN_SERVER_VIDEO_PIPELINE_FRAME_REC_SETUP_FAILED = 0x800E8602,
        NVST_DISCONN_SERVER_VIDEO_PIPELINE_EXIT_FRAME_REC_SETUP_FAILED = 0x800E8603,
        NVST_DISCONN_SERVER_VIDEO_PIPELINE_ALLOCATE_SCALING_FRAME_FAILED = 0x800E8604,
        NVST_DISCONN_SERVER_VIDEO_PIPELINE_SCALE_FRAME_FAILED = 0x800E8605,
        NVST_DISCONN_SERVER_VIDEO_PIPELINE_PREPARE_BLEND_BACKGROUND_FAILED = 0x800E8606,
        NVST_DISCONN_SERVER_VIDEO_PIPELINE_ALLOCATE_BLEND_FRAME_FAILED = 0x800E8607,
        NVST_DISCONN_SERVER_VIDEO_PIPELINE_BLEND_FRAME_FAILED = 0x800E8608,
        NVST_DISCONN_SERVER_VIDEO_PIPELINE_ALLOCATE_ALPHA_FRAME_FAILED = 0x800E8609,
        NVST_DISCONN_SERVER_VIDEO_PIPELINE_ENCODE_ALPHA_FRAME_FAILED = 0x800E860A,

        /// Encoder Utility Disconnection Reasons.
        NVST_DISCONN_SERVER_VIDEO_ENCODER_UTILITY_GENERIC = 0x800E8700,
        NVST_DISCONN_SERVER_VIDEO_ENCODER_UTILITY_ALLOCATE_FAILED = 0x800E8701,

        /// Runtime Video Frame Provider Disconnection Reasons
        NVST_DISCONN_SERVER_VFP_ANOTHER_INSTANCE_ACTIVE = 0x800E9300,
        NVST_DISCONN_SERVER_VFP_MEM_RESERVE_FAILED = 0x800E9301,
        NVST_DISCONN_SERVER_VFP_PROTECTED_CONTENT = 0x800E9302,
        NVST_DISCONN_SERVER_VFP_DRIVER_FAILURE = 0x800E9303,
        NVST_DISCONN_SERVER_VFP_GRAB_FAILED = 0x800E9304,
        NVST_DISCONN_SERVER_VFP_REALLOC_FAILED = 0x800E9305,
        NVST_DISCONN_SERVER_VFP_INVALIDATED_RESOLUTION_CHANGED = 0x800E9308,
        NVST_DISCONN_SERVER_VFP_INVALIDATED_DEVICE_LOST = 0x800E9309,
        NVST_DISCONN_SERVER_VFP_INVALIDATED_UNSUPPORTED_RESOLUTION = 0x800E930A,
        NVST_DISCONN_SERVER_VFP_INVALIDATED_UNKNOWN = 0x800E930B,
        NVST_DISCONN_SERVER_VFP_RECREATE_FAILED = 0x800E930C,
        NVST_DISCONN_SERVER_VFP_RECREATE_GRAB_FAILED = 0x800E930E,
        NVST_DISCONN_SERVER_VFP_COPY_RECTS_FROM_FAILED = 0x800E9311,
        NVST_DISCONN_SERVER_VFP_WAC_ERROR_GENERIC = 0x800E9312,
        NVST_DISCONN_SERVER_VFP_WAC_ERROR_INVALID_PARAM = 0x800E9313,
        NVST_DISCONN_SERVER_VFP_WAC_ERROR_CAPTURE_FAILURE = 0x800E9314,
        NVST_DISCONN_SERVER_VFP_WAC_ERROR_INVALID_CALL = 0x800E9315,
        NVST_DISCONN_SERVER_VFP_WAC_ERROR_SYSTEM_ERROR = 0x800E9316,
        NVST_DISCONN_SERVER_VFP_WAC_ERROR_INVALID_WINDOW_HANDLE = 0x800E9317,
        NVST_DISCONN_SERVER_VFP_WAC_ERROR_WINDOW_Z_ORDER_CHANGED = 0x800E9318,
        NVST_DISCONN_SERVER_VFP_WAC_ERROR_OPEN_PROCESS_FAILED_ACCESS_DENIED = 0x800E9319,
        NVST_DISCONN_SERVER_VFP_WAC_ERROR_OPEN_PROCESS_FAILED = 0x800E931A,
        NVST_DISCONN_SERVER_VFP_WAC_ERROR_MAG_STRATEGY_INTERNAL_POINTERS_FAILURE = 0x800E931B,
        NVST_DISCONN_SERVER_VFP_WAC_ERROR_MAG_STRATEGY_UPDATE_FAILURE = 0x800E931C,
        NVST_DISCONN_SERVER_VFP_DDA_ERROR_RESOURCE_ACQUIRE_FAILURE = 0x800E931D,
        NVST_DISCONN_SERVER_VFP_DDA_ERROR_FRAME_ACQUIRE_FAILURE = 0x800E931E,

        // Runtime Encoder Disconnection Reasons
        NVST_DISCONN_SERVER_VIDEO_ENCODER_REALLOCATE_INPUT_FRAME_FAILED = 0X800E9400,
        NVST_DISCONN_SERVER_VIDEO_ENCODER_RESOLUTION_CHANGE_FAILED = 0x800E9401,
        NVST_DISCONN_SERVER_VIDEO_ENCODER_SDR_HDR_CHANGE_FAILED = 0x800E9402,
        NVST_DISCONN_SERVER_VIDEO_ENCODER_CONVERT_INPUT_FRAME_FAILED = 0x800E9403,
        NVST_DISCONN_SERVER_VIDEO_ENCODER_MAP_RESOURCE_FAILED = 0x800E9404,
        NVST_DISCONN_SERVER_VIDEO_ENCODER_UNMAP_RESOURCE_FAILED = 0x800E9405,
        NVST_DISCONN_SERVER_VIDEO_ENCODER_ENCODE_PICTURE_FAILED = 0x800E9406,
        NVST_DISCONN_SERVER_VIDEO_ENCODER_COPY_OUTPUT_FAILED = 0x800E9407,
        NVST_DISCONN_SERVER_VIDEO_ENCODER_LOCK_BITSTREAM_FAILED = 0x800E9408,
        NVST_DISCONN_SERVER_VIDEO_ENCODER_UNLOCK_BITSTREAM_FAILED = 0x800E9409,
        NVST_DISCONN_SERVER_VIDEO_ENCODER_RECONFIGURE_FAILED = 0x800E940A,
        NVST_DISCONN_SERVER_VIDEO_ENCODER_ASYNC_MODE_EVENT_WAIT_FAILED = 0x800E940B,

        /// Passthrough group for SessionTerminationReason due to audio, defined in ProtocolDefines.h.
        NVSTE_GROUP_SERVER_TERMINATED_AUDIO = 0x800F0000,
        // Server terminated due to audio issues passthrough error codes.

        /// Errors occuring due to issues in client audio pipeline.
        NVST_R_ERROR_CLIENT_AUDIO_GENERIC_ERROR = 0x80100000,
        // Unknown error occured in audio pipeline.
        /// Unsupported channel configuration was received.
        NVST_R_ERROR_CLIENT_AUDIO_UNSUPPORTED_CHANNEL_CONFIG = 0x80100001,
        /// Encountered error opening/accessing audio sink.
        NVST_R_ERROR_CLIENT_AUDIO_SINK_ERROR = 0x80100002,
        /// Encountered error accessing audio source.
        NVST_R_ERROR_CLIENT_AUDIO_SOURCE_ERROR = 0x80100003,

        // Remote input error codes.
        NVST_R_RI_PROTOCOL_UPDATE_FAILED = 0x800D0000,
        NVST_R_RI_FAILED_TO_CONNECT = 0x800D0001,

        // NVSC VIDEO QOS
        NVST_R_ERROR_BUFFER_FULL = 0x80060000,

        // ENET errors.
        NVST_R_ERROR_ENET_COULD_NOT_CREATE_HOST = 0x80070000,
        NVST_R_ERROR_ENET_COULD_NOT_SET_ADDRESS = 0x80070001,
        NVST_R_ERROR_ENET_COULD_NOT_CREATE_PEER = 0x80070002,
        NVST_R_ERROR_ENET_COULD_NOT_CONNECT_TO_HOST = 0x80070003,
        NVST_R_ERROR_ENET_MC_START_FAILED = 0x80070004,
        NVST_ENET_DISCONNECTED = 0x80470023,

        // EnetSocket start errors
        NVST_ENET_ERROR = 0x8047000b,
        NVST_ENET_CREATE_HOST_ERROR = 0x80470029,

        // DcSctp Errors
        NVST_DCSCTP_ERROR = 80470030,
        NVST_DCSCTP_SOCKET_ALREADY_STARTED_ERROR = 80470031,
        NVST_DCSCTP_TRANSPORT_NOT_READY = 80470031,

        /// Received data from a peer different than the one connected to.
        NVST_ENET_SECURITY_INVALID_READ = 0x80470000,
        /// Received disconnect event from a peer different than the one connected to.
        NVST_ENET_SECURITY_INVALID_DISCONNECT = 0x80470001,
        /// Received invalid event from enet with a peer different than the one connected to.
        NVST_ENET_SECURITY_INVALID_EVENT = 0x80470002,

        // RTSP error subgroups.
        NVSTE_RTSP_SETUP_AUDIO = 0x80084000,
        /// client reporting error codes, in case of RTSP SETUP_AUDIO FAIL
        /// RTSP Client <---error--> rp/rtsp server
        /// RTSP(S)client not able to connect to server ( send error, recv error, ssl level errors)
        NVSTE_RTSP_SETUP_AUDIO_RTSPE_COULDNT_CONNECT = 0x80084007,
        NVSTE_RTSP_SETUP_AUDIO_RTSPE_OPERATION_TIMEDOUT = 0x8008401C,
        NVSTE_RTSP_SETUP_AUDIO_RTSPE_SSL_CONNECT_ERROR = 0x80084023,
        NVSTE_RTSP_SETUP_AUDIO_RTSPE_SEND_ERROR = 0x80084037,
        NVSTE_RTSP_SETUP_AUDIO_RTSPE_RECV_ERROR = 0x80084038,
        NVSTE_RTSP_SETUP_AUDIO_RTSPE_SSL_INVALIDCERTSTATUS = 0x8008405B,
        NVSTE_RTSP_SETUP_AUDIO_RTSPE_NOT_AUTHENTICATED_EXCEPTION = 0x8008405E,
        NVSTE_RTSP_SETUP_AUDIO_RTSPE_POCO_NET_SSLCONTEXT_EXCEPTION = 0x8008405F,
        NVSTE_RTSP_SETUP_AUDIO_RTSPE_POCO_NET_CERTVALIDATION_EXCEPTION = 0x80084060,
        NVSTE_RTSP_SETUP_AUDIO_RTSPE_POCO_NET_SSLCON_UNEXPECTEDLYCLOSED_EXCEPTION = 0x80084061,

        /// relevant only incase of rtsps
        /// RTSPS Client <---> [rp] <---error-> rtsp server
        /// client is able to send/recv to RP and RP is sending the error
        NVSTE_RTSP_SETUP_AUDIO_RESPONSE_CODE_SERVICE_UNAVAILABLE = 0x800841f7,
        NVSTE_RTSP_SETUP_AUDIO_RESPONSE_CODE_GATEWAY_TIMEOUT = 0x800841f8,

        /// RTSPS client <-> [rp] <-> RTSP server
        /// RTSP client <-> RTSP server
        /// client is able to send/recv to server and server response came, but Error reported
        NVSTE_RTSP_SETUP_AUDIO_RESPONSE_CODE_BAD_REQUEST = 0x80084190,
        NVSTE_RTSP_SETUP_AUDIO_RESPONSE_CODE_UNAUTHORIZED = 0x80084191,
        NVSTE_RTSP_SETUP_AUDIO_RESPONSE_CODE_FORBIDDEN = 0x80084193,
        NVSTE_RTSP_SETUP_AUDIO_RESPONSE_CODE_NOT_FOUND = 0x80084194,
        NVSTE_RTSP_SETUP_AUDIO_RESPONSE_CODE_NOT_ALLOWED = 0x80084195,
        NVSTE_RTSP_SETUP_AUDIO_RESPONSE_CODE_AUTHENTICATION_REQUIRED = 0x80084197,
        NVSTE_RTSP_SETUP_AUDIO_RESPONSE_CODE_TIMEOUT = 0x80084198,
        NVSTE_RTSP_SETUP_AUDIO_RESPONSE_CODE_NOT_ENOUGH_BANDWIDTH = 0x800841C5,
        NVSTE_RTSP_SETUP_AUDIO_RESPONSE_CODE_SESSION_NOT_FOUND = 0x800841C6,
        NVSTE_RTSP_SETUP_AUDIO_RESPONSE_CODE_NOT_VALID_IN_THIS_STATE = 0x800841C7,
        NVSTE_RTSP_SETUP_AUDIO_RESPONSE_CODE_UNSUPPORTED_TRANSPORT = 0x800841CD,
        NVSTE_RTSP_SETUP_AUDIO_RESPONSE_CODE_DESTINATION_UNREACHABLE = 0x800841CE,
        NVSTE_RTSP_SETUP_AUDIO_RESPONSE_CODE_INVALID_CSEQ = 0x800841CF,
        NVSTE_RTSP_SETUP_AUDIO_RESPONSE_CODE_MALFORMED_RESPONSE = 0x800841D0,
        NVSTE_RTSP_SETUP_AUDIO_RESPONSE_CODE_INTERNAL_SERVER_ERROR = 0x800841F4,
        NVSTE_RTSP_SETUP_AUDIO_RESPONSE_CODE_NOT_IMPLEMENTED = 0x800841F5,
        NVSTE_RTSP_SETUP_AUDIO_RESPONSE_CODE_VERSION_NOT_SUPPORTED = 0x800841F9,
        NVSTE_RTSP_SETUP_AUDIO_RESPONSE_CODE_SETUP_AUDIO_NOT_SUPPORTED = 0x80084227,
        NVSTE_RTSP_SETUP_AUDIO_RESPONSE_CODE_STREAMER_COMMUNICATION_FAILED = 0x80084228,

        NVSTE_RTSP_SETUP_VIDEO = 0x80087000,

        /// client reporting error codes, in case of RTSP SETUP_VIDEO FAIL
        /// RTSP Client <---error--> rp/rtsp server
        /// RTSP(S)client not able to connect to server ( send error, recv error, ssl level errors)
        NVSTE_RTSP_SETUP_VIDEO_RTSPE_COULDNT_CONNECT = 0x80087007,
        NVSTE_RTSP_SETUP_VIDEO_RTSPE_OPERATION_TIMEDOUT = 0x8008701C,
        NVSTE_RTSP_SETUP_VIDEO_RTSPE_SSL_CONNECT_ERROR = 0x80087023,
        NVSTE_RTSP_SETUP_VIDEO_RTSPE_SEND_ERROR = 0x80087037,
        NVSTE_RTSP_SETUP_VIDEO_RTSPE_RECV_ERROR = 0x80087038,
        NVSTE_RTSP_SETUP_VIDEO_RTSPE_SSL_INVALIDCERTSTATUS = 0x8008705B,
        NVSTE_RTSP_SETUP_VIDEO_RTSPE_NOT_AUTHENTICATED_EXCEPTION = 0x8008705E,
        NVSTE_RTSP_SETUP_VIDEO_RTSPE_POCO_NET_SSLCONTEXT_EXCEPTION = 0x8008705F,
        NVSTE_RTSP_SETUP_VIDEO_RTSPE_POCO_NET_CERTVALIDATION_EXCEPTION = 0x80087060,
        NVSTE_RTSP_SETUP_VIDEO_RTSPE_POCO_NET_SSLCON_UNEXPECTEDLYCLOSED_EXCEPTION = 0x80087061,

        /// relevant only incase of rtsps
        /// RTSPS Client <---> [rp] <---error-> rtsp server
        /// client is able to send/recv to RP and RP is sending the error
        NVSTE_RTSP_SETUP_VIDEO_RESPONSE_CODE_SERVICE_UNAVAILABLE = 0x800871f7,
        NVSTE_RTSP_SETUP_VIDEO_RESPONSE_CODE_GATEWAY_TIMEOUT = 0x800871f8,

        /// RTSPS client <-> [rp] <-> RTSP server
        /// RTSP client <-> RTSP server
        /// client is able to send/recv to server and server response came, but Error reported
        NVSTE_RTSP_SETUP_VIDEO_RESPONSE_CODE_BAD_REQUEST = 0x80087190,
        NVSTE_RTSP_SETUP_VIDEO_RESPONSE_CODE_UNAUTHORIZED = 0x80087191,
        NVSTE_RTSP_SETUP_VIDEO_RESPONSE_CODE_FORBIDDEN = 0x80087193,
        NVSTE_RTSP_SETUP_VIDEO_RESPONSE_CODE_NOT_FOUND = 0x80087194,
        NVSTE_RTSP_SETUP_VIDEO_RESPONSE_CODE_NOT_ALLOWED = 0x80087195,
        NVSTE_RTSP_SETUP_VIDEO_RESPONSE_CODE_AUTHENTICATION_REQUIRED = 0x80087197,
        NVSTE_RTSP_SETUP_VIDEO_RESPONSE_CODE_TIMEOUT = 0x80087198,
        NVSTE_RTSP_SETUP_VIDEO_RESPONSE_CODE_NOT_ENOUGH_BANDWIDTH = 0x800871C5,
        NVSTE_RTSP_SETUP_VIDEO_RESPONSE_CODE_SESSION_NOT_FOUND = 0x800871C6,
        NVSTE_RTSP_SETUP_VIDEO_RESPONSE_CODE_NOT_VALID_IN_THIS_STATE = 0x800871C7,
        NVSTE_RTSP_SETUP_VIDEO_RESPONSE_CODE_UNSUPPORTED_TRANSPORT = 0x800871CD,
        NVSTE_RTSP_SETUP_VIDEO_RESPONSE_CODE_DESTINATION_UNREACHABLE = 0x800871CE,
        NVSTE_RTSP_SETUP_VIDEO_RESPONSE_CODE_INVALID_CSEQ = 0x800871CF,
        NVSTE_RTSP_SETUP_VIDEO_RESPONSE_CODE_MALFORMED_RESPONSE = 0x800871D0,
        NVSTE_RTSP_SETUP_VIDEO_RESPONSE_CODE_INTERNAL_SERVER_ERROR = 0x800871F4,
        NVSTE_RTSP_SETUP_VIDEO_RESPONSE_CODE_NOT_IMPLEMENTED = 0x800871F5,
        NVSTE_RTSP_SETUP_VIDEO_RESPONSE_CODE_VERSION_NOT_SUPPORTED = 0x800871F9,
        NVSTE_RTSP_SETUP_VIDEO_RESPONSE_CODE_SETUP_VIDEO_NOT_SUPPORTED = 0x80087227,
        NVSTE_RTSP_SETUP_VIDEO_RESPONSE_CODE_STREAMER_COMMUNICATION_FAILED = 0x80087228,

        NVSTE_RTSP_SETUP_CONTROL = 0x80088000,

        /// client reporting error codes, in case of RTSP SETUP_CONTROL FAIL
        /// RTSP Client <---error--> rp/rtsp server
        /// RTSP(S)client not able to connect to server ( send error, recv error, ssl level errors)
        NVSTE_RTSP_SETUP_CONTROL_RTSPE_COULDNT_CONNECT = 0x80088007,
        NVSTE_RTSP_SETUP_CONTROL_RTSPE_OPERATION_TIMEDOUT = 0x8008801C,
        NVSTE_RTSP_SETUP_CONTROL_RTSPE_SSL_CONNECT_ERROR = 0x80088023,
        NVSTE_RTSP_SETUP_CONTROL_RTSPE_SEND_ERROR = 0x80088037,
        NVSTE_RTSP_SETUP_CONTROL_RTSPE_RECV_ERROR = 0x80088038,
        NVSTE_RTSP_SETUP_CONTROL_RTSPE_SSL_INVALIDCERTSTATUS = 0x8008805B,
        NVSTE_RTSP_SETUP_CONTROL_RTSPE_NOT_AUTHENTICATED_EXCEPTION = 0x8008805E,
        NVSTE_RTSP_SETUP_CONTROL_RTSPE_POCO_NET_SSLCONTEXT_EXCEPTION = 0x8008805F,
        NVSTE_RTSP_SETUP_CONTROL_RTSPE_POCO_NET_CERTVALIDATION_EXCEPTION = 0x80088060,
        NVSTE_RTSP_SETUP_CONTROL_RTSPE_POCO_NET_SSLCON_UNEXPECTEDLYCLOSED_EXCEPTION = 0x80088061,

        /// relevant only incase of rtsps
        /// RTSPS Client <---> [rp] <---error-> rtsp server
        /// client is able to send/recv to RP and RP is sending the error
        NVSTE_RTSP_SETUP_CONTROL_RESPONSE_CODE_SERVICE_UNAVAILABLE = 0x800881f7,
        NVSTE_RTSP_SETUP_CONTROL_RESPONSE_CODE_GATEWAY_TIMEOUT = 0x800881f8,

        /// RTSPS client <-> [rp] <-> RTSP server
        /// RTSP client <-> RTSP server
        /// client is able to send/recv to server and server response came, but Error reported
        NVSTE_RTSP_SETUP_CONTROL_RESPONSE_CODE_BAD_REQUEST = 0x80088190,
        NVSTE_RTSP_SETUP_CONTROL_RESPONSE_CODE_UNAUTHORIZED = 0x80088191,
        NVSTE_RTSP_SETUP_CONTROL_RESPONSE_CODE_FORBIDDEN = 0x80088193,
        NVSTE_RTSP_SETUP_CONTROL_RESPONSE_CODE_NOT_FOUND = 0x80088194,
        NVSTE_RTSP_SETUP_CONTROL_RESPONSE_CODE_NOT_ALLOWED = 0x80088195,
        NVSTE_RTSP_SETUP_CONTROL_RESPONSE_CODE_AUTHENTICATION_REQUIRED = 0x80088197,
        NVSTE_RTSP_SETUP_CONTROL_RESPONSE_CODE_TIMEOUT = 0x80088198,
        NVSTE_RTSP_SETUP_CONTROL_RESPONSE_CODE_NOT_ENOUGH_BANDWIDTH = 0x800881C5,
        NVSTE_RTSP_SETUP_CONTROL_RESPONSE_CODE_SESSION_NOT_FOUND = 0x800881C6,
        NVSTE_RTSP_SETUP_CONTROL_RESPONSE_CODE_NOT_VALID_IN_THIS_STATE = 0x800881C7,
        NVSTE_RTSP_SETUP_CONTROL_RESPONSE_CODE_UNSUPPORTED_TRANSPORT = 0x800881CD,
        NVSTE_RTSP_SETUP_CONTROL_RESPONSE_CODE_DESTINATION_UNREACHABLE = 0x800881CE,
        NVSTE_RTSP_SETUP_CONTROL_RESPONSE_CODE_INVALID_CSEQ = 0x800881CF,
        NVSTE_RTSP_SETUP_CONTROL_RESPONSE_CODE_MALFORMED_RESPONSE = 0x800881D0,
        NVSTE_RTSP_SETUP_CONTROL_RESPONSE_CODE_INTERNAL_SERVER_ERROR = 0x800881F4,
        NVSTE_RTSP_SETUP_CONTROL_RESPONSE_CODE_NOT_IMPLEMENTED = 0x800881F5,
        NVSTE_RTSP_SETUP_CONTROL_RESPONSE_CODE_VERSION_NOT_SUPPORTED = 0x800881F9,
        NVSTE_RTSP_SETUP_CONTROL_RESPONSE_CODE_SETUP_CONTROL_NOT_SUPPORTED = 0x80088227,
        NVSTE_RTSP_SETUP_CONTROL_RESPONSE_CODE_STREAMER_COMMUNICATION_FAILED = 0x80088228,

        NVSTE_RTSP_SETUP_MIC = 0x80089000,

        /// client reporting error codes, in case of RTSP SETUP_MIC FAIL
        /// RTSP Client <---error--> rp/rtsp server
        /// RTSP(S)client not able to connect to server ( send error, recv error, ssl level errors)
        NVSTE_RTSP_SETUP_MIC_RTSPE_COULDNT_CONNECT = 0x80089007,
        NVSTE_RTSP_SETUP_MIC_RTSPE_OPERATION_TIMEDOUT = 0x8008901C,
        NVSTE_RTSP_SETUP_MIC_RTSPE_SSL_CONNECT_ERROR = 0x80089023,
        NVSTE_RTSP_SETUP_MIC_RTSPE_SEND_ERROR = 0x80089037,
        NVSTE_RTSP_SETUP_MIC_RTSPE_RECV_ERROR = 0x80089038,
        NVSTE_RTSP_SETUP_MIC_RTSPE_SSL_INVALIDCERTSTATUS = 0x8008905B,
        NVSTE_RTSP_SETUP_MIC_RTSPE_NOT_AUTHENTICATED_EXCEPTION = 0x8008905E,
        NVSTE_RTSP_SETUP_MIC_RTSPE_POCO_NET_SSLCONTEXT_EXCEPTION = 0x8008905F,
        NVSTE_RTSP_SETUP_MIC_RTSPE_POCO_NET_CERTVALIDATION_EXCEPTION = 0x80089060,
        NVSTE_RTSP_SETUP_MIC_RTSPE_POCO_NET_SSLCON_UNEXPECTEDLYCLOSED_EXCEPTION = 0x80089061,

        /// relevant only incase of rtsps
        /// RTSPS Client <---> [rp] <---error-> rtsp server
        /// client is able to send/recv to RP and RP is sending the error
        NVSTE_RTSP_SETUP_MIC_RESPONSE_CODE_SERVICE_UNAVAILABLE = 0x800891f7,
        NVSTE_RTSP_SETUP_MIC_RESPONSE_CODE_GATEWAY_TIMEOUT = 0x800891f8,

        /// RTSPS client <-> [rp] <-> RTSP server
        /// RTSP client <-> RTSP server
        /// client is able to send/recv to server and server response came, but Error reported
        NVSTE_RTSP_SETUP_MIC_RESPONSE_CODE_BAD_REQUEST = 0x80089190,
        NVSTE_RTSP_SETUP_MIC_RESPONSE_CODE_UNAUTHORIZED = 0x80089191,
        NVSTE_RTSP_SETUP_MIC_RESPONSE_CODE_FORBIDDEN = 0x80089193,
        NVSTE_RTSP_SETUP_MIC_RESPONSE_CODE_NOT_FOUND = 0x80089194,
        NVSTE_RTSP_SETUP_MIC_RESPONSE_CODE_NOT_ALLOWED = 0x80089195,
        NVSTE_RTSP_SETUP_MIC_RESPONSE_CODE_AUTHENTICATION_REQUIRED = 0x80089197,
        NVSTE_RTSP_SETUP_MIC_RESPONSE_CODE_TIMEOUT = 0x80089198,
        NVSTE_RTSP_SETUP_MIC_RESPONSE_CODE_NOT_ENOUGH_BANDWIDTH = 0x800891C5,
        NVSTE_RTSP_SETUP_MIC_RESPONSE_CODE_SESSION_NOT_FOUND = 0x800891C6,
        NVSTE_RTSP_SETUP_MIC_RESPONSE_CODE_NOT_VALID_IN_THIS_STATE = 0x800891C7,
        NVSTE_RTSP_SETUP_MIC_RESPONSE_CODE_UNSUPPORTED_TRANSPORT = 0x800891CD,
        NVSTE_RTSP_SETUP_MIC_RESPONSE_CODE_DESTINATION_UNREACHABLE = 0x800891CE,
        NVSTE_RTSP_SETUP_MIC_RESPONSE_CODE_INVALID_CSEQ = 0x800891CF,
        NVSTE_RTSP_SETUP_MIC_RESPONSE_CODE_MALFORMED_RESPONSE = 0x800891D0,
        NVSTE_RTSP_SETUP_MIC_RESPONSE_CODE_INTERNAL_SERVER_ERROR = 0x800891F4,
        NVSTE_RTSP_SETUP_MIC_RESPONSE_CODE_NOT_IMPLEMENTED = 0x800891F5,
        NVSTE_RTSP_SETUP_MIC_RESPONSE_CODE_VERSION_NOT_SUPPORTED = 0x800891F9,
        NVSTE_RTSP_SETUP_MIC_RESPONSE_CODE_SETUP_MIC_NOT_SUPPORTED = 0x80089227,
        NVSTE_RTSP_SETUP_MIC_RESPONSE_CODE_STREAMER_COMMUNICATION_FAILED = 0x80089228,

        NVSTE_RTSP_EXTENDED = 0x8008A000,

        /// client reporting error codes, in case of RTSP SETUP_EXTENDED FAIL
        /// RTSP Client <---error--> rp/rtsp server
        /// RTSP(S)client not able to connect to server ( send error, recv error, ssl level errors)
        NVSTE_RTSP_SETUP_EXTENDED_RTSPE_COULDNT_CONNECT = 0x8008A007,
        NVSTE_RTSP_SETUP_EXTENDED_RTSPE_OPERATION_TIMEDOUT = 0x8008A01C,
        NVSTE_RTSP_SETUP_EXTENDED_RTSPE_SSL_CONNECT_ERROR = 0x8008A023,
        NVSTE_RTSP_SETUP_EXTENDED_RTSPE_SEND_ERROR = 0x8008A037,
        NVSTE_RTSP_SETUP_EXTENDED_RTSPE_RECV_ERROR = 0x8008A038,
        NVSTE_RTSP_SETUP_EXTENDED_RTSPE_SSL_INVALIDCERTSTATUS = 0x8008A05B,
        NVSTE_RTSP_SETUP_EXTENDED_RTSPE_NOT_AUTHENTICATED_EXCEPTION = 0x8008A05E,
        NVSTE_RTSP_SETUP_EXTENDED_RTSPE_POCO_NET_SSLCONTEXT_EXCEPTION = 0x8008A05F,
        NVSTE_RTSP_SETUP_EXTENDED_RTSPE_POCO_NET_CERTVALIDATION_EXCEPTION = 0x8008A060,
        NVSTE_RTSP_SETUP_EXTENDED_RTSPE_POCO_NET_SSLCON_UNEXPECTEDLYCLOSED_EXCEPTION = 0x8008A061,

        /// relevant only incase of rtsps
        /// RTSPS Client <---> [rp] <---error-> rtsp server
        /// client is able to send/recv to RP and RP is sending the error
        NVSTE_RTSP_SETUP_EXTENDED_RESPONSE_CODE_SERVICE_UNAVAILABLE = 0x8008A1f7,
        NVSTE_RTSP_SETUP_EXTENDED_RESPONSE_CODE_GATEWAY_TIMEOUT = 0x8008A1f8,

        /// RTSPS client <-> [rp] <-> RTSP server
        /// RTSP client <-> RTSP server
        /// client is able to send/recv to server and server response came, but Error reported
        NVSTE_RTSP_SETUP_EXTENDED_RESPONSE_CODE_BAD_REQUEST = 0x8008A190,
        NVSTE_RTSP_SETUP_EXTENDED_RESPONSE_CODE_UNAUTHORIZED = 0x8008A191,
        NVSTE_RTSP_SETUP_EXTENDED_RESPONSE_CODE_FORBIDDEN = 0x8008A193,
        NVSTE_RTSP_SETUP_EXTENDED_RESPONSE_CODE_NOT_FOUND = 0x8008A194,
        NVSTE_RTSP_SETUP_EXTENDED_RESPONSE_CODE_NOT_ALLOWED = 0x8008A195,
        NVSTE_RTSP_SETUP_EXTENDED_RESPONSE_CODE_AUTHENTICATION_REQUIRED = 0x8008A197,
        NVSTE_RTSP_SETUP_EXTENDED_RESPONSE_CODE_TIMEOUT = 0x8008A198,
        NVSTE_RTSP_SETUP_EXTENDED_RESPONSE_CODE_NOT_ENOUGH_BANDWIDTH = 0x8008A1C5,
        NVSTE_RTSP_SETUP_EXTENDED_RESPONSE_CODE_SESSION_NOT_FOUND = 0x8008A1C6,
        NVSTE_RTSP_SETUP_EXTENDED_RESPONSE_CODE_NOT_VALID_IN_THIS_STATE = 0x8008A1C7,
        NVSTE_RTSP_SETUP_EXTENDED_RESPONSE_CODE_UNSUPPORTED_TRANSPORT = 0x8008A1CD,
        NVSTE_RTSP_SETUP_EXTENDED_RESPONSE_CODE_DESTINATION_UNREACHABLE = 0x8008A1CE,
        NVSTE_RTSP_SETUP_EXTENDED_RESPONSE_CODE_INVALID_CSEQ = 0x8008A1CF,
        NVSTE_RTSP_SETUP_EXTENDED_RESPONSE_CODE_MALFORMED_RESPONSE = 0x8008A1D0,
        NVSTE_RTSP_SETUP_EXTENDED_RESPONSE_CODE_INTERNAL_SERVER_ERROR = 0x8008A1F4,
        NVSTE_RTSP_SETUP_EXTENDED_RESPONSE_CODE_NOT_IMPLEMENTED = 0x8008A1F5,
        NVSTE_RTSP_SETUP_EXTENDED_RESPONSE_CODE_VERSION_NOT_SUPPORTED = 0x8008A1F9,
        NVSTE_RTSP_SETUP_EXTENDED_RESPONSE_CODE_SETUP_EXTENDED_NOT_SUPPORTED = 0x8008A227,
        NVSTE_RTSP_SETUP_EXTENDED_RESPONSE_CODE_STREAMER_COMMUNICATION_FAILED = 0x8008A228,

        // RTSP_OPTIONS error codes
        NVSTE_RTSP_OPTIONS = 0x80082000,

        /// client reporting error codes, in case of RTSP OPTIONS FAIL
        /// RTSP Client <---error--> rp/rtsp server
        /// RTSP(S)client not able to connect to server ( send error, recv error, ssl level errors)
        NVSTE_RTSP_OPTIONS_RTSPE_COULDNT_CONNECT = 0x80082007,
        NVSTE_RTSP_OPTIONS_RTSPE_OPERATION_TIMEDOUT = 0x8008201C,
        NVSTE_RTSP_OPTIONS_RTSPE_SSL_CONNECT_ERROR = 0x80082023,
        NVSTE_RTSP_OPTIONS_RTSPE_SEND_ERROR = 0x80082037,
        NVSTE_RTSP_OPTIONS_RTSPE_RECV_ERROR = 0x80082038,
        NVSTE_RTSP_OPTIONS_RTSPE_SSL_INVALIDCERTSTATUS = 0x8008205B,
        NVSTE_RTSP_OPTIONS_RTSPE_NOT_AUTHENTICATED_EXCEPTION = 0x8008205E,
        NVSTE_RTSP_OPTIONS_RTSPE_POCO_NET_SSLCONTEXT_EXCEPTION = 0x8008205F,
        NVSTE_RTSP_OPTIONS_RTSPE_POCO_NET_CERTVALIDATION_EXCEPTION = 0x80082060,
        NVSTE_RTSP_OPTIONS_RTSPE_POCO_NET_SSLCON_UNEXPECTEDLYCLOSED_EXCEPTION = 0x80082061,

        /// relevant only incase of rtsps
        /// RTSPS Client <---> [rp] <---error-> rtsp server
        /// client is able to send/recv to RP and RP is sending the error
        NVSTE_RTSP_OPTIONS_RESPONSE_CODE_SERVICE_UNAVAILABLE = 0x800821f7,
        NVSTE_RTSP_OPTIONS_RESPONSE_CODE_GATEWAY_TIMEOUT = 0x800821f8,

        /// RTSPS client <-> [rp] <-> RTSP server
        /// RTSP client <--> RTSP server
        /// client is able to send/recv to server and server response came, but Error reported
        NVSTE_RTSP_OPTIONS_RESPONSE_CODE_BAD_REQUEST = 0x80082190,
        NVSTE_RTSP_OPTIONS_RESPONSE_CODE_UNAUTHORIZED = 0x80082191,
        NVSTE_RTSP_OPTIONS_RESPONSE_CODE_FORBIDDEN = 0x80082193,
        NVSTE_RTSP_OPTIONS_RESPONSE_CODE_NOT_FOUND = 0x80082194,
        NVSTE_RTSP_OPTIONS_RESPONSE_CODE_NOT_ALLOWED = 0x80082195,
        NVSTE_RTSP_OPTIONS_RESPONSE_CODE_AUTHENTICATION_REQUIRED = 0x80082197,
        NVSTE_RTSP_OPTIONS_RESPONSE_CODE_TIMEOUT = 0x80082198,
        NVSTE_RTSP_OPTIONS_RESPONSE_CODE_NOT_ENOUGH_BANDWIDTH = 0x800821C5,
        NVSTE_RTSP_OPTIONS_RESPONSE_CODE_SESSION_NOT_FOUND = 0x800821C6,
        NVSTE_RTSP_OPTIONS_RESPONSE_CODE_NOT_VALID_IN_THIS_STATE = 0x800821C7,
        NVSTE_RTSP_OPTIONS_RESPONSE_CODE_UNSUPPORTED_TRANSPORT = 0x800821CD,
        NVSTE_RTSP_OPTIONS_RESPONSE_CODE_DESTINATION_UNREACHABLE = 0x800821CE,
        NVSTE_RTSP_OPTIONS_RESPONSE_CODE_INVALID_CSEQ = 0x800821CF,
        NVSTE_RTSP_OPTIONS_RESPONSE_CODE_MALFORMED_RESPONSE = 0x800821D0,
        NVSTE_RTSP_OPTIONS_RESPONSE_CODE_INTERNAL_SERVER_ERROR = 0x800821F4,
        NVSTE_RTSP_OPTIONS_RESPONSE_CODE_NOT_IMPLEMENTED = 0x800821F5,
        NVSTE_RTSP_OPTIONS_RESPONSE_CODE_VERSION_NOT_SUPPORTED = 0x800821F9,
        NVSTE_RTSP_OPTIONS_RESPONSE_CODE_OPTIONS_NOT_SUPPORTED = 0x80082227,
        NVSTE_RTSP_OPTIONS_RESPONSE_CODE_STREAMER_COMMUNICATION_FAILED = 0x80082228,

        NVST_RTSP_OPTIONS_OK = 0x004820C8,
        NVST_RTSP_OPTIONS_BAD_REQUEST = 0x80482190,
        NVST_RTSP_OPTIONS_VERSION_NOT_SUPPORTED = 0x804821F9,

        // RTSP_DESCRIBE error codes
        NVSTE_RTSP_DESCRIBE = 0x80083000,

        /// client reporting error codes, in case of RTSP DESCRIBE FAIL
        /// RTSP Client <---error--> rp/rtsp server
        /// RTSP(S)client not able to connect to server ( send error, recv error, ssl level errors)
        NVSTE_RTSP_DESCRIBE_RTSPE_COULDNT_CONNECT = 0x80083007,
        NVSTE_RTSP_DESCRIBE_RTSPE_OPERATION_TIMEDOUT = 0x8008301C,
        NVSTE_RTSP_DESCRIBE_RTSPE_SSL_CONNECT_ERROR = 0x80083023,
        NVSTE_RTSP_DESCRIBE_RTSPE_SEND_ERROR = 0x80083037,
        NVSTE_RTSP_DESCRIBE_RTSPE_RECV_ERROR = 0x80083038,
        NVSTE_RTSP_DESCRIBE_RTSPE_SSL_INVALIDCERTSTATUS = 0x8008305B,
        NVSTE_RTSP_DESCRIBE_RTSPE_NOT_AUTHENTICATED_EXCEPTION = 0x8008305E,
        NVSTE_RTSP_DESCRIBE_RTSPE_POCO_NET_SSLCONTEXT_EXCEPTION = 0x8008305F,
        NVSTE_RTSP_DESCRIBE_RTSPE_POCO_NET_CERTVALIDATION_EXCEPTION = 0x80083060,
        NVSTE_RTSP_DESCRIBE_RTSPE_POCO_NET_SSLCON_UNEXPECTEDLYCLOSED_EXCEPTION = 0x80083061,

        /// relevant only incase of rtsps
        /// RTSPS Client <---> [rp] <---error-> rtsp server
        /// client is able to send/recv to RP and RP is sending the error
        NVSTE_RTSP_DESCRIBE_RESPONSE_CODE_SERVICE_UNAVAILABLE = 0x800831f7,
        NVSTE_RTSP_DESCRIBE_RESPONSE_CODE_GATEWAY_TIMEOUT = 0x800831f8,

        /// RTSPS client <-> [rp] <-> RTSP server
        /// RTSP client <--> RTSP server
        /// client is able to send/recv to server and server response came, but Error reported
        NVSTE_RTSP_DESCRIBE_RESPONSE_CODE_BAD_REQUEST = 0x80083190,
        NVSTE_RTSP_DESCRIBE_RESPONSE_CODE_UNAUTHORIZED = 0x80083191,
        NVSTE_RTSP_DESCRIBE_RESPONSE_CODE_FORBIDDEN = 0x80083193,
        NVSTE_RTSP_DESCRIBE_RESPONSE_CODE_NOT_FOUND = 0x80083194,
        NVSTE_RTSP_DESCRIBE_RESPONSE_CODE_NOT_ALLOWED = 0x80083195,
        NVSTE_RTSP_DESCRIBE_RESPONSE_CODE_AUTHENTICATION_REQUIRED = 0x80083197,
        NVSTE_RTSP_DESCRIBE_RESPONSE_CODE_TIMEOUT = 0x80083198,
        NVSTE_RTSP_DESCRIBE_RESPONSE_CODE_NOT_ENOUGH_BANDWIDTH = 0x800831C5,
        NVSTE_RTSP_DESCRIBE_RESPONSE_CODE_SESSION_NOT_FOUND = 0x800831C6,
        NVSTE_RTSP_DESCRIBE_RESPONSE_CODE_NOT_VALID_IN_THIS_STATE = 0x800831C7,
        NVSTE_RTSP_DESCRIBE_RESPONSE_CODE_UNSUPPORTED_TRANSPORT = 0x800831CD,
        NVSTE_RTSP_DESCRIBE_RESPONSE_CODE_DESTINATION_UNREACHABLE = 0x800831CE,
        NVSTE_RTSP_DESCRIBE_RESPONSE_CODE_INVALID_CSEQ = 0x800831CF,
        NVSTE_RTSP_DESCRIBE_RESPONSE_CODE_MALFORMED_RESPONSE = 0x800831D0,
        NVSTE_RTSP_DESCRIBE_RESPONSE_CODE_INTERNAL_SERVER_ERROR = 0x800831F4,
        NVSTE_RTSP_DESCRIBE_RESPONSE_CODE_NOT_IMPLEMENTED = 0x800831F5,
        NVSTE_RTSP_DESCRIBE_RESPONSE_CODE_VERSION_NOT_SUPPORTED = 0x800831F9,
        NVSTE_RTSP_DESCRIBE_RESPONSE_CODE_DESCRIBE_NOT_SUPPORTED = 0x80083227,
        NVSTE_RTSP_DESCRIBE_RESPONSE_CODE_STREAMER_COMMUNICATION_FAILED = 0x80083228,
        NVST_RTSP_DESCRIBE_OK = 0x004830C8,
        NVST_RTSP_DESCRIBE_BAD_REQUEST = 0x80483190,
        NVST_RTSP_DESCRIBE_INTERNAL_SERVER_ERROR = 0x804831F4,

        // RTSP_PLAY error codes
        NVSTE_RTSP_PLAY = 0x80085000,

        /// client reporting error codes, in case of RTSP PLAY FAIL
        /// RTSP Client <---error--> rp/rtsp server
        /// RTSP(S)client not able to connect to server ( send error, recv error, ssl level errors)
        NVSTE_RTSP_PLAY_RTSPE_COULDNT_CONNECT = 0x80085007,
        NVSTE_RTSP_PLAY_RTSPE_OPERATION_TIMEDOUT = 0x8008501C,
        NVSTE_RTSP_PLAY_RTSPE_SSL_CONNECT_ERROR = 0x80085023,
        NVSTE_RTSP_PLAY_RTSPE_SEND_ERROR = 0x80085037,
        NVSTE_RTSP_PLAY_RTSPE_RECV_ERROR = 0x80085038,
        NVSTE_RTSP_PLAY_RTSPE_SSL_INVALIDCERTSTATUS = 0x8008505B,
        NVSTE_RTSP_PLAY_RTSPE_NOT_AUTHENTICATED_EXCEPTION = 0x8008505E,
        NVSTE_RTSP_PLAY_RTSPE_POCO_NET_SSLCONTEXT_EXCEPTION = 0x8008505F,
        NVSTE_RTSP_PLAY_RTSPE_POCO_NET_CERTVALIDATION_EXCEPTION = 0x80085060,
        NVSTE_RTSP_PLAY_RTSPE_POCO_NET_SSLCON_UNEXPECTEDLYCLOSED_EXCEPTION = 0x80085061,

        /// relevant only incase of rtsps
        /// RTSPS Client <---> [rp] <---error-> rtsp server
        /// client is able to send/recv to RP and RP is sending the error
        NVSTE_RTSP_PLAY_RESPONSE_CODE_SERVICE_UNAVAILABLE = 0x800851f7,
        NVSTE_RTSP_PLAY_RESPONSE_CODE_GATEWAY_TIMEOUT = 0x800851f8,

        /// RTSPS client <-> [rp] <-> RTSP server
        /// RTSP client <--> RTSP server
        /// client is able to send/recv to server and server response came, but Error reported
        NVSTE_RTSP_PLAY_RESPONSE_CODE_BAD_REQUEST = 0x80085190,
        NVSTE_RTSP_PLAY_RESPONSE_CODE_UNAUTHORIZED = 0x80085191,
        NVSTE_RTSP_PLAY_RESPONSE_CODE_FORBIDDEN = 0x80085193,
        NVSTE_RTSP_PLAY_RESPONSE_CODE_NOT_FOUND = 0x80085194,
        NVSTE_RTSP_PLAY_RESPONSE_CODE_NOT_ALLOWED = 0x80085195,
        NVSTE_RTSP_PLAY_RESPONSE_CODE_AUTHENTICATION_REQUIRED = 0x80085197,
        NVSTE_RTSP_PLAY_RESPONSE_CODE_TIMEOUT = 0x80085198,
        NVSTE_RTSP_PLAY_RESPONSE_CODE_NOT_ENOUGH_BANDWIDTH = 0x800851C5,
        NVSTE_RTSP_PLAY_RESPONSE_CODE_SESSION_NOT_FOUND = 0x800851C6,
        NVSTE_RTSP_PLAY_RESPONSE_CODE_NOT_VALID_IN_THIS_STATE = 0x800851C7,
        NVSTE_RTSP_PLAY_RESPONSE_CODE_UNSUPPORTED_TRANSPORT = 0x800851CD,
        NVSTE_RTSP_PLAY_RESPONSE_CODE_DESTINATION_UNREACHABLE = 0x800851CE,
        NVSTE_RTSP_PLAY_RESPONSE_CODE_INVALID_CSEQ = 0x800851CF,
        NVSTE_RTSP_PLAY_RESPONSE_CODE_MALFORMED_RESPONSE = 0x800851D0,
        NVSTE_RTSP_PLAY_RESPONSE_CODE_INTERNAL_SERVER_ERROR = 0x800851F4,
        NVSTE_RTSP_PLAY_RESPONSE_CODE_NOT_IMPLEMENTED = 0x800851F5,
        NVSTE_RTSP_PLAY_RESPONSE_CODE_VERSION_NOT_SUPPORTED = 0x800851F9,
        NVSTE_RTSP_PLAY_RESPONSE_CODE_PLAY_NOT_SUPPORTED = 0x80085227,
        NVSTE_RTSP_PLAY_RESPONSE_CODE_STREAMER_COMMUNICATION_FAILED = 0x80085228,
        NVST_RTSP_PLAY_OK = 0x004850C8,
        NVST_RTSP_PLAY_BAD_REQUEST = 0x80485190,
        NVST_RTSP_PLAY_NOT_VALID_IN_THIS_STATE = 0x804851C7,
        NVST_RTSP_PLAY_SESSION_NOT_FOUND = 0x804851C6,

        /// RTSP_ANNOUNCE error codes
        NVSTE_RTSP_ANNOUNCE = 0x80086000,

        /// client reporting error codes, in case of RTSP ANNOUNCE FAIL
        /// RTSP Client <---error--> rp/rtsp server
        /// RTSP(S)client not able to connect to server ( send error, recv error, ssl level errors)
        NVSTE_RTSP_ANNOUNCE_RTSPE_COULDNT_CONNECT = 0x80086007,
        NVSTE_RTSP_ANNOUNCE_RTSPE_OPERATION_TIMEDOUT = 0x8008601C,
        NVSTE_RTSP_ANNOUNCE_RTSPE_SSL_CONNECT_ERROR = 0x80086023,
        NVSTE_RTSP_ANNOUNCE_RTSPE_SEND_ERROR = 0x80086037,
        NVSTE_RTSP_ANNOUNCE_RTSPE_RECV_ERROR = 0x80086038,
        NVSTE_RTSP_ANNOUNCE_RTSPE_SSL_INVALIDCERTSTATUS = 0x8008605B,
        NVSTE_RTSP_ANNOUNCE_RTSPE_NOT_AUTHENTICATED_EXCEPTION = 0x8008605E,
        NVSTE_RTSP_ANNOUNCE_RTSPE_POCO_NET_SSLCONTEXT_EXCEPTION = 0x8008605F,
        NVSTE_RTSP_ANNOUNCE_RTSPE_POCO_NET_CERTVALIDATION_EXCEPTION = 0x80086060,
        NVSTE_RTSP_ANNOUNCE_RTSPE_POCO_NET_SSLCON_UNEXPECTEDLYCLOSED_EXCEPTION = 0x80086061,

        /// relevant only incase of rtsps
        /// RTSPS Client <---> [rp] <---error-> rtsp server
        /// client is able to send/recv to RP and RP is sending the error
        NVSTE_RTSP_ANNOUNCE_RESPONSE_CODE_SERVICE_UNAVAILABLE = 0x800861f7,
        NVSTE_RTSP_ANNOUNCE_RESPONSE_CODE_GATEWAY_TIMEOUT = 0x800861f8,

        /// RTSPS client <-> [rp] <-> RTSP server
        /// RTSP client <--> RTSP server
        /// client is able to send/recv to server and server response came, but Error reported
        NVSTE_RTSP_ANNOUNCE_RESPONSE_CODE_BAD_REQUEST = 0x80086190,
        NVSTE_RTSP_ANNOUNCE_RESPONSE_CODE_UNAUTHORIZED = 0x80086191,
        NVSTE_RTSP_ANNOUNCE_RESPONSE_CODE_FORBIDDEN = 0x80086193,
        NVSTE_RTSP_ANNOUNCE_RESPONSE_CODE_NOT_FOUND = 0x80086194,
        NVSTE_RTSP_ANNOUNCE_RESPONSE_CODE_NOT_ALLOWED = 0x80086195,
        NVSTE_RTSP_ANNOUNCE_RESPONSE_CODE_AUTHENTICATION_REQUIRED = 0x80086197,
        NVSTE_RTSP_ANNOUNCE_RESPONSE_CODE_TIMEOUT = 0x80086198,
        NVSTE_RTSP_ANNOUNCE_RESPONSE_CODE_NOT_ENOUGH_BANDWIDTH = 0x800861C5,
        NVSTE_RTSP_ANNOUNCE_RESPONSE_CODE_SESSION_NOT_FOUND = 0x800861C6,
        NVSTE_RTSP_ANNOUNCE_RESPONSE_CODE_NOT_VALID_IN_THIS_STATE = 0x800861C7,
        NVSTE_RTSP_ANNOUNCE_RESPONSE_CODE_UNSUPPORTED_TRANSPORT = 0x800861CD,
        NVSTE_RTSP_ANNOUNCE_RESPONSE_CODE_DESTINATION_UNREACHABLE = 0x800861CE,
        NVSTE_RTSP_ANNOUNCE_RESPONSE_CODE_INVALID_CSEQ = 0x800861CF,
        NVSTE_RTSP_ANNOUNCE_RESPONSE_CODE_MALFORMED_RESPONSE = 0x800861D0,
        NVSTE_RTSP_ANNOUNCE_RESPONSE_CODE_INTERNAL_SERVER_ERROR = 0x800861F4,
        NVSTE_RTSP_ANNOUNCE_RESPONSE_CODE_NOT_IMPLEMENTED = 0x800861F5,
        NVSTE_RTSP_ANNOUNCE_RESPONSE_CODE_VERSION_NOT_SUPPORTED = 0x800861F9,
        NVSTE_RTSP_ANNOUNCE_RESPONSE_CODE_ANNOUNCE_NOT_SUPPORTED = 0x80086227,
        NVSTE_RTSP_ANNOUNCE_RESPONSE_CODE_STREAMER_COMMUNICATION_FAILED = 0x80086228,

        NVST_RTSP_ANNOUNCE_OK = 0x004860C8,
        NVST_RTSP_ANNOUNCE_BAD_REQUEST = 0x80486190,
        NVST_RTSP_ANNOUNCE_NOT_VALID_IN_THIS_STATE = 0x804861C7,
        NVST_RTSP_ANNOUNCE_NOT_FOUND = 0x80486194,
        NVST_RTSP_ANNOUNCE_INTERNAL_SERVER_ERROR = 0x804861F4,

        /// RTSP_TEARDOWN error codes
        NVSTE_RTSP_TEARDOWN = 0x8008B000,

        /// client reporting error codes, in case of RTSP TEARDOWN FAIL
        /// RTSP Client <---error--> rp/rtsp server
        /// RTSP(S)client not able to connect to server ( send error, recv error, ssl level errors)
        NVSTE_RTSP_TEARDOWN_RTSPE_COULDNT_CONNECT = 0x8008B007,
        NVSTE_RTSP_TEARDOWN_RTSPE_OPERATION_TIMEDOUT = 0x8008B01C,
        NVSTE_RTSP_TEARDOWN_RTSPE_SSL_CONNECT_ERROR = 0x8008B023,
        NVSTE_RTSP_TEARDOWN_RTSPE_SEND_ERROR = 0x8008B037,
        NVSTE_RTSP_TEARDOWN_RTSPE_RECV_ERROR = 0x8008B038,
        NVSTE_RTSP_TEARDOWN_RTSPE_SSL_INVALIDCERTSTATUS = 0x8008B05B,
        NVSTE_RTSP_TEARDOWN_RTSPE_NOT_AUTHENTICATED_EXCEPTION = 0x8008B05E,
        NVSTE_RTSP_TEARDOWN_RTSPE_POCO_NET_SSLCONTEXT_EXCEPTION = 0x8008B05F,
        NVSTE_RTSP_TEARDOWN_RTSPE_POCO_NET_CERTVALIDATION_EXCEPTION = 0x8008B060,
        NVSTE_RTSP_TEARDOWN_RTSPE_POCO_NET_SSLCON_UNEXPECTEDLYCLOSED_EXCEPTION = 0x8008B061,

        /// relevant only incase of rtsps
        /// RTSPS Client <---> [rp] <---error-> rtsp server
        /// client is able to send/recv to RP and RP is sending the error
        NVSTE_RTSP_TEARDOWN_RESPONSE_CODE_SERVICE_UNAVAILABLE = 0x8008B1f7,
        NVSTE_RTSP_TEARDOWN_RESPONSE_CODE_GATEWAY_TIMEOUT = 0x8008B1f8,

        /// RTSPS client <-> [rp] <-> RTSP server
        /// RTSP Client <---> RTSP Server
        /// client is able to send/recv to server and server response came, but Error reported
        NVSTE_RTSP_TEARDOWN_RESPONSE_CODE_BAD_REQUEST = 0x8008B190,
        NVSTE_RTSP_TEARDOWN_RESPONSE_CODE_UNAUTHORIZED = 0x8008B191,
        NVSTE_RTSP_TEARDOWN_RESPONSE_CODE_FORBIDDEN = 0x8008B193,
        NVSTE_RTSP_TEARDOWN_RESPONSE_CODE_NOT_FOUND = 0x8008B194,
        NVSTE_RTSP_TEARDOWN_RESPONSE_CODE_NOT_ALLOWED = 0x8008B195,
        NVSTE_RTSP_TEARDOWN_RESPONSE_CODE_AUTHENTICATION_REQUIRED = 0x8008B197,
        NVSTE_RTSP_TEARDOWN_RESPONSE_CODE_TIMEOUT = 0x8008B198,
        NVSTE_RTSP_TEARDOWN_RESPONSE_CODE_NOT_ENOUGH_BANDWIDTH = 0x8008B1C5,
        NVSTE_RTSP_TEARDOWN_RESPONSE_CODE_SESSION_NOT_FOUND = 0x8008B1C6,
        NVSTE_RTSP_TEARDOWN_RESPONSE_CODE_NOT_VALID_IN_THIS_STATE = 0x8008B1C7,
        NVSTE_RTSP_TEARDOWN_RESPONSE_CODE_UNSUPPORTED_TRANSPORT = 0x8008B1CD,
        NVSTE_RTSP_TEARDOWN_RESPONSE_CODE_DESTINATION_UNREACHABLE = 0x8008B1CE,
        NVSTE_RTSP_TEARDOWN_RESPONSE_CODE_INVALID_CSEQ = 0x8008B1CF,
        NVSTE_RTSP_TEARDOWN_RESPONSE_CODE_MALFORMED_RESPONSE = 0x8008B1D0,
        NVSTE_RTSP_TEARDOWN_RESPONSE_CODE_INTERNAL_SERVER_ERROR = 0x8008B1F4,
        NVSTE_RTSP_TEARDOWN_RESPONSE_CODE_NOT_IMPLEMENTED = 0x8008B1F5,
        NVSTE_RTSP_TEARDOWN_RESPONSE_CODE_VERSION_NOT_SUPPORTED = 0x8008B1F9,
        NVSTE_RTSP_TEARDOWN_RESPONSE_CODE_TEARDOWN_NOT_SUPPORTED = 0x8008B227,
        NVSTE_RTSP_TEARDOWN_RESPONSE_CODE_STREAMER_COMMUNICATION_FAILED = 0x8008B228,

        NVST_RTSP_TEARDOWN_OK = 0x0048B0C8,
        NVST_RTSP_TEARDOWN_BAD_REQUEST = 0x8048B190,
        NVST_RTSP_TEARDOWN_NOT_VALID_IN_THIS_STATE = 0x8048B1C7,

        // RTSP_SETUP error codes
        NVSTE_RTSP_SETUP = 0x8008C000,
        NVST_RTSP_SETUP_OK = 0x0048C0C8,
        NVST_RTSP_SETUP_BAD_REQUEST = 0x8048C190,
        NVST_RTSP_SETUP_NOT_VALID_IN_THIS_STATE = 0x8048C1C7,
        NVST_RTSP_SETUP_SESSION_NOT_FOUND = 0x8048C1C6,
        NVST_RTSP_SETUP_NOT_FOUND = 0x8048C194,
        NVST_RTSP_SETUP_INTERNAL_SERVER_ERROR = 0x8048C1F4,

        // Connection attempt error codes.
        /// Unspecified reason.
        NVST_R_CONNERR_UNSPECIFIED = 0x80020000,
        /// Server is old.
        NVST_R_CONNERR_VERSION_SERVER_OLD = 0x80020001,
        /// Client is old.
        NVST_R_CONNERR_VERSION_CLIENT_OLD = 0x80020002,
        /// Hole punching failed.
        NVST_R_CONNERR_HOLE_PUNCH_FAILED = 0x80020003,
        /// RTSP STUNing failed.
        NVST_R_CONNERR_STUN_FAILED = 0x80020004,
        /// Validation of server certificate failed.
        NVST_R_CONNERR_CERT_VALIDATION_FAILED = 0x80020005,
        /// Generaion of server certificate failed.
        NVST_R_CONNERR_CERT_GENERATION_FAILED = 0x80020006,

        // Streamer server terminate reason codes.
        NVSTE_GROUP_SERVER_TERMINATED_INTENDED = 0x80030000,
        /// Server Quit game.
        NVST_DISCONN_INTENDED = 0x80030000,
        /// Game focus change on server.
        NVST_DISCONN_HOTKEY = 0x80030001,
        /// Client no longer alive.
        NVST_DISCONN_TIMEOUT = 0x80030002,
        /// Disconnected failed on launch failure on lock screen.
        NVST_DISCONN_LAUNCH_FAILED = 0x80030003,
        /// User change after remote login.
        NVST_DISCONN_USER_CHANGED = 0x80030004,
        /// Unspecified reason termination.
        NVST_DISCONN_UNSPECIFIED = 0x80030005,
        /// Disconnected due to shifting to windows mode.
        NVST_DISCONN_WINDOWED_MODE = 0x80030006,
        /// Disconnected due to idle user.
        NVST_DISCONN_USER_IDLE = 0x80030007,
        /// Disconnected by PM for unknown reason.
        NVST_DISCONN_UNKNOWN_FROM_PM = 0x80030008,
        /// Disconnected after user entitlement timeout.
        NVST_DISCONN_USER_ENTITLEMENT_TIMEOUT = 0x80030009,
        /// Disconnected after client could not reconnect.
        NVST_DISCONN_CLIENT_RECONNECT_TIMEOUT = 0x8003000A,
        /// Disconnected commanded by operator.
        NVST_DISCONN_OPERATOR_COMMANDED_TERMINATION = 0x8003000B,
        /// Disconnected after reaching max session time limit.
        NVST_DISCONN_MAX_SESSION_TIME_LIMIT = 0x8003000C,
        /// Disconnected after reaching the max concurrent session limit.
        NVST_DISCONN_MAX_CONCURRENT_SESSION_LIMIT = 0x8003000D,
        /// SDK requested a session pause.
        NVST_DISCONN_BIFROST_INITIATED_SESSION_PAUSE = 0x8003000E,
        /// Frame grab failed.
        NVST_DISCONN_SERVER_TERMINATED_FRAME_GRAB_FAILED = 0x8003000F,
        /// Unknown intended disconnection coming from the client.
        NVST_DISCONN_UNKNOWN_TRI_CLIENT = 0x80030010,
        /// Unknown intended disconnection coming from the server.
        NVST_DISCONN_UNKNOWN_TRI_SERVER = 0x80030011,
        /// Unknown intended disconnection coming from the PM.
        NVST_DISCONN_UNKNOWN_TRI_PM = 0x80030012,
        /// Disconnection commanded by the PM.
        NVST_DISCONN_PM_COMMANDED_TERMINATION = 0x80030013,
        /// Game process terminated.
        NVST_DISCONN_SERVER_TERMINATED_GAME_PROCESS_EXITED = 0x80030014,
        /// Unauthorized process detected by the server.
        NVST_DISCONN_SERVER_TERMINATED_INTENDED_UNAUTHORIZED_PROCESS = 0x80030015,
        /// User has log in from another system.
        NVST_DISCONN_SERVER_TERMINATED_MULTIPLE_LOGIN = 0x80030016,
        /// Malicious process detected by the server.
        NVST_DISCONN_SERVER_TERMINATED_INTENDED_MALICIOUS_PROCESS = 0x80030017,
        /// Unknown process detected by the server.
        NVST_DISCONN_SERVER_TERMINATED_INTENDED_UNKNOWN_PROCESS = 0x80030018,
        /// Miner process detected by the server.
        NVST_DISCONN_SERVER_TERMINATED_INTENDED_MINER_PROCESS = 0x80030019,
        /// Timeout Detection and Recovery(TDR) occurred on the server.
        NVST_DISCONN_SERVER_TERMINATED_FULL_TDR_OCCURRED = 0x8003001A,
        /// Timeout during frame grab.
        NVST_DISCONN_SERVER_TERMINATED_FRAME_GRAB_TIMEOUT = 0x8003001B,
        /// Timeout during frame encoding.
        NVST_DISCONN_SERVER_TERMINATED_FRAME_ENCODE_TIMEOUT = 0x8003001C,
        /// Timeout during frame sending.
        NVST_DISCONN_SERVER_TERMINATED_FRAME_SEND_TIMEOUT = 0x8003001D,
        /// User has log in from another system.
        NVST_DISCONN_SERVER_TERMINATED_ANOTHER_CLIENT = 0x8003001E,
        /// Server detected code integrity violation.
        NVST_DISCONN_SERVER_TERMINATED_CODE_INTEGRITY_VIOLATION = 0x8003001F,
        /// Server detected unauthorized activity.
        NVST_DISCONN_SERVER_TERMINATED_UNAUTHORIZED_ACTIVITY_DETECTED = 0x80030020,
        /// Server detected unsupported display topology change.
        NVST_DISCONN_SERVER_TERMINATED_DISPLAY_TOPOLOGY_CHANGED = 0x80030021,
        /// Server was closed by nvstStopServer.
        NVST_DISCONN_SERVER_TERMINATED_CLOSED = 0x80030023,
        /// DTLS connection was gracefully closed.
        NVST_DISCONN_SERVER_TERMINATED_DTLS_CLOSED_DEPRECATED = 0x80030024,
        /// DTLS fingerprint mismatched.
        NVST_DISCONN_SERVER_TERMINATED_DTLS_FINGERPRINT_MISMATCH_DEPRECATED = 0x80030025,
        /// DTLS completed without an SRTP profile.
        NVST_DISCONN_SERVER_TERMINATED_NO_SRTP_PROFILE_DEPRECATED = 0x80030026,
        /// SRTP profile unsupported.
        NVST_DISCONN_SERVER_TERMINATED_SRTP_PROFILE_UNSUPPORTED_DEPRECATED = 0x80030027,
        /// Failure in SSL_export_keying_material.
        NVST_DISCONN_SERVER_TERMINATED_SSL_KEYING_MATERIAL_EXPORT_FAILED_DEPRECATED = 0x80030028,
        /// SRTP Setup failed.
        NVST_DISCONN_SERVER_TERMINATED_SRTP_SETUP_FAILED_DEPRECATED = 0x80030029,
        /// General failure in DTLS context initialization.
        NVST_DISCONN_SERVER_TERMINATED_DTLS_CONTEXT_ERROR_DEPRECATED = 0x8003002A,
        /// Peer did not send required DTLS certificate (or there was an error extracting it).
        NVST_DISCONN_SERVER_TERMINATED_DTLS_NO_PEER_CERTIFICATE_DEPRECATED = 0x8003002B,

        // Network error codes.
        NVSTE_GROUP_NETERR = 0x80040000,
        /// Unknown network error.
        NVST_NETERR_UNSPECIFIED = 0x80040000,
        /// Network operation timed out.
        NVST_NETERR_TIMEOUT = 0x80040001,
        /// Lost network connectivity.
        NVST_NETERR_CONNLOST = 0x80040002,
        /// Unintended disconnection.
        NVST_NETERR_UNINTENDED = 0x80040003,
        /// Lost network connectivity.
        NVST_NETERR_CLIENT_TIMED_OUT = 0x80040004,
        /// Client-side unintended disconnect.
        NVST_NETERR_CLIENT_DISCONNECT_UNINTENDED = 0x80040005,
        /// Client-side RTP ping packets not received.
        NVST_NETERR_CLIENT_DISCONNECT_UNINTENDED_NO_RTP_PING = 0x80040006,
        /// No video frames assembled from received video packets due to lossy network.
        NVST_NETERR_CLIENT_DISCONNECT_UNINTENDED_NO_VIDEO_FRAMES_LOSSY_NETWORK = 0x80040007,
        /// No valid packets (neither data nor ping) seen in the video stream during the entire session.
        NVST_NETERR_CLIENT_DISCONNECT_UNINTENDED_NO_VIDEO_PACKETS_EVER = 0x80040008,
        /// No server ping packets.
        NVST_NETERR_CLIENT_DISCONNECT_UNINTENDED_NO_SERVER_HEART_BEAT = 0x80040009,
        /// Control socket error.
        NVST_NETERR_CLIENT_DISCONNECT_UNINTENDED_CONTROL_SOCKET_ERROR = 0x8004000A,
        /// Server-side unintended termination.
        NVST_NETERR_SERVER_TERMINATED_UNINTENDED = 0x8004000B,
        /// Remote input server error.
        NVST_NETERR_SERVER_TERMINATED_REMOTE_INPUT_ERROR = 0x8004000C,
        /// Config unavailable.
        NVST_NETERR_SERVER_TERMINATED_CONFIG_UNAVAILABLE = 0x8004000D,
        /// Invalid command.
        NVST_NETERR_SERVER_TERMINATED_INVALID_COMMAND = 0x8004000E,
        /// Invalid mouse state.
        NVST_NETERR_SERVER_TERMINATED_INVALID_MOUSE_STATE = 0x8004000F,
        /// No valid packets (neither data nor ping) seen in the video stream recently.
        NVST_NETERR_CLIENT_DISCONNECT_UNINTENDED_NO_VIDEO_PACKETS_RECENTLY = 0x80040010,
        /// Failed to start encoder.
        NVST_NETERR_SERVER_TERMINATED_ENCODER_START_FAILED = 0x80040011,
        /// Unknown error coming from the client.
        NVST_NETERR_UNKNOWN_TRU_CLIENT = 0x80040012,
        /// Unknown error coming from the server.
        NVST_NETERR_UNKNOWN_TRU_SERVER = 0x80040013,
        /// STUNing on video port failed.
        NVST_NETERR_CLIENT_DISCONNECT_UNINTENDED_STUN_FAILED_DEPRECATED = 0x80040014,
        /// Sent Audio Buffer is NULL.
        NVST_NULL_AUDIO_BUFFER = 0x80040015,
        /// Invalid RTSP signaling port
        NVST_INVALID_RTSP_SIGNALING_PORT = 0x80040016,
        /// No packets on downstream channels (Audio, ENET, Video) prior to timeout.
        NVST_NETERR_CLIENT_DISCONNECT_UNINTENDED_DOWNSTREAM_BLACKOUT = 0x80040017,
        /// No packets received on control channel by server app, but exist in underlying libraries.
        /// They couldn't be dispatched to the app because not all sequence numbers are received etc.
        NVST_NETERR_CLIENT_TIMED_OUT_PACKETS_IN_DISPATCH_QUEUE = 0x80040018,

        /// Intended Client Disconnection Reasons.
        NVST_DISCONN_CLIENT_INTENDED = 0x80110000,
        /// Client disconnected.
        /// \todo Need to add mapping for this code only because client isn't currently tranlating both end_session
        /// and intended to NVST_DISCONN_INTENDED
        /// (At Bifrost layer)
        /// this and NVST_DISCONN_INTENDED mapped to same
        NVST_DISCONN_CLIENT_END_SESSION = 0x80110000,
        /// Client disconnected but keep the session alive.
        NVST_DISCONN_CLIENT_KEEP_SESSION_ALIVE = 0x80110001,
        /// Client disconnected the websocket intentionally or because of some error to stop the session.
        NVST_DISCONN_CLIENT_ON_SIGNALING = 0x80110002,
        /// For use with NvstCallbacks.onSignalingHeaders. Return one of the following NVST_SIGERR_* results to reject
        /// the signaling connection and provide more information to the client.
        //
        /// Generic reason for rejecting a signaling connection.
        NVST_SIGERR_GENERIC = 0x80420000,
        /// Client did not provide the necessary header(s) for authentication.
        NVST_SIGERR_FORBIDDEN = 0x80420001,

        /// Native RTC peer disconnect codes.
        NVSTE_GROUP_NATIVE_RTC_PEER_TRANSPORT_DISCONNECT = 0x80120000,
        /// DTLS connection was gracefully closed.
        NVST_DISCONN_PEER_TRANSPORT_TERMINATED_DTLS_CLOSED = 0x80120001,
        /// DTLS fingerprint mismatched.
        NVST_DISCONN_PEER_TRANSPORT_TERMINATED_DTLS_FINGERPRINT_MISMATCH = 0x80120002,
        /// DTLS completed without an SRTP profile.
        NVST_DISCONN_PEER_TRANSPORT_TERMINATED_NO_SRTP_PROFILE = 0x80120003,
        /// SRTP profile unsupported.
        NVST_DISCONN_PEER_TRANSPORT_TERMINATED_SRTP_PROFILE_UNSUPPORTED = 0x80120004,
        /// Failure in SSL_export_keying_material.
        NVST_DISCONN_PEER_TRANSPORT_TERMINATED_SSL_KEYING_MATERIAL_EXPORT_FAILED = 0x80120005,
        /// SRTP Setup failed.
        NVST_DISCONN_PEER_TRANSPORT_TERMINATED_SRTP_SETUP_FAILED = 0x80120006,
        /// failure in DTLS context initialization SSL_accept.
        NVST_DISCONN_PEER_TRANSPORT_TERMINATED_SSL_ACCEPT_ERROR = 0x80120007,
        /// failure in DTLS context initialization SSL_CTX_use_certificate.
        NVST_DISCONN_PEER_TRANSPORT_TERMINATED_USE_CERTIFICATE_ERROR = 0x80120008,
        /// failure in DTLS context initialization SSL_CTX_use_PrivateKey.
        NVST_DISCONN_PEER_TRANSPORT_TERMINATED_USE_PRIVATE_KEY_ERROR = 0x80120009,
        /// failure in DTLS context initialization SSL_CTX_check_private_key.
        NVST_DISCONN_PEER_TRANSPORT_TERMINATED_CHECK_PRIVATE_KEY_ERROR = 0x8012000a,
        /// failure in DTLS context initialization SSL_CTX_set_tlsext_use_srtp.
        NVST_DISCONN_PEER_TRANSPORT_TERMINATED_SSL_USE_SRTP_ERROR = 0x8012000b,
        /// Peer did not send required DTLS certificate (or there was an error extracting it).
        NVST_DISCONN_PEER_TRANSPORT_TERMINATED_DTLS_NO_PEER_CERTIFICATE = 0x8012000c,
        /// DTLS connection close alert received.
        NVST_DISCONN_PEER_TRANSPORT_TERMINATED_DTLS_ALERT = 0x8012000d,
        /// Peer disconnected the websocket intentionally or because of some error to stop the session.
        NVST_DISCONN_PEER_TRANSPORT_TERMINATED_ON_SIGNALING = 0x8012000e,
        /// Dtls handshake timed out
        NVST_DISCONN_PEER_TRANSPORT_TIMEOUT_ON_SIGNALING = 0x8012000f,

        /// Native RTC network error codes.
        NVSTE_GROUP_NATIVE_RTC_NETWORK_ERROR = 0x80130000,
        /// Peer unintended termination.
        NVST_NETERR_PEER_TRANSPORT_TERMINATED_UNINTENDED = 0x80130001,
        /// Unknown network error.
        NVST_NETERR_PEER_TRANSPORT_UNSPECIFIED = 0x80130002,
        /// Lost network connectivity.
        NVST_NETERR_PEER_TRANSPORT_TIMED_OUT = 0x80130003,
    } NvstResult;

    /// Alias the type while transitioning. \sa NvstResult.
    typedef NvstResult NvssResult;

    /// Helper function which converts NvstResult into const char *
    NVST_API const char* nvstResultToString(NvstResult result);

    /// @}

#if defined(__cplusplus)
}
#endif
